#!/usr/bin/env python
import user
import importer
import base

import cPickle
import os
import sys
import traceback

class Scandal(base.Base):
  def __init__(self, clArgs = None, argDB = None):
    return base.Base.__init__(self, clArgs, argDB)

  def clear(self):
    self.debugger       = None
    self.parser       = None
    self.visitors       = []
    self.codeGenerators = {}
    self.fileOutputs    = {}
    self.fileReports    = {}
    self.outputFiles    = []
    return

  def importModule(moduleName):
    '''Import the named module, and return the module object'''
    module     = __import__(moduleName)
    components = moduleName.split('.')
    for comp in components[1:]:
      module = getattr(module, comp)
    return module
  importModule = staticmethod(importModule)

  def setupArgDB(self, argDB, clArgs):
    import nargs

    argDB.setType('debug',            nargs.ArgBool(None, 0, 'Flag to spawn the debugger'),        forceLocal = 1)
    argDB.setType('noStackTrace',     nargs.ArgBool(None, 1, 'Suppress the stack trace on error'), forceLocal = 1)
    argDB.setType('updateRepository', nargs.ArgBool(None, 0, 'Update the online repository of ASTs',         isTemporary = 1), forceLocal = 1)
    argDB.setType('includes',         nargs.Arg(None, [], 'List of include files',                           isTemporary = 1), forceLocal = 1) # TODO: Should be a list
    argDB.setType('client',           nargs.Arg(None, [], 'Client language or list of languages',            isTemporary = 1), forceLocal = 1)
    argDB.setType('clientDirs',       nargs.Arg(None, {}, 'Dictionary of client directories by language',    isTemporary = 1), forceLocal = 1) # TODO: Should be a dict
    argDB.setType('server',           nargs.Arg(None, [], 'Server language or list of languages',            isTemporary = 1), forceLocal = 1)
    argDB.setType('serverDirs',       nargs.Arg(None, {}, 'Dictionary of server directories by language',    isTemporary = 1), forceLocal = 1) # TODO: Should be a dict
    argDB.setType('implTransfer',     nargs.Arg(None, '', 'Exchange implementations between SIDL compilers', isTemporary = 1), forceLocal = 1)

    base.Base.setupArgDB(self, argDB, clArgs)
    return argDB

  def setupDebug(self):
    if not self.argDB['debug']: return
    import SIDL.Loader

    try:
      import PETSc.Debug.Debugger
    except ImportError, e:
      raise RuntimeError('Could not load Petsc debugger module: '+str(e))

    cls      = 'PETSc.DebugI.GDB.Debugger'
    debugger = PETSc.Debug.Debugger.Debugger(SIDL.Loader.createClass(cls))
    if debugger:
      debugger.setProgram('/usr/local/python/bin/python')
      debugger.setUseXterm(1)
      debugger.setDebugger('/usr/local/gdb-old/bin/gdb')
      debugger.attach()
    else:
      raise RuntimeError('Could not load debugger: '+cls)
    return debugger

  def setupParser(self):
    import SIDL.Loader
    import SIDLLanguage.Parser

    cls    = 'ANL.SIDLCompilerI.SIDLCompiler'
    parser = SIDLLanguage.Parser.Parser(SIDL.Loader.createClass(cls))
    if not parser:
      raise RuntimeError('Unable to instantiate SIDL parser: '+cls)
    return parser

  def loadVisitor(self, cls, defaultClass = None):
    import SIDL.Loader
    import SIDLLanguage.Visitor

    visitor = SIDLLanguage.Visitor.Visitor(SIDL.Loader.createClass(cls))
    if not visitor:
      if defaultClass:
        visitor = SIDLLanguage.Visitor.Visitor(SIDL.Loader.createClass(defaultClass))
        if not visitor:
          raise RuntimeError('Unable to instantiate visitor: '+cls+' or '+defaultClass)
      else:
        raise RuntimeError('Unable to instantiate visitor: '+cls)
    return visitor

  def getCodeGenerators(self, language):
    if not self.codeGenerators.has_key(language):
      import SIDLLanguage.CodeGenerator

      generators = []
      try:
        A = Scandal.importModule('ANL.CodeGenerationI.'+language)
      except:
        raise RuntimeError('Could not load any code generators for '+language)

      for a in A.all:
        l = self.loadVisitor('ANL.CodeGenerationI.'+language+'.'+a)
        try:
          l = SIDLLanguage.CodeGenerator.CodeGenerator(l)
          generators.append(l)
          self.debugPrint('Loaded code generator: '+'ANL.CodeGenerationI.'+language+'.'+a, 3, 'sidl')
        except: pass
      if len(generators) == 0:
        raise RuntimeError('Could not load any code generators for '+language)
      self.codeGenerators[language] = generators
      for generator in self.codeGenerators[language]:
        self.visitors.append(generator)
    return self.codeGenerators[language]

  def getFileOutput(self, language):
    if not self.fileOutputs.has_key(language):
      import SIDLLanguage.FileOutput
      outputters = []
      try:
        A = Scandal.importModule('ANL.CodeGenerationI.'+language)
      except ImportError, e:
        raise RuntimeError('Could not load any code generators for '+language+': '+str(e))

      for a in A.all:
        l = self.loadVisitor('ANL.CodeGenerationI.'+language+'.'+a)
        try:
          l = SIDLLanguage.FileOutput.FileOutput(l)
          outputters.append(l)
          self.debugPrint('Loaded code outputer: '+'ANL.CodeGenerationI.'+language+'.'+a, 3, 'sidl')
        except: pass
      if len(outputters) == 0:
        raise RuntimeError('Could not load any code file outputters for '+language)
      self.fileOutputs[language] = outputters
      for outputter in self.fileOutputs[language]:
        self.visitors.append(outputter)
    return self.fileOutputs[language]

  def getIOR(self):
    if not hasattr(self, 'ior'):
      import SIDLLanguage.CodeGenerator

      self.ior = SIDLLanguage.CodeGenerator.CodeGenerator(self.loadVisitor('ANL.CodeGenerationI.IOR.Generator'))
      self.ior.setPurposeActive('IOR', 1)
      self.visitors.append(self.ior)
      self.debugPrint('Loaded IOR generator', 3, 'sidl')
    return self.ior

  def getIORFileOutput(self):
    import SIDLLanguage.FileOutput

    fileOutput = SIDLLanguage.FileOutput.FileOutput(self.loadVisitor('ANL.CodeGenerationI.IOR.FileOutput'))
    self.visitors.append(fileOutput)
    self.debugPrint('Loaded IOR outputer', 3, 'sidl')
    return fileOutput

  def getIORFileReport(self):
    import SIDLLanguage.FileOutput

    fileReport = SIDLLanguage.FileOutput.FileOutput(self.loadVisitor('ANL.CodeGenerationI.IOR.FileReport'))
    self.visitors.append(fileReport)
    self.debugPrint('Loaded IOR reporter', 3, 'sidl')
    return fileReport

  def setupImplTransfer(self, fileOutput):
    if self.argDB['implTransfer']:
      import ANL.SplicerI.Splicer
      import ANL.SplicerI.TransferMode

      splicer = ANL.SplicerI.Splicer.Splicer(fileOutput.getSplicer())
      splicer.setDoTransfer(1)
      mode    = self.argDB['implTransfer']
      if mode == 'babelToScandal' or mode == '1':
        splicer.setTransferMode(ANL.SplicerI.TransferMode.BABEL_TO_SCANDAL)
      elif self.argDB['implTransfer'] == 'scandalToBabel':
        splicer.setTransferMode(ANL.SplicerI.TransferMode.SCANDAL_TO_BABEL)
      else:
        raise RuntimeError('Invalid implementation transfer mode: '+str(mode))
    return

  def setupClient(self):
    if len(self.argDB['client']):
      ior = self.getIOR()
      ior.setDeclarationActive(1)
      clients = self.argDB['client']
      if not isinstance(clients, list): clients = [clients]
      for client in clients:
        clientLang = client.replace('+', 'x')
        try:
          clientDir = self.argDB['clientDirs'][client]
        except:
          self.debugPrint('Using default since clientDirs = \''+str(self.argDB['clientDirs'])+'\' client = '+client, 1, 'sidl')
          clientDir = os.path.abspath(client.lower())

        for codeGenerator in self.getCodeGenerators(clientLang):
          codeGenerator.setPurposeActive('stub', 1)
          codeGenerator.setPurposeActive('cart', 1)
          codeGenerator.setDeclarationActive(1)
          codeGenerator.setDefinitionActive(1)

        fileOutput = self.getIORFileOutput()
        fileOutput.setRoot('IOR', clientDir)
        fileOutput.setOutputSource(0)
        for fileOutput in self.getFileOutput(clientLang):
          fileOutput.setRoot('stub', clientDir)
          fileOutput.setRoot('cart', clientDir)
    return

  def setupServer(self):
    if len(self.argDB['server']):
      ior = self.getIOR()
      ior.setDeclarationActive(1)
      ior.setDefinitionActive(1)
      servers = self.argDB['server']
      if not isinstance(servers, list): servers = [servers]
      for server in servers:
        serverLang = server.replace('+', 'x')
        try:
          serverDir = self.argDB['serverDirs'][server]
        except:
          serverDir = os.path.abspath('server-'+server.lower())
        for codeGenerator in self.getCodeGenerators(serverLang):
          codeGenerator.setPurposeActive('skel', 1)
          codeGenerator.setPurposeActive('impl', 1)
          codeGenerator.setDeclarationActive(1)
          codeGenerator.setDefinitionActive(1)

        fileOutput = self.getIORFileOutput()
        fileOutput.setRoot('IOR', serverDir)
        for fileOutput in self.getFileOutput(serverLang):
          fileOutput.setRoot('skel', serverDir)
          fileOutput.setRoot('impl', serverDir)
          self.setupImplTransfer(fileOutput)
    return

  def setupReportClient(self):
    if len(self.argDB['client']):
      clients = self.argDB['client']
      if not isinstance(clients, list): clients = [clients]
      for client in clients:
        clientLang = client.replace('+', 'x')
        try:
          clientDir = self.argDB['clientDirs'][client]
        except:
          self.debugPrint('Using default since clientDirs = \''+str(self.argDB['clientDirs'])+'\' client = '+client, 1, 'sidl')
          clientDir = os.path.abspath(client.lower())

        fileReport = self.getIORFileReport()
        fileReport.setRoot('IOR', clientDir)
        fileReport.setOutputSource(0)
        for fileReport in self.getFileOutput(clientLang):
          fileReport.setRoot('stub', clientDir)
          fileReport.setRoot('cart', clientDir)
    return

  def setupReportServer(self):
    if len(self.argDB['server']):
      servers = self.argDB['server']
      if not isinstance(servers, list): servers = [servers]
      for server in servers:
        serverLang = server.replace('+', 'x')
        try:
          serverDir = self.argDB['serverDirs'][server]
        except:
          serverDir = os.path.abspath('server-'+server.lower())

        fileReport = self.getIORFileReport()
        fileReport.setRoot('IOR', serverDir)
        for fileReport in self.getFileOutput(serverLang):
          fileReport.setRoot('skel', serverDir)
          fileReport.setRoot('impl', serverDir)
    return

  def checkRun(self):
    if self.argDB['updateRepository']:
      # This should update the online repository of ASTs
      return 0
    # Check for loader and parser
    try:
      import SIDL.Loader
      import SIDLLanguage.Parser
    except ImportError, e:
      raise RuntimeError('Cannot locate libraries for SIDL compiler: '+str(e))
    return 1

  def setupResolution(self):
    self.debugger = self.setupDebug()
    self.parser = self.setupParser()
    if len(self.argDB['client']) or len(self.argDB['server']):
      self.visitors.append(self.loadVisitor('ANL.CodeGenerationI.Resolver'))
      self.debugPrint('Loaded resolver', 3, 'sidl')
      self.visitors.append(self.loadVisitor('ANL.CodeGenerationI.DependencyAnalyzer'))
      self.debugPrint('Loaded dependency analyzer', 3, 'sidl')
      if self.argDB['client'] in ['F90', 'Mathematica'] or self.argDB['server'] in ['F90', 'Mathematica']:
        self.visitors.append(self.loadVisitor('ANL.CodeGenerationI.InitConstructor'))
        self.debugPrint('Loaded constructor creator', 3, 'sidl')
    return

  def setupRun(self):
    self.setupResolution()
    self.setupClient()
    self.setupServer()
    return

  def setupReport(self):
    self.setupResolution()
    self.setupReportClient()
    self.setupReportServer()
    return

  def processIncludes(self):
    '''Read the include filenames from the argument database and return a list of AST objects'''
    includes = []
    for include in self.argDB['includes']:
      includes.append(self.parser.parseFile(include))
    return includes

  def processTargets(self):
    '''Read the target filenames from the argument database'''
    if self.argDB.target == ['default']:
      print 'Usage: '+sys.argv[0]+' [options] <sidl files>'
      sys.exit(1)
    trees = []
    for target in self.argDB.target:
      if os.path.isfile(target):
        trees.append(self.parser.parseFile(target))
    return trees

  def processTree(self):
    import SIDL.BaseException
    import SIDLLanguage.ASTException
    import SIDLLanguage.FileOutput

    includes  = self.processIncludes()
    trees     = self.processTargets()
    includes += trees

    try:
      for visitor in self.visitors:
        visitor.setRepository(includes)
        if visitor.isInstanceOf('SIDLLanguage.FileOutput'):
          SIDLLanguage.FileOutput.FileOutput(visitor).clearOutputFiles()
        for tree in trees:
          tree.accept(visitor)
        if visitor.isInstanceOf('SIDLLanguage.FileOutput'):
          self.outputFiles.extend(SIDLLanguage.FileOutput.FileOutput(visitor).getOutputFiles())
    except (SIDL.BaseException.Exception, SIDLLanguage.ASTException.Exception), e:
      print e.getMessage()
      if not self.argDB['noStackTrace']:
        sys.stderr.write(e.getStackTrace())
        print traceback.print_tb(sys.exc_info()[2])
      sys.exit(1)
    return

  def run(self):
    self.clear()
    if not self.checkRun(): return
    self.setupRun()
    self.processTree()
    return

  def report(self):
    self.clear()
    if not self.checkRun(): return
    self.setupReport()
    self.processTree()
    return

if __name__ == '__main__': Scandal(sys.argv[1:]).run()
