#!/usr/bin/env python
import user
import build.framework
import project

import os

class PetscMake(build.framework.Framework):
  def __init__(self, clArgs = None, argDB = None):
    build.framework.Framework.__init__(self, project.Project('bk://sidl.bkbits.net/Compiler', self.getRoot()), clArgs, argDB)
    self.project.setWebDirectory('petsc@terra.mcs.anl.gov://mcs/www-unix/ase')
    if not self.argDB.has_key('installedLanguages'):
      self.argDB['installedLanguages'] = []
    if not self.argDB.has_key('clientLanguages'):
      self.argDB['clientLanguages'] = []
    return

  def t_getDependencies(self):
    return ['bk://ply.bkbits.net/ply-dev', 'bk://sidl.bkbits.net/Runtime']

  def t_install(self):
    installedLanguages = self.argDB['installedLanguages']
    clientLanguages    = self.argDB['clientLanguages']
    for language in ['Python', 'Cxx']:
      if not language in installedLanguages:
        installedLanguages.append(language)
      if not language in clientLanguages:
        clientLanguages.append(language)
    self.argDB['installedLanguages'] = installedLanguages
    self.argDB['clientLanguages']    = clientLanguages
    return build.framework.Framework.t_install(self)

  def defineSIDL(self):
    '''Define SIDL files sets before activation since they are necessary to calculate the compiler template'''
    import build.fileset
    self.filesets['sidl']  = build.fileset.ExtensionFileSet(os.path.join(self.project.getRoot(), 'sidl'), '.sidl')
    self.filesets['etags'] = build.fileset.FileSet(['TAGS'], mustExist = 0)
    # Must build pythonGenerator first
    generator = os.path.join(os.path.join(self.project.getRoot(), 'sidl'), 'pythonGenerator.sidl')
    self.filesets['sidl'].remove(generator)
    self.filesets['sidl'].insert(0, generator)
    return

  def defineSource(self):
    '''Define rooted files sets here since they require project activation'''
    import build.fileset
    self.filesets['python'] = build.fileset.RootedFileSet(self.project.getUrl(),
                                                          map(lambda f: os.path.join(self.sidlTemplate.usingSIDL.getServerRootDir('Python', 'pythonGenerator'), f),
                                                              ['SIDL_Python.c', 'SIDL_Python.h']), tag = 'python server pythonGenerator')
    self.filesets['sidl'].children.append(self.filesets['python'])
    return

  def setupProject(self):
    self.defineSIDL()
    return

  def setupBuild(self):
    self.defineSource()

    self.sidlTemplate.addServer('Python')
    return

if __name__ ==  '__main__':
  import sys
  PetscMake(sys.argv[1:]).main()
