#!/usr/bin/env python

# -----------------------------------------------------------------------------
# sidlparse.py
#
# Simple parser for SIDL
# -----------------------------------------------------------------------------
import user
import importer
import SIDLLanguage.ParameterAccessMode
import ANL.SIDLCompilerI.Comment
import ANL.SIDLCompilerI.Specification
import ANL.SIDLCompilerI.Version
import ANL.SIDLCompilerI.Package
import ANL.SIDLCompilerI.Interface
import ANL.SIDLCompilerI.Class
import ANL.SIDLCompilerI.Method
import ANL.SIDLCompilerI.Type
import ANL.SIDLCompilerI.Enumeration
import ANL.SIDLCompilerI.Enumerator
import ANL.SIDLCompilerI.Array
import ANL.SIDLCompilerI.Parameter

import yacc
import sidllex
import sys

# Get the token map
tokens = sidllex.tokens
sdebug = 0

def addNullIORException(node):
    exception = ANL.SIDLCompilerI.Class.Class()
    exception.setIdentifier('SIDL.NullIORException')
    exception.setBaseType(0)
    exception.setParent(node)
    node.addExceptions([exception])
    return node

# specification:

def p_specification_1(t):
    'specification : version_list import_list package_specifier_list'
    if sdebug: print "in p_specification_1 t[1] = %s" % t[1]
    spec = ANL.SIDLCompilerI.Specification.Specification()
    spec.addChildren(t[1])
    spec.addChildren(t[2])
    spec.addChildren(t[3])
    t[0] = spec

def p_specification_2(t):
    'specification : import_list package_specifier_list'
    if sdebug: print "in p_specification_2 t[1] = %s" % t[1]
    spec = ANL.SIDLCompilerI.Specification.Specification()
    spec.addChildren(t[1])
    spec.addChildren(t[2])
    t[0] = spec

def p_specification_3(t):
    'specification : version_list package_specifier_list'
    if sdebug: print "in p_specification_3 t[1] = %s t[2] = %s" % (t[1], t[2])
    spec = ANL.SIDLCompilerI.Specification.Specification()
    spec.addChildren(t[1])
    spec.addChildren(t[2])
    t[0] = spec

def p_specification_4(t):
    'specification : package_specifier_list'
    if sdebug: print "in p_specification_4 t[1] = %s" % t[1]
    spec = ANL.SIDLCompilerI.Specification.Specification()
    spec.addChildren(t[1])
    t[0] = spec

def p_specification_empty(t):
    'specification : empty'
    if sdebug: print "in p_specification_empty t[1] = %s" % t[1]
    spec = ANL.SIDLCompilerI.Specification.Specification()
    t[0] = spec

def p_version_list_1(t):
    'version_list : version_list version'
    if sdebug: print "in p_version_list_1: t[1] = %s" % t[1]
    t[0] = t[1]+[t[2]]

def p_version_list_2(t):
    'version_list : version'
    if sdebug: print "in p_version_list_2 t[1] = %s" % t[1]
    t[0] = [t[1]]

def p_version_1(t):
    'version : VERSION ID VERSION_STRING SEMI'
    if sdebug: print "in p_version_1 t[1]=%s, t[2]=%s, t[3]=%s, t[4]=%s" % (t[1],t[2],t[3],t[4])
    version = ANL.SIDLCompilerI.Version.Version()
    version.setIdentifier(t[2])
    version.setNumber(t[3])
    t[0] = version

def p_version_2(t):
    'version : VERSION ID ICONST_ALT SEMI'
    if sdebug: print "in p_version_1 t[1]=%s, t[2]=%s, t[3]=%s, t[4]=%s" % (t[1],t[2], t[3],t[4])
    version = ANL.SIDLCompilerI.Version.Version()
    version.setIdentifier(t[2])
    version.setNumber(t[3])
    t[0] = version

def p_version_3(t):
    'version : comment_list version'
    if sdebug: print "in p_version_3: t[1] = %s" % t[1]
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

def p_import_list_1(t):
    'import_list : import_list import'
    if sdebug: print "in p_import_list_1, t[1]=%s" % t[1]
    pass

def p_import_list_2(t):
    'import_list : import'
    if sdebug: print "in p_import_list_2, t[1]=%s" % t[1]
    t[0] = t[1]

def p_import(t):
    'import : IMPORT ID SEMI'
    if sdebug: print "in p_import, t[1]=%s, t[2]=%s, t[3]=%s" % (t[1],t[2],t[3])
    pass

def p_import_2(t):
    'import : comment_list import'
    if sdebug: print "in p_import_2, t[1]=%s, t[2]=%s" % (t[1],t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

# package-specifier
def p_package_specifier_list(t):
    'package_specifier_list : package_specifier_list package_specifier'
    if sdebug: print "in package_specifier_list: t[1]=%s, t[2]=%s" % (t[1],t[2])
    t[0] = t[1]+[t[2]]

def p_package_specifier_list_2(t):
    'package_specifier_list : package_specifier'
    if sdebug: print "in package_specifier_list_2: t[1]=%s" % t[1]
    t[0] = [t[1]]

def p_package_specifier(t):
    'package_specifier : PACKAGE ID LBRACE definition_list RBRACE optional_semi'
    if sdebug: print "in p_package_specifier t[1]=%s, t[2]=%s,t[3]=%s, t[4]=%s, t[5]=%s" % (t[1], t[2], t[3], t[4], t[5])
    package = ANL.SIDLCompilerI.Package.Package()
    package.setIdentifier(t[2])
    if t[4]:
        package.addChildren(t[4])
    t[0] = package

def p_package_specifier_2(t):
    'package_specifier : comment_list package_specifier'
    if sdebug: print "in p_package_specifier_2 t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

def p_definition_list_1(t):
    'definition_list : enum_specifier definition_list'
    if sdebug: print "in p_definition_list_1: t[1]=%s" % t[1]
    t[0] = [t[1]]+t[2]

def p_definition_list_2(t):
    'definition_list : class_specifier definition_list'
    if sdebug: print "in p_definition_list_2: t[1]=%s" % t[1]
    t[0] = [t[1]]+t[2]

def p_definition_list_3(t):
    'definition_list : interface_specifier definition_list'
    if sdebug: print "in p_definition_list_3: t[1]=%s" % t[1]
    t[0] = [t[1]]+t[2]

def p_definition_list_4(t):
    'definition_list : package_specifier definition_list'
    if sdebug: print "in p_definition_list_4: t[1]=%s" % t[1]
    t[0] = [t[1]]+t[2]

def p_definition_list_empty(t):
    'definition_list : empty'
    if sdebug: print "in p_definition_list_empty: t[1]=%s" % t[1]
    t[0] = []

# enum-specifier:
def p_enum_specifier_1(t):
    'enum_specifier : ENUM ID LBRACE enumerator_list RBRACE optional_semi'
    if sdebug: print "in p_enum_specifier_1"
    enum = ANL.SIDLCompilerI.Enumeration.Enumeration()
    enum.setIdentifier(t[2])
    enum.setBaseType(0)
    if t[4]:
        enum.addChildren(t[4])
    t[0] = enum

def p_enum_specifier_2(t):
    'enum_specifier : ENUM LBRACE enumerator_list RBRACE optional_semi'
    enum = ANL.SIDLCompilerI.Enumeration.Enumeration()
    enum.setBaseType(0)
    if t[3]:
        enum.addChildren(t[3])
    t[0] = enum

def p_enum_specifier_3(t):
    'enum_specifier : ENUM ID'
    enum = ANL.SIDLCompilerI.Enumeration.Enumeration()
    enum.setIdentifier(t[2])
    enum.setBaseType(0)
    t[0] = enum

def p_enum_specifier_4(t):
    'enum_specifier : comment_list enum_specifier'
    if sdebug: print "in p_enum_specifier_4 t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

# enumerator_list:
def p_enumerator_list_1(t):
    'enumerator_list : enumerator'
    t[0] = [t[1]]

def p_enumerator_list_2(t):
    'enumerator_list : enumerator_list COMMA enumerator'
    t[0] = t[1]+[t[3]]

# enumerator:
def p_enumerator_1(t):
    'enumerator : ID'
    enum = ANL.SIDLCompilerI.Enumerator.Enumerator()
    enum.setIdentifier(t[1])
    t[0] = enum

def p_enumerator_2(t):
    '''enumerator : ID EQUALS ICONST
                  | ID EQUALS ICONST_ALT'''
    enum = ANL.SIDLCompilerI.Enumerator.Enumerator()
    enum.setIdentifier(t[1])
    enum.setValue(int(t[3]))
    enum.setInitialized(1)
    t[0] = enum

# class-specifier
def p_class_specifier_1(t):
    'class_specifier : CLASS ID extends_specifier implements_specifier LBRACE class_declaration_list RBRACE optional_semi'
    if sdebug: print "in p_class_specifier_1 t[1]=%s, t[2]=%s,t[3]=%s, t[4]=%s, t[5]=%s, t[6]=%s, t[7]=%s" % (t[1], t[2], t[3], t[4], t[5], t[6], t[7])
    sclass = ANL.SIDLCompilerI.Class.Class()
    sclass.setIdentifier(t[2])
    sclass.setBaseType(0)
    if t[3]:
        interfaces = []
        for name in t[3]:
            interface = ANL.SIDLCompilerI.Interface.Interface()
            interface.setIdentifier(name)
            interface.setBaseType(0)
            interfaces.append(interface)
        sclass.addParents(interfaces)
        for node in interfaces:
            node.setParent(sclass)
    if t[4]:
        interfaces = []
        for name in t[4]:
            interface = ANL.SIDLCompilerI.Interface.Interface()
            interface.setIdentifier(name)
            interface.setBaseType(0)
            interfaces.append(interface)
        sclass.addInterfaces(interfaces)
        for node in interfaces:
            node.setParent(sclass)
    if t[6]:
        sclass.addChildren(t[6])
        for node in t[6]:
            if not node.getStatic():
                addNullIORException(node)
    t[0] = sclass

def p_class_specifier_2(t):
    'class_specifier : ABSTRACT CLASS ID extends_specifier implements_specifier LBRACE class_declaration_list RBRACE optional_semi'
    if sdebug: print "in p_class_specifier_2 t[1]=%s, t[2]=%s,t[3]=%s, t[4]=%s, t[5]=%s, t[6]=%s, t[7]=%1, t[8]=%s" % (t[1], t[2], t[3], t[4], t[5], t[6], t[7], t[8])
    sclass = ANL.SIDLCompilerI.Class.Class()
    sclass.setAbstract(1)
    sclass.setIdentifier(t[3])
    sclass.setBaseType(0)
    if t[4]:
        interfaces = []
        for name in t[4]:
            interface = ANL.SIDLCompilerI.Interface.Interface()
            interface.setIdentifier(name)
            interface.setBaseType(0)
            interfaces.append(interface)
        sclass.addParents(interfaces)
        for node in interfaces:
            node.setParent(sclass)
    if t[5]:
        interfaces = []
        for name in t[5]:
            interface = ANL.SIDLCompilerI.Interface.Interface()
            interface.setIdentifier(name)
            interface.setBaseType(0)
            interfaces.append(interface)
        sclass.addInterfaces(interfaces)
        for node in interfaces:
            node.setParent(sclass)
    if t[7]:
        sclass.addChildren(t[7])
        for node in t[7]:
            if not node.getStatic():
                addNullIORException(node)
    t[0] = sclass

def p_class_specifier_3(t):
    'class_specifier : comment_list class_specifier'
    if sdebug: print "in p_class_specifier_3 t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

# interface-specifier
def p_interface_specifier(t):
    'interface_specifier : INTERFACE ID extends_specifier LBRACE declaration_list RBRACE optional_semi'
    if sdebug: print "in p_interface_specifier: t[1]=%s, t[2]=%s,t[3]=%s, t[4]=%s, t[5]=%s, t[6]=%s" % (t[1], t[2], t[3], t[4], t[5], t[6])
    interface = ANL.SIDLCompilerI.Interface.Interface()
    interface.setIdentifier(t[2])
    interface.setBaseType(0)
    if t[3]:
        interfaces = []
        for name in t[3]:
            interf = ANL.SIDLCompilerI.Interface.Interface()
            interf.setIdentifier(name)
            interface.setBaseType(0)
            interfaces.append(interf)
        interface.addParents(interfaces)
        for node in interfaces:
            node.setParent(interface)
    if t[5]:
        interface.addChildren(t[5])
        for node in t[5]:
            addNullIORException(node)
    t[0] = interface

def p_interface_specifier_2(t):
    'interface_specifier : comment_list interface_specifier'
    if sdebug: print "in p_interface_specifier_2 t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = t[2]

def p_extends_specifier(t):
    'extends_specifier : EXTENDS scoped_identifier_list'
    if sdebug: print "in p_extends_specifier: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = t[2]

def p_extends_specifier_empty(t):
    'extends_specifier : empty'
    if sdebug: print "in p_extends_specifier_empty: t[1]=%s" % t[1]
    t[0] = []

def p_implements_specifier_1(t):
    'implements_specifier : IMPLEMENTSALL scoped_identifier_list'
    if sdebug: print "in p_implements_specifier_1: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = t[2]

def p_implements_specifier_2(t):
    'implements_specifier : IMPLEMENTS scoped_identifier_list'
    if sdebug: print "in p_implements_specifier_2: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = t[2]

def p_implements_specifier_empty(t):
    'implements_specifier : empty'
    if sdebug: print "in p_implements_specifier_empty: t[1]=%s" % t[1]
    t[0] = []

# class-declaration-list
def p_class_declaration_list_1(t):
    'class_declaration_list : class_method class_declaration_list'
    if sdebug: print "in p_class_declaration_list_1: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = [t[1]]+t[2]

def p_class_declaration_list_2(t):
    'class_declaration_list : method class_declaration_list'
    if sdebug: print "in p_class_declaration_list_2: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = [t[1]]+t[2]

def p_class_declaration_list_3(t):
    'class_declaration_list : comment_list class_method class_declaration_list'
    if sdebug: print "in p_class_declaration_list_3: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = [t[2]]+t[3]

def p_class_declaration_list_4(t):
    'class_declaration_list : comment_list method class_declaration_list'
    if sdebug: print "in p_class_declaration_list_4: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = [t[2]]+t[3]

def p_class_declaration_list_empty(t):
    'class_declaration_list : empty'
    if sdebug: print "in p_class_declaration_list_empty: t[1]=%s" % t[1]
    t[0] = []

# class-method
def p_class_method_1(t):
    'class_method : ABSTRACT method'
    if sdebug: print "in p_class_method_1: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].setAbstract(1)
    t[0] = t[2]
    
def p_class_method_2(t):
    'class_method : FINAL method'
    if sdebug: print "in p_class_method_2: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].setFinal(1)
    t[0] = t[2]
    
def p_class_method_3(t):
    'class_method : STATIC method'
    if sdebug: print "in p_class_method_3: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].setStatic(1)
    t[0] = t[2]
    
# declaration-list
def p_declaration_list_1(t):
    'declaration_list : method declaration_list'
    if sdebug: print "in p_declaration_list_1: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[0] = [t[1]]+t[2]

def p_declaration_list_2(t):
    'declaration_list : comment_list method declaration_list'
    if sdebug: print "in p_declaration_list_2: t[1]=%s, t[2]=%s" % (t[1], t[2])
    t[2].addComments(t[1])
    for node in t[1]:
        node.setParent(t[2])
    t[0] = [t[2]]+t[3]

def p_declaration_list_empty(t):
    'declaration_list : empty'
    if sdebug: print "in p_declaration_list_empty: t[1]=%s" % t[1]
    t[0] = []

# method
def p_method_1(t):
    'method : return_type_specifier ID LPAREN parameter_list RPAREN local_qualifier throws_specifier SEMI'
    t[0] = createMethod(t[1], t[2], t[4], t[7])
    return

def p_method_2(t):
    'method : return_type_specifier ID LBRACKET ID RBRACKET LPAREN parameter_list RPAREN local_qualifier throws_specifier SEMI'
    t[0] = createMethod(t[1], t[2], t[7], t[10], [t[4]])
    return

def createMethod(returnType, name, parameters, exceptions, templateIdentifiers = None):
    method = ANL.SIDLCompilerI.Method.Method()
    returnType.setParent(method)
    method.setReturnType(returnType)
    method.setIdentifier(name)
    if parameters:
        method.addChildren(parameters)
    if exceptions:
        excs = []
        for id in exceptions:
            exception = ANL.SIDLCompilerI.Class.Class()
            exception.setIdentifier(id)
            exception.setBaseType(0)
            excs.append(exception)
        method.addExceptions(excs)
        for node in excs:
            node.setParent(method)
    if not templateIdentifiers is None:
        method.setTemplateIdentifiers(templateIdentifiers)
    return method

def p_return_type_specifier(t):
    '''return_type_specifier : type_specifier 
                             | REPLICATE type_specifier
                             | VOID'''
    if sdebug: print "in p_return_type_specifier: t[1]=%s" % t[1]
    if t[1] == 'replicate':
        t[0] = t[2]
    elif t[1] == 'void':
        type = ANL.SIDLCompilerI.Type.Type()
        type.setIdentifier(t[1])
        type.setBaseType(1)
        t[0] = type
    else:
        t[0] = t[1]

def p_local_qualifier_1(t):
    'local_qualifier : LOCAL'
    if sdebug: print "in p_local_qualifier_1: t[1]=%s" % t[1]    
    t[0] = t[1]

def p_local_qualifier_2(t):
    'local_qualifier : ONEWAY'
    if sdebug: print "in p_local_qualifier_2: t[1]=%s" % t[1]    
    t[0] = t[1]

def p_local_qualifier_empty(t):
    'local_qualifier : empty'
    if sdebug: print "in p_local_qualifier_empty: t[1]=%s" % t[1]    
    t[0] = None

def p_throws_specifier(t):
    'throws_specifier : THROWS scoped_identifier_list'
    if sdebug: print "in p_throws_specifier: t[1]=%s, t[2]=%s" % (t[1],t[2])
    t[0] = t[2]

def p_throws_specifier_empty(t):
    'throws_specifier : empty'
    if sdebug: print "in p_throws_specifier_empty: t[1]=%s" % t[1]    
    t[0] = None

# parameter-list
def p_parameter_list_1(t): 
    'parameter_list : parameter COMMA parameter_list'
    if sdebug: print "in p_parameter_list_1: t[1]=%s, t[2]=%s, t[3]=%s" % (t[1],t[2],t[3])
    t[0] = [t[1]]+t[3]

def p_parameter_list_2(t): 
    'parameter_list : parameter'
    if sdebug: print "in p_parameter_list_2: t[1]=%s" % t[1]    
    t[0] = [t[1]]

def p_parameter_list_empty(t): 
    'parameter_list : empty'
    if sdebug: print "in p_parameter_list_empty: t[1]=%s" % t[1]    
    t[0] = []

def p_parameter_1(t):
    'parameter : REPLICATE access_specifier type_specifier ID'
    if sdebug: print "in p_parameter_1: t[1]=%s, t[2]=%s, t[3]=%s, t[4]=%s" % (t[1],t[2],t[3],t[4])
    parameter = ANL.SIDLCompilerI.Parameter.Parameter()
    parameter.setIdentifier(t[4])
    t[3].setParent(parameter)
    parameter.setType(t[3])
    parameter.setAccessMode(int(t[2]))
    parameter.setRequired(1)
    parameter.setCopy(1)
    t[0] = parameter

def p_parameter_2(t):
    'parameter : access_specifier type_specifier ID'
    if sdebug: print "in p_parameter_2: t[1]=%s, t[2]=%s, t[3]=%s" % (t[1],t[2],t[3])
    parameter = ANL.SIDLCompilerI.Parameter.Parameter()
    parameter.setIdentifier(t[3])
    t[2].setParent(parameter)
    parameter.setType(t[2])
    parameter.setAccessMode(int(t[1]))
    parameter.setRequired(1)
    parameter.setCopy(0)
    t[0] = parameter

def p_access_specifier(t):
    '''access_specifier : IN
                        | INOUT
                        | OUT'''
    if sdebug: print "in p_access_specifier: t[1]=%s" % t[1]
    if t[1] == 'in':
        t[0] = SIDLLanguage.ParameterAccessMode.IN
    elif t[1] == 'inout':
        t[0] = SIDLLanguage.ParameterAccessMode.INOUT
    elif t[1] == 'out':
        t[0] = SIDLLanguage.ParameterAccessMode.OUT
    else:
        raise RuntimeError('Invalid access specifier: '+t[1])

# type-specifier:
def p_type_specifier_1(t):
    '''type_specifier : BOOL
                      | CHAR
                      | DCOMPLEX
                      | DOUBLE
                      | FCOMPLEX
                      | FLOAT
                      | INT
                      | LONG
                      | OPAQUE
                      | STRING
                      | TYPEID
                      '''
    type = ANL.SIDLCompilerI.Type.Type()
    type.setIdentifier(t[1])
    type.setBaseType(1)
    t[0] = type

def p_type_specifier_2(t):
    '''type_specifier : scoped_identifier'''
    type = ANL.SIDLCompilerI.Type.Type()
    type.setIdentifier(t[1])
    type.setBaseType(0)
    t[0] = type

def p_type_specifier_3(t):
    '''type_specifier : array_specifier'''
    t[0] = t[1]

# array-specifier

def p_array_specifier_1(t):
    'array_specifier : ARRAY LT type_specifier COMMA ICONST_ALT GT'
    if sdebug: print "in array_specifier_1: t[1]=%s, t[2]=%s, t[3]=%s, t[4]=%s, t[5]=%s, t[6]=%s" % (t[1],t[2],t[3],t[4],t[5],t[6])
    array = ANL.SIDLCompilerI.Array.Array()
    array.setIdentifier('array')
    array.setBaseType(0)
    t[3].setParent(array)
    array.setType(t[3])
    array.setDimension(int(t[5]))
    t[0] = array

def p_array_specifier_2(t):
    'array_specifier : ARRAY LT type_specifier GT'
    if sdebug: print "in array_specifier_2: t[1]=%s, t[2]=%s, t[3]=%s, t[4]=%s" % (t[1],t[2],t[3],t[4])
    array = ANL.SIDLCompilerI.Array.Array()
    array.setIdentifier('array')
    array.setBaseType(0)
    t[3].setParent(array)
    array.setType(t[3])
    t[0] = array

# scoped-identifier-list
def p_scoped_identifier_list_1(t):
    'scoped_identifier_list : scoped_identifier_list COMMA scoped_identifier'
    if sdebug: print "in p_scoped_identifier_list"
    t[0] = t[1]+[t[3]]

def p_scoped_identifier_list_2(t):
    'scoped_identifier_list : scoped_identifier'
    if sdebug: print "in scoped_identifier_list_2: t[1] = %s" % t[1]
    t[0] = [t[1]]

def p_scoped_identifier(t):
    '''scoped_identifier : SCOPEDID
                               | ID'''
    if sdebug: print "in scoped_identifier: t[1] = %s" % t[1]
    t[0] = t[1]

def p_optional_semi(t):
    '''optional_semi : SEMI 
                     | empty''' 
    pass

def p_comment_list_1(t):
    'comment_list : comment_list comment'
    if sdebug: print "in comment_list_1: t[1] = %s" % t[1]
    t[0] = t[1] + [t[2]]

def p_comment_list_2(t):
    'comment_list : comment'
    if sdebug: print "in comment_list_2: t[1] = %s" % t[1]
    t[0] = [t[1]]

def p_comment(t):
    'comment : COMMENT'
    if sdebug: print "in comment: t[1] = %s" % t[1]
    comment = ANL.SIDLCompilerI.Comment.Comment()
    comment.setComment(t[1])
    t[0] = comment

def p_empty(t):
    'empty : '
    pass

# Parse error
def p_error(t):
    if t is None:
        print 'Error: None, possibly a missing closing brace'
    else:
        print 'Error: token %s, line number %d' % (t, t.lineno)

# Driver
def setup():
    global parser

    parser = yacc.yacc(debug=1)

if __name__ == '__main__':
    import ANL.SIDL.PrettyPrinter

    #import profile
    # Build the grammar
    #profile.run("yacc.yacc()")

    setup()

    for i in range(1, len(sys.argv)):
        print "About to parse %s" % sys.argv[i]
        f = open(sys.argv[i],"r")
        s = f.read()
        f.close()
        # print "Contents of %s: %s" % (sys.argv[i], s)
        ast = parser.parse(s)
        ast.accept(ANL.SIDL.PrettyPrinter.PrettyPrinter())
