# DO-NOT-DELETE splicer.begin(_initial)
import SIDL.Loader
import SIDLLanguage.CodePurpose
# DO-NOT-DELETE splicer.end(_initial)

import SIDL.NullIORException
import SIDL.BaseInterface
import SIDL.BaseException
import SIDLLanguage.AST
import SIDLLanguage.VisitorException
import SIDLLanguage.Specification
import SIDLLanguage.Version
import SIDLLanguage.Package
import SIDLLanguage.Interface
import SIDLLanguage.Class
import SIDLLanguage.Method
import SIDLLanguage.Type
import SIDLLanguage.Enumeration
import SIDLLanguage.Enumerator
import SIDLLanguage.Array
import SIDLLanguage.Parameter
import SIDLLanguage.Comment

# DO-NOT-DELETE splicer.begin(_before_type)
import sys
# DO-NOT-DELETE splicer.end(_before_type)

class Generator:
  '''
  // This class generates all code for the IOR
  '''
  # All calls to SIDL methods should use __IORself

  def __init__(self, IORself):
    self.__IORself = IORself
    # DO-NOT-DELETE splicer.begin(__init__)
    self.language      = 'C'
    self.indentString  = '  '
    self.markName      = 'iorGeneratorMark'
    self.typeOnly      = [0]
    self.purposeActive = {'stub': [0],
                          'IOR':  [0],
                          'skel': [0],
                          'impl': [0]}
    self.declActive    = 0
    self.defnActive    = 0
    self.IOR           = SIDLLanguage.CodePurpose.CodePurpose(SIDL.Loader.createClass('ANL.SIDLCompilerI.IOR'))
    self.baseException          = None
    self.generatedBaseException = 0
    # Map SIDL types to IOR types
    self.typeMap = {"bool":     "SIDL_bool",
                    "char":     "char",
                    "dcomplex": "struct SIDL_dcomplex",
                    "double":   "double",
                    "fcomplex": "struct SIDL_fcomplex",
                    "float":    "float",
                    "int":      "int32_t",
                    "long":     "int64_t",
                    "opaque":   "void *",
                    "string":   "char *",
                    "void":     "void"}
    self.defaultValues = {"bool":     "FALSE",
                          "char":     "'\\0'",
                          "dcomplex": "{0.0, 0.0}",
                          "double":   "0.0",
                          "fcomplex": "{0.0, 0.0}",
                          "float":    "0.0",
                          "int":      "0",
                          "long":     "0L",
                          "opaque":   "NULL",
                          "string":   "NULL",
                          "void":     ""}
    # DO-NOT-DELETE splicer.end(__init__)

  def getPurposeActive(self, purpose):
    '''
    // Get the activity state of the given generation purpose
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # purpose string
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getPurposeActive)
    return self.purposeActive[purpose][-1]
    # DO-NOT-DELETE splicer.end(getPurposeActive)

  def setPurposeActive(self, purpose, isActive):
    '''
    // Set the activity state of the given generation purpose
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # purpose string
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setPurposeActive)
    if self.purposeActive.has_key(purpose):
      self.purposeActive[purpose][-1] = isActive
    else:
      self.purposeActive[purpose] = [isActive]
    # DO-NOT-DELETE splicer.end(setPurposeActive)

  def getDeclarationActive(self):
    '''
    // Get the flag determining whether to generate declarations (headers)
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getDeclarationActive)
    return self.declActive
    # DO-NOT-DELETE splicer.end(getDeclarationActive)

  def setDeclarationActive(self, isActive):
    '''
    // Set the flag determining whether to generate declarations (headers)
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setDeclarationActive)
    self.declActive = isActive
    # DO-NOT-DELETE splicer.end(setDeclarationActive)

  def getDefinitionActive(self):
    '''
    // Get the flag determining whether to generate definitions (source)
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getDefinitionActive)
    return self.defnActive
    # DO-NOT-DELETE splicer.end(getDefinitionActive)

  def setDefinitionActive(self, isActive):
    '''
    // Set the flag determining whether to generate definitions (source)
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setDefinitionActive)
    self.defnActive = isActive
    # DO-NOT-DELETE splicer.end(setDefinitionActive)

  def getRepository(self):
    '''
    // Retrieve auxilliary ASTs
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # array _return
    # DO-NOT-DELETE splicer.begin(getRepository)
    return []
    # DO-NOT-DELETE splicer.end(getRepository)

  def setRepository(self, repository):
    '''
    // Set auxilliary ASTs
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # repository array
    # DO-NOT-DELETE splicer.begin(setRepository)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(setRepository)

  def visitAST(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.AST
    # DO-NOT-DELETE splicer.begin(visitAST)
    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitAST)

  def visitSpecification(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Specification
    # DO-NOT-DELETE splicer.begin(visitSpecification)
    if self.isMarked(node): return
    self.markNode(node)

    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitSpecification)

  def visitVersion(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Version
    # DO-NOT-DELETE splicer.begin(visitVersion)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitVersion)

  def visitPackage(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Package
    # DO-NOT-DELETE splicer.begin(visitPackage)
    if self.isMarked(node): return
    self.markNode(node)

    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitPackage)

  def visitInterface(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Interface
    # DO-NOT-DELETE splicer.begin(visitInterface)
    self.generateTypeInformation(node)
    self.generateInterfaceAccessors(node);

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    # All method code must be generated
    try:
      for interface in node.getParents():
        interface.accept(self.__IORself)
      for dep in node.getDependencies():
        if dep.isInstanceOf('SIDLLanguage.Enumeration'):
          dep.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    self.defaultVisit(node)

    purpose = self.IOR
    if self.getPurposeActive(purpose.getName()):
      self.generateSource(node)
      self.generateHeader(node, purpose)
    # DO-NOT-DELETE splicer.end(visitInterface)

  def visitClass(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Class
    # DO-NOT-DELETE splicer.begin(visitClass)
    self.generateTypeInformation(node)
    self.generateInterfaceAccessors(node);

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    # All method code must be generated
    try:
      for interface in node.getInterfaces():
        interface.accept(self.__IORself)
      for interface in node.getParents():
        interface.accept(self.__IORself)
      for dep in node.getDependencies():
        if dep.isInstanceOf('SIDLLanguage.Enumeration'):
          dep.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    self.defaultVisit(node)

    purpose = self.IOR
    if self.getPurposeActive(purpose.getName()):
      self.generateSource(node)
      self.generateHeader(node, purpose)
    # DO-NOT-DELETE splicer.end(visitClass)

  def visitMethod(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Method
    # DO-NOT-DELETE splicer.begin(visitMethod)
    if self.isMarked(node): return
    self.markNode(node)

    try:
      self.typeOnly.append(1)
      node.getReturnType().accept(self.__IORself)
      for exception in node.getExceptions():
        exception.accept(self.__IORself)
        self.generateBaseException(exception)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    self.defaultVisit(node)

    purpose = self.IOR
    if self.getPurposeActive(purpose.getName()):
      self.generateMethodCast(node)
    # DO-NOT-DELETE splicer.end(visitMethod)

  def visitType(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Type
    # DO-NOT-DELETE splicer.begin(visitType)
    if self.isMarked(node): return
    self.markNode(node)

    self.generateTypeInformation(node)
    # DO-NOT-DELETE splicer.end(visitType)

  def visitEnumeration(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumeration
    # DO-NOT-DELETE splicer.begin(visitEnumeration)
    self.generateTypeInformation(node)

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    purpose = self.IOR
    if self.getPurposeActive(purpose.getName()):
      self.generateSource(node)
      self.generateHeader(node, purpose)
    # DO-NOT-DELETE splicer.end(visitEnumeration)

  def visitEnumerator(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumerator
    # DO-NOT-DELETE splicer.begin(visitEnumerator)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitEnumerator)

  def visitArray(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Array
    # DO-NOT-DELETE splicer.begin(visitArray)
    if self.isMarked(node): return
    self.markNode(node)

    try:
      self.typeOnly.append(1)
      node.getType().accept(self.__IORself)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    self.generateTypeInformation(node)
    # DO-NOT-DELETE splicer.end(visitArray)

  def visitParameter(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Parameter
    # DO-NOT-DELETE splicer.begin(visitParameter)
    if self.isMarked(node): return
    self.markNode(node)

    try:
      self.typeOnly.append(1)
      node.getType().accept(self.__IORself)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    name = node.getIdentifier()
    mode = node.getAccessMode()

    iorParam = self.getCode('ior parameter')
    iorParam.write(self.getModeComment(mode))
    if not self.isOut(mode) and node.getType().getBaseType() and node.getType().getIdentifier() == 'string':
      iorParam.write('const ')
    self.composeCode(iorParam, node.getType(), 'ior type')
    if self.isOut(mode):
      iorParam.write(' *'+name)
    else:
      iorParam.write(' '+name)
    node.addCode(iorParam)
    # DO-NOT-DELETE splicer.end(visitParameter)

  def visitComment(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Comment
    # DO-NOT-DELETE splicer.begin(visitComment)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitComment)

# DO-NOT-DELETE splicer.begin(_final)
  ##########################
  # Basic Visitor Operations
  def raiseVisitorException(self, message, trace = ''):
    e = SIDLLanguage.VisitorException.VisitorException()
    e.setMessage(message)
    if isinstance(trace, list):
      for line in trace: e.addToStackTrace(line.strip())
    elif trace:
      e.addToStackTrace(trace)
    e.addToTrace(__file__, sys._getframe(1).f_lineno, sys._getframe().f_code.co_name)
    raise SIDLLanguage.VisitorException.Exception, e

  def defaultVisit(self, node):
    try:
      for child in node.getChildren():
        child.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException('In '+str(node.getIdentifier())+': '+e.getMessage(), e.getStackTrace().strip())
    return

  def isMarked(self, node, name = ''):
    if not name: name = self.markName
    return node.hasAttribute(name)

  def markNode(self, node, name = ''):
    if not name: name = self.markName
    seen = SIDLLanguage.Attribute.Attribute(SIDL.Loader.createClass('ANL.SIDLCompilerI.Attribute'))
    seen.setName(name)
    node.addAttribute(seen)
    return

  def unmarkNode(self, node, name = ''):
    if not name: name = self.markName
    if node.hasAttribute(name):
      node.deleteAttribute(name)
    return

  def unmarkNetwork(self, node, name = ''):
    self.unmarkNode(node, name)
    for interface in node.getParents():
      self.unmarkNetwork(interface, name)
    if node.isInstanceOf('SIDLLanguage.Class'):
      for interface in SIDLLanguage.Class.Class(node).getInterfaces():
        self.unmarkNetwork(interface, name)
    return

  ##################################
  # Basic Code Generation Operations
  def getCode(self, label = '', language = ''):
    code = SIDLLanguage.Code.Code(SIDL.Loader.createClass('ANL.SIDLCompilerI.Code'))
    if language:
      code.setLanguage(language)
    else:
      code.setLanguage(self.language)
    if label:
      code.setLabel(label)
    return code

  def isIn(self, mode):
    return mode == SIDLLanguage.ParameterAccessMode.IN or mode == SIDLLanguage.ParameterAccessMode.INOUT

  def isOut(self, mode):
    return mode == SIDLLanguage.ParameterAccessMode.OUT or mode == SIDLLanguage.ParameterAccessMode.INOUT

  def isBase(self, interface):
    return interface.getFullIdentifier() in ['SIDL.BaseInterface', 'SIDL.BaseClass', 'SIDL.BaseException', 'SIDL.NullIORException', 'SIDL.Loader', 'SIDL.DLL']

  def hasStaticMethod(self, interface):
    for method in interface.getMethods():
      if method.getStatic() and not method.getInternal():
        return 1
    return 0

  def hasMethodException(self, interface):
    for method in interface.getMethods():
      if len(self.getIORExceptions(method)) > 0:
        return 1
    return 0

  #########################
  # Should go in Code Class
  def composeCode(self, parent, node, label, fail = 1, forceBreak = 0):
    if node.hasCodeBlock(self.language, label):
      self.composeCodeBlock(parent, node.getCodeBlock(self.language, label), forceBreak)
    elif fail:
      self.raiseVisitorException('Unable to find code block "'+label+'" in '+node.getFullIdentifier())
    return

  def composeCodeBlock(self, parent, code, forceBreak = 0):
    indent = ''
    for i in range(parent.getTabLevel()):
      indent += self.indentString
    source = code.getSource()
    if indent and (source.find('\n') >= 0 or forceBreak):
      # Indent start and all newlines
      source = indent+source.replace('\n', '\n'+indent)
      # Remove extra space from the end
      index  = 0
      while source[index-1] == ' ': index -= 1
      if index < 0:
        source = source[:index]
    if forceBreak and not source[-1] == '\n':
      source = source+'\n'
    parent.write(source)
    return parent

  # HACK: This is just until I rewrite everything to use augmented ASTs
  def hasCtorNode(self, sclass):
    for method in sclass.getImplementedMethods():
      if method.getIdentifier() == '_ctor':
        return 1
    return 0

  #############################
  # C specific code structures
  def getModeComment(self, mode):
    if mode == SIDLLanguage.ParameterAccessMode.IN:
      return '/*in*/ '
    elif mode == SIDLLanguage.ParameterAccessMode.INOUT:
      return '/*inout*/ '
    elif mode == SIDLLanguage.ParameterAccessMode.OUT:
      return '/*out*/ '
    else:
      self.raiseVisitorException('Invalid parameter access mode: '+str(mode))

  #########################
  # IOR specific methods
  def getIORName(self, type):
    return type.getFullIdentifier().replace('.', '_')

  def getIORArrayName(self, type):
    if type.getBaseType():
      return 'struct SIDL_'+type.getIdentifier()+'__array *'
    else:
      return 'struct '+self.getIORName(type)+'__array *'

  def getBaseObject(self, node, identifier):
    # Recurse to parents
    for interface in node.getParents():
      base = self.getBaseObject(interface, identifier)
      if base: return base
    # Find SIDL package
    parent = node.getParent()
    if parent.isInstanceOf('SIDLLanguage.Package') and parent.getFullIdentifier() == 'SIDL':
      for child in parent.getChildren():
        if child.getFullIdentifier() == 'SIDL.'+identifier:
          return child
    return None

  def getBaseException(self, exception):
    if exception.getFullIdentifier() == 'SIDL.BaseException':
      return exception
    for sclass in exception.getParents():
      base = self.getBaseException(sclass)
      if base:
        return base
    return None

  def getIORExceptions(self, method):
    exceptions = []
    for exception in method.getExceptions():
      if not exception.getFullIdentifier() == 'SIDL.NullIORException':
        exceptions.append(exception)
    return exceptions

  def generateBaseException(self, exception):
    if not self.generatedBaseException:
      self.baseException = self.getBaseException(exception)
      if self.baseException:
        self.typeOnly.append(1)
        self.baseException.accept(self.__IORself)
        self.typeOnly.pop()
        self.generatedBaseException = 1
    return

  #################################
  # SECTION: Marshalling Operations
  def marshallReturnType(self, node, type):
    '''No marshalling is done inside the IOR'''
    return

  ##########################
  # SECTION: Type operations
  def generateArrayTypes(self, type):
    iorType = self.getCode('ior array type')
    iorType.write(self.getIORArrayName(type))
    type.addCode(iorType)

    iorType = self.getCode('ior array struct')
    iorType.write(self.getIORArrayName(type)[:-2])
    type.addCode(iorType)
    return

  def generateTypeInformation(self, type):
    name = self.getIORName(type)

    try:
      iorType  = self.getCode('ior type')
      defValue = self.getCode('ior default value')
      type.addCode(iorType)
      type.addCode(defValue)
      if type.getBaseType():
        try:
          iorType.write(self.typeMap[type.getIdentifier()])
          defValue.write(self.defaultValues[type.getIdentifier()])
        except KeyError:
          self.raiseVisitorException('No base type: '+str(type.getIdentifier()))
      elif type.isInstanceOf('SIDLLanguage.Array'):
        iorType.write(self.getIORArrayName(SIDLLanguage.Array.Array(type).getType()))
        defValue.write('NULL')
      elif type.isInstanceOf('SIDLLanguage.Enumeration'):
        iorType.write('enum '+name+'__enum')
        defValue.write(SIDLLanguage.Enumerator.Enumerator(type.getChildren()[0]).getIdentifier())
      else:
        iorType.write('struct '+name+'__object *')
        defValue.write('NULL')

        code = self.getCode('ior type struct')
        code.write('struct '+name+'__object')
        type.addCode(code)

        code     = self.getCode('ior function table type')
        defValue = self.getCode('ior function table default value')
        code.write('struct '+name+'__external *')
        defValue.write('NULL')
        type.addCode(code)
        type.addCode(defValue)

        code = self.getCode('ior function table type struct')
        code.write('struct '+name+'__external')
        type.addCode(code)

        code     = self.getCode('ior epv table type')
        defValue = self.getCode('ior epv table default value')
        code.write('struct '+name+'__epv *')
        defValue.write('NULL')
        type.addCode(code)
        type.addCode(defValue)

        code = self.getCode('ior epv table type struct')
        code.write('struct '+name+'__epv')
        type.addCode(code)

        if type.isInstanceOf('SIDLLanguage.Class'):
          code     = self.getCode('ior static function table type')
          defValue = self.getCode('ior static function table default value')
          code.write('struct '+name+'__sepv *')
          defValue.write('NULL')
          type.addCode(code)
          type.addCode(defValue)

          code = self.getCode('ior static function table type struct')
          code.write('struct '+name+'__sepv')
          type.addCode(code)
      self.generateArrayTypes(type)

      self.marshallReturnType(type, type)
    except SIDLLanguage.CodeGeneratorException.Exception, e:
      self.raiseVisitorException('In '+type.getIdentifier()+': '+e.getMessage(), e.getStackTrace().strip())
    return

  ###############################
  # SECTION: Interface operations
  def generateInterfaceAccessors(self, interface):
    name = self.getIORName(interface)

    iorAccessor = self.getCode('ior function table accessor')
    iorAccessor.write(name+'__externals')
    interface.addCode(iorAccessor)

    iorAccessor = self.getCode('ior epv table accessor')
    iorAccessor.write(name+'__set_epv')
    interface.addCode(iorAccessor)

    if interface.isInstanceOf('SIDLLanguage.Class'):
      iorAccessor = self.getCode('ior static function table accessor')
      iorAccessor.write(name+'__statics')
      interface.addCode(iorAccessor)

      iorAccessor = self.getCode('ior static epv table accessor')
      iorAccessor.write(name+'__set_sepv')
      interface.addCode(iorAccessor)
    return

  ##########################
  # SECTION: Builtin Methods
  def generateMethodCast(self, method):
    parameters = method.getChildren()
    exceptions = self.getIORExceptions(method)

    iorCast = self.getCode('ior cast begin')
    iorCast.write('(')
    self.composeCode(iorCast, method.getReturnType(), 'ior type')
    iorCast.write(' (*)(')
    method.addCode(iorCast)

    iorCast = self.getCode('ior cast end')
    for child in parameters:
      parameter = SIDLLanguage.Parameter.Parameter(child)
      iorCast.write(', ')
      if self.isOut(parameter.getAccessMode()): 
        self.composeCode(iorCast, parameter.getType(), 'ior type')
        iorCast.write('* ')
      elif parameter.getType().getBaseType() and parameter.getType().getIdentifier() == 'string':
        iorCast.write('const ')
        self.composeCode(iorCast, parameter.getType(), 'ior type')
      else:
        self.composeCode(iorCast, parameter.getType(), 'ior type')
    if len(exceptions) > 0:
      iorCast.write(', ')
      self.composeCode(iorCast, self.getBaseException(exceptions[0]), 'ior type')
      iorCast.write('*')
    iorCast.write('))')
    method.addCode(iorCast)
    return

  ###########################
  # Header generation methods
  def generateHeader(self, interface, purpose):
    if not self.declActive: return
    self.generateHeaderBegin(interface, purpose)
    self.generateDefines(interface, purpose)
    self.generateIncludes(interface, purpose)
    self.generateExternBegin(interface, purpose)
    self.generateForwardDeclarations(interface, purpose)
    if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
      self.generateExports(interface, purpose)
      self.generateSEPVDeclaration(interface, purpose)
      self.generateEPVDeclaration(interface, purpose)
      self.generateObjectDeclaration(interface, purpose)
    else:
      self.generateEnumDeclaration(interface, purpose)
    self.generateExternEnd(interface, purpose)
    self.generateHeaderEnd(interface, purpose)
    return

  def generateHeaderBegin(self, interface, purpose):
    code  = self.getCode(purpose.getName()+' begin header')
    guard = 'included_'+self.getIORName(interface)+'_'+purpose.getName()
    code.write('#ifndef '+guard+'\n')
    code.write('#define '+guard+'\n')
    code.println('')
    interface.addCode(code)
    return

  def generateDefines(self, interface, purpose):
    code = self.getCode(purpose.getName()+' defines')
    code.write('#ifndef NULL\n')
    code.write('#define NULL 0\n')
    code.write('#endif\n')
    code.println('')
    interface.addCode(code)
    return

  def generateIncludes(self, interface, purpose):
    code = self.getCode(purpose.getName()+' includes')
    code.write('#include <stdlib.h>\n')
    code.write('#include <stdio.h>\n')
    code.write('#include <string.h>\n')
    code.println('')
    interface.addCode(code)
    return

  def generateExternBegin(self, interface, purpose):
    code = self.getCode(purpose.getName()+' begin extern')
    code.write('#ifdef __cplusplus\n')
    code.println('extern "C" {')
    code.write('#endif\n')
    code.println('')
    code.println('/* Symbol '+interface.getFullIdentifier()+' */')
    code.println('')
    interface.addCode(code)
    return

  def generateExports(self, interface, purpose):
    code = self.getCode(purpose.getName()+' exports')
    code.println('/* These structures are defined in this header */')
    code.indent()
    self.composeCode(code, interface, 'ior type struct')
    code.write(';\n')
    if self.hasStaticMethod(interface):
      code.indent()
      self.composeCode(code, interface, 'ior static function table type struct')
      code.write(';\n')
    code.println('')
    if interface.isInstanceOf('SIDLLanguage.Class'):
      sclass = SIDLLanguage.Class.Class(interface)
      code.println('/* These functions are defined in the implementation */')
      if not sclass.getAbstract():
        code.indent()
        code.write('extern ')
        self.composeCode(code, interface, 'ior type')
        code.write(self.getIORName(sclass)+'__new(void);\n')
      if self.hasStaticMethod(sclass):
        code.indent()
        code.write('extern ')
        self.composeCode(code, interface, 'ior static function table type')
        code.write(' '+self.getIORName(sclass)+'__statics(void);\n')
      code.indent()
      code.write('extern void '+self.getIORName(sclass)+'__init(')
      self.composeCode(code, interface, 'ior type')
      code.write('self);\n')
      code.indent()
      code.write('extern void '+self.getIORName(sclass)+'__fini(')
      self.composeCode(code, interface, 'ior type')
      code.write('self);\n')
      code.println('')
    interface.addCode(code)
    return

  def generateForwardDeclarations(self, interface, purpose):
    code = self.getCode(purpose.getName()+' forward declarations')
    code.println('/* Forward declarations for array structure and external classes and interfaces */')
    code.indent()
    self.composeCode(code, interface, 'ior array struct')
    code.write(';\n')
    if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
      for dep in interface.getDependencies():
        if dep.isInstanceOf('SIDLLanguage.Interface'):
          self.typeOnly.append(1)
          dep.accept(self.__IORself)
          self.typeOnly.pop()

          code.indent()
          self.composeCode(code, dep, 'ior type struct')
          code.write(';\n')
          code.indent()
          self.composeCode(code, dep, 'ior array struct')
          code.write(';\n')
      if self.hasMethodException(interface):
        code.indent()
        self.composeCode(code, self.baseException, 'ior type struct')
        code.write(';\n')
        code.indent()
        self.composeCode(code, self.baseException, 'ior array struct')
        code.write(';\n')
    code.println('')
    interface.addCode(code)
    return

  def generateEPVEntry(self, code, holder, interface, method, static = 0):
    if static == (not method.getStatic()): return
    args = []
    if not method.getStatic():
      if holder.isInstanceOf('SIDLLanguage.Class'):
        args.append(holder.getCodeBlock(self.language, 'ior type').getSource()+'self')
      else:
        args.append('void *self')
    for parameter in method.getChildren():
      args.append(parameter.getCodeBlock(self.language, 'ior parameter').getSource())
    if len(self.getIORExceptions(method)) > 0:
      args.append(self.baseException.getCodeBlock(self.language, 'ior type').getSource()+'*_exception')

    code.indent()
    # TODO: Kludge until IOR is changed
    if method.getIdentifier() == '_ctor':
      code.write('void')
    else:
      self.composeCode(code, method.getReturnType(), 'ior type')
    code.write(' (*f_'+method.getIdentifier()+')')
    if len(args) > 0:
      code.write('(\n')
      code.increaseTabLevel()
      for arg in args[:-1]:
        code.println(arg+',')
      code.println(args[-1]+');')
      code.decreaseTabLevel()
    else:
      code.write('(void);\n')
    return

  # TODO: Must make these Method nodes
  def generateEPVBuiltins(self, code, interface):
    code.println('/* Builtin methods */')
    code.println('void *(*f__cast_ior)(')
    code.increaseTabLevel()
    code.indent()
    if interface.isInstanceOf('SIDLLanguage.Class'):
      self.composeCode(code, interface, 'ior type')
    else:
      code.write('void *')
    code.write('self,\n')
    code.println('const char *name);')
    code.decreaseTabLevel()

    code.println('void (*f__delete)(')
    code.increaseTabLevel()
    code.indent()
    if interface.isInstanceOf('SIDLLanguage.Class'):
      self.composeCode(code, interface, 'ior type')
    else:
      code.write('void *')
    code.write('self);\n')
    code.decreaseTabLevel()

    if interface.isInstanceOf('SIDLLanguage.Class'):
      if self.hasCtorNode(SIDLLanguage.Class.Class(interface)): return

      code.println('void (*f__ctor)(')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('self);\n')
      code.decreaseTabLevel()

      code.println('void (*f__dtor)(')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('self);\n')
      code.decreaseTabLevel()
    return

  def generateClassEPVEntries(self, code, holder, sclass, markName, methodsSeen = None, static = 0):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)

    code.println('/* Methods contributed by '+sclass.getFullIdentifier()+' */')
    if methodsSeen is None: methodsSeen = []
    for child in sclass.getChildren():
      method = SIDLLanguage.Method.Method(child)
      same   = 0
      for m in methodsSeen:
        if method.getIdentifier() == m.getIdentifier():
          same = 1
          break
      if not same:
        self.generateEPVEntry(code, holder, sclass, method, static)
        methodsSeen.append(method)

    for cls in sclass.getParents():
      self.generateClassEPVEntries(code, holder, SIDLLanguage.Class.Class(cls), markName, methodsSeen, static)
    for interface in sclass.getInterfaces():
      self.generateInterfaceEPVEntries(code, holder, interface, markName, methodsSeen, static)
    return

  def generateInterfaceEPVEntries(self, code, holder, interface, markName, methodsSeen = None, static = 0):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)

    code.println('/* Methods contributed by '+interface.getFullIdentifier()+' */')
    if methodsSeen is None: methodsSeen = []
    for child in interface.getChildren():
      method = SIDLLanguage.Method.Method(child)
      same   = 0
      for m in methodsSeen:
        if method.getIdentifier() == m.getIdentifier():
          same = 1
          break
      if not same:
        self.generateEPVEntry(code, holder, interface, method, static)
        methodsSeen.append(method)

    for inf in interface.getParents():
      self.generateInterfaceEPVEntries(code, holder, inf, markName, methodsSeen, static)
    return

  def generateEPVDeclaration(self, interface, purpose):
    code = self.getCode(purpose.getName()+' epv declaration')
    code.println('/* Declare the method entry point vector */')
    code.indent()
    self.composeCode(code, interface, 'ior epv table type struct')
    code.write(' {\n')
    code.increaseTabLevel()
    self.generateEPVBuiltins(code, interface)
    markName = interface.getFullIdentifier()+'.epvMark'
    if interface.isInstanceOf('SIDLLanguage.Class'):
      self.generateClassEPVEntries(code, interface, interface, markName)
    else:
      self.generateInterfaceEPVEntries(code, interface, interface, markName)
    self.unmarkNetwork(interface, markName)
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    interface.addCode(code)
    return

  def generateSEPVDeclaration(self, sclass, purpose):
    if not self.hasStaticMethod(sclass): return
    code = self.getCode(purpose.getName()+' sepv declaration')
    code.println('/* Declare the static entry point vector */')
    code.indent()
    self.composeCode(code, sclass, 'ior static function table type struct')
    code.write(' {\n')
    code.increaseTabLevel()
    markName = sclass.getFullIdentifier()+'.epvMark'
    self.generateClassEPVEntries(code, sclass, sclass, markName, static = 1)
    self.unmarkNetwork(sclass, markName)
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    sclass.addCode(code)
    return

  def generateClassObjectEntries(self, code, sclass, level, markName):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    if level == 1:
      code.indent()
      self.composeCode(code, sclass, 'ior type struct')
      code.write(' d_'+self.getIORName(sclass).lower()+';\n')
    for cls in sclass.getParents():
      self.generateClassObjectEntries(code, SIDLLanguage.Class.Class(cls), level+1, markName)
    for interface in sclass.getInterfaces():
      self.generateInterfaceObjectEntries(code, interface, level, markName)
    return

  def generateInterfaceObjectEntries(self, code, interface, level, markName):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    if level == 0:
      code.indent()
      self.composeCode(code, interface, 'ior type struct')
      code.write(' d_'+self.getIORName(interface).lower()+';\n')
    for inf in interface.getParents():
      self.generateInterfaceObjectEntries(code, inf, level, markName)
    return

  def generateObjectDeclaration(self, interface, purpose):
    code = self.getCode(purpose.getName()+' object declaration')
    code.println('/* Define the object structure */')
    code.indent()
    self.composeCode(code, interface, 'ior type struct')
    code.write(' {\n')
    code.increaseTabLevel()
    # Declare subclass structures
    markName = interface.getFullIdentifier()+'.objMark'
    if interface.isInstanceOf('SIDLLanguage.Class'):
      self.generateClassObjectEntries(code, interface, 0, markName)
    self.unmarkNetwork(interface, markName)
    # Declare EPV, object, and backend
    code.indent()
    self.composeCode(code, interface, 'ior epv table type')
    code.write(' d_epv;\n')
    code.println('void *d_object;')
    if interface.isInstanceOf('SIDLLanguage.Class'):
      code.println('void *d_data;')
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    interface.addCode(code)
    return

  def generateEnumDeclaration(self, enum, purpose):
    code = self.getCode(purpose.getName()+' enumeration declaration')
    code.indent()
    self.composeCode(code, enum, 'ior type')
    code.write(' {\n')
    code.increaseTabLevel()
    children = enum.getChildren()
    for i in range(len(children)):
      enumerator = SIDLLanguage.Enumerator.Enumerator(children[i])
      code.indent()
      code.write(enumerator.getIdentifier()+' = '+str(enumerator.getValue()))
      if i < len(children)-1: code.write(',')
      code.write('\n')
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    enum.addCode(code)
    return

  def generateExternEnd(self, interface, purpose):
    code = self.getCode(purpose.getName()+' end extern')
    code.write('#ifdef __cplusplus\n')
    code.println('}')
    code.write('#endif\n')
    code.println('')
    interface.addCode(code)
    return

  def generateHeaderEnd(self, interface, purpose):
    code  = self.getCode(purpose.getName()+' end header')
    guard = 'included_'+self.getIORName(interface)+'_'+purpose.getName()
    code.write('#endif /* '+guard+' */\n')
    interface.addCode(code)
    return

  ##################################
  # Source generation helper methods
  def generateInterfaceVTables(self, code, interface, markName, inherited = 0):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    name = self.getIORName(interface).lower()

    code.indent()
    code.write('static ')
    self.composeCode(code, interface, 'ior epv table type struct')
    if inherited > 1: code.write(' ')
    code.write(' s_new__'+name+';\n')
    if inherited > 0:
      code.indent()
      code.write('static ')
      self.composeCode(code, interface, 'ior epv table type')
      code.write(' s_old__'+name+';\n')
    else:
      if self.hasStaticMethod(interface):
        code.indent()
        code.write('static ')
        self.composeCode(code, interface, 'ior static function table type struct')
        code.write(' s_stc__'+name+';\n')
        code.println('')
    code.println('')

    if interface.isInstanceOf('SIDLLanguage.Class'):
      for child in interface.getParents():
        self.generateInterfaceVTables(code, child, markName, inherited+1)
      for inf in SIDLLanguage.Class.Class(interface).getInterfaces():
        self.generateInterfaceVTables(code, inf, markName, inherited)
    else:
      for child in interface.getParents():
        self.generateInterfaceVTables(code, child, markName, inherited)
    return

  def generateVTables(self, code, sclass, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.vtableMark'
    self.generateInterfaceVTables(code, sclass, markName)
    self.unmarkNetwork(sclass, markName)
    return

  def generateClassSelf(self, code, sclass, level = 0, markName = '', selfName = 'self'):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)

    code.indent()
    self.composeCode(code, sclass, 'ior type')
    code.write('s'+str(level)+' = ')
    if level == 0:
      code.write(selfName+';\n')
    else:
      code.write('&s'+str(level-1)+'->d_'+self.getIORName(sclass).lower()+';\n')
    for cls in sclass.getParents():
      self.generateClassSelf(code, cls, level+1, markName)
    return

  def generateSelf(self, code, sclass, markName = '', selfName = 'self'):
    if not sclass.isInstanceOf('SIDLLanguage.Class'): return
    if not markName: markName = sclass.getFullIdentifier()+'.vtableMark'
    self.generateClassSelf(code, sclass, markName = markName, selfName = selfName)
    self.unmarkNetwork(sclass, markName)
    return

  def generateClassCastCheck(self, code, sclass, level, markName):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    name     = sclass.getFullIdentifier()
    selfName = 's'+str(level)

    code.indent()
    if level > 0: code.write('} else ')
    code.write('if (!strcmp(name, "'+name+'")) {\n')
    code.increaseTabLevel()
    code.println('cast = (void *) '+selfName+';')
    code.decreaseTabLevel()

    for cls in sclass.getParents():
      self.generateClassCastCheck(code, SIDLLanguage.Class.Class(cls), level+1, markName)
    for interface in sclass.getInterfaces():
      self.generateInterfaceCastCheck(code, interface, level, markName)
    return

  def generateInterfaceCastCheck(self, code, interface, level, markName):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    name     = self.getIORName(interface).lower()
    selfName = 's'+str(level)

    code.println('} else if (!strcmp(name, "'+interface.getFullIdentifier()+'")) {')
    code.increaseTabLevel()
    code.println('cast = (void *) &'+selfName+'->d_'+name+';')
    code.decreaseTabLevel()

    for inf in interface.getParents():
      self.generateInterfaceCastCheck(code, inf, level, markName)
    return

  def generateCastCheck(self, code, sclass, level = 0, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.castMark'
    self.generateClassCastCheck(code, sclass, level, markName)
    code.println('}')
    self.unmarkNetwork(sclass, markName)
    return

  def retargetClassEPV(self, code, sclass, level, markName, new):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    name     = self.getIORName(sclass).lower()
    selfName = 's'+str(level)

    for cls in sclass.getParents():
      self.retargetClassEPV(code, SIDLLanguage.Class.Class(cls), level+1, markName, new)
    for interface in sclass.getInterfaces():
      self.retargetInterfaceEPV(code, interface, level, markName, new)

    if new:
      code.println(selfName+'->d_epv = &s_new__'+name+';')
    else:
      code.println(selfName+'->d_epv = s_old__'+name+';')
    code.println('')
    return

  def retargetInterfaceEPV(self, code, interface, level, markName, new):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    name     = self.getIORName(interface).lower()
    selfName = 's'+str(level)

    for inf in interface.getParents():
      self.retargetInterfaceEPV(code, inf, level, markName, new)

    if new:
      code.println(selfName+'->d_'+name+'.d_epv = &s_new__'+name+';')
    else:
      code.println(selfName+'->d_'+name+'.d_epv = s_old__'+name+';')
    return

  def retargetEPV(self, code, sclass, level = 0, markName = '', new = 1):
    if not markName: markName = sclass.getFullIdentifier()+'.epvMark'
    code.println('/* Setup d_epv members */')
    self.retargetClassEPV(code, sclass, level, markName, new)
    self.unmarkNetwork(sclass, markName)
    return

  def setupClassData(self, code, sclass, level, markName):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    selfName = 's'+str(level)

    for cls in sclass.getParents():
      self.setupClassData(code, SIDLLanguage.Class.Class(cls), level+1, markName)

    if level > 0 and sclass.getAbstract():
      code.println(selfName+'->d_data = self->d_data;')
    return

  def setupData(self, code, sclass, level = 0, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.abstractMark'
    code.println('/* Setup d_data members */')
    code.println('s0->d_data = NULL;')
    if not sclass.getAbstract():
      code.println('(*(self->d_epv->f__ctor))(self);')
    self.setupClassData(code, sclass, level, markName)
    self.unmarkNetwork(sclass, markName)
    return

  def setupClassObject(self, code, sclass, level, markName):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    selfName = 's'+str(level)

    for cls in sclass.getParents():
      self.setupClassObject(code, SIDLLanguage.Class.Class(cls), level+1, markName)
    for interface in sclass.getInterfaces():
      self.setupInterfaceObject(code, interface, level, markName)

    code.println(selfName+'->d_object = self;')
    return

  def setupInterfaceObject(self, code, interface, level, markName):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    name     = self.getIORName(interface).lower()
    selfName = 's'+str(level)

    for inf in interface.getParents():
      self.setupInterfaceObject(code, inf, level, markName)

    code.println(selfName+'->d_'+name+'.d_object = self;')
    return

  def setupObjects(self, code, sclass, level = 0, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.objectMark'
    code.println('/* Setup d_object members */')
    self.setupClassObject(code, sclass, level, markName)
    self.unmarkNetwork(sclass, markName)
    return

  def aliasInterfaceEPV(self, code, interface, count, markName):
    if self.isMarked(interface, name = markName): return count
    self.markNode(interface, name = markName)
    name  = self.getIORName(interface).lower()
    if count < 0:
      alias = 'epv'
    else:
      alias = 'e'+str(count)

    code.indent()
    self.composeCode(code, interface, 'ior epv table type')
    code.write(' '+alias+' = &s_new__'+name+';\n')

    for parent in interface.getParents():
      count = self.aliasInterfaceEPV(code, parent, count+1, markName)
    if interface.isInstanceOf('SIDLLanguage.Class'):
      for inf in SIDLLanguage.Class.Class(interface).getInterfaces():
        count = self.aliasInterfaceEPV(code, inf, count+1, markName)
    return count

  def aliasEPV(self, code, sclass, count = -1, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.epvMark'
    self.aliasInterfaceEPV(code, sclass, count, markName)
    code.println('')
    self.unmarkNetwork(sclass, markName)
    return

  def storeClassEPV(self, code, sclass, level, markName):
    if self.isMarked(sclass, name = markName): return
    self.markNode(sclass, name = markName)
    name     = self.getIORName(sclass).lower()
    selfName = 's'+str(level)

    for cls in sclass.getParents():
      self.storeClassEPV(code, SIDLLanguage.Class.Class(cls), level+1, markName)
    for interface in sclass.getInterfaces():
      self.storeInterfaceEPV(code, interface, level, markName)

    code.println('s_old__'+name+' = '+selfName+'->d_epv;')
    return

  def storeInterfaceEPV(self, code, interface, level, markName):
    if self.isMarked(interface, name = markName): return
    self.markNode(interface, name = markName)
    name     = self.getIORName(interface).lower()
    selfName = 's'+str(level)

    for inf in interface.getParents():
      self.storeInterfaceEPV(code, inf, level, markName)

    code.println('s_old__'+name+' = '+selfName+'->d_'+name+'.d_epv;')
    return

  def storeEPV(self, code, sclass, level = 0, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.epvMark'
    self.storeClassEPV(code, sclass, level, markName)
    code.println('')
    self.unmarkNetwork(sclass, markName)
    return

  def updateClassEPV(self, code, sclass, count, markName):
    if self.isMarked(sclass, name = markName): return count
    self.markNode(sclass, name = markName)
    name  = self.getIORName(sclass).lower()
    alias = 'e'+str(count)

    if count >= 0:
      code.indent()
      code.write(alias+'->f__cast_ior = (void *(*)(')
      self.composeCode(code, sclass, 'ior type')
      code.write(', const char *)) epv->f__cast_ior;\n')
      code.indent()
      code.write(alias+'->f__delete = (void (*)(')
      self.composeCode(code, sclass, 'ior type')
      code.write(')) epv->f__delete;\n')
      for method in sclass.getMethods():
        if method.getStatic(): continue
        code.indent()
        code.write(alias+'->f_'+method.getIdentifier()+' = ')
        self.composeCode(code, method, 'ior cast begin')
        self.composeCode(code, sclass, 'ior type')
        self.composeCode(code, method, 'ior cast end')
        code.write(' epv->f_'+method.getIdentifier()+';\n')
      code.println('')

    for cls in sclass.getParents():
      count = self.updateClassEPV(code, SIDLLanguage.Class.Class(cls), count+1, markName)
    for interface in sclass.getInterfaces():
      count = self.updateInterfaceEPV(code, interface, count+1, markName)
    return count

  def updateInterfaceEPV(self, code, interface, count, markName):
    if self.isMarked(interface, name = markName): return count
    self.markNode(interface, name = markName)
    name  = self.getIORName(interface).lower()
    alias = 'e'+str(count)

    code.println(alias+'->f__cast_ior = (void *(*)(void *, const char *)) epv->f__cast_ior;')
    code.println(alias+'->f__delete = (void (*)(void *)) epv->f__delete;')
    for method in interface.getMethods():
      if method.getStatic(): continue
      code.indent()
      code.write(alias+'->f_'+method.getIdentifier()+' = ')
      self.composeCode(code, method, 'ior cast begin')
      code.write('void *')
      self.composeCode(code, method, 'ior cast end')
      code.write(' epv->f_'+method.getIdentifier()+';\n')
    code.println('')

    for inf in interface.getParents():
      count = self.updateInterfaceEPV(code, inf, count+1, markName)
    return count

  def updateEPV(self, code, sclass, count = -1, markName = ''):
    if not markName: markName = sclass.getFullIdentifier()+'.epvMark'
    self.updateClassEPV(code, sclass, count, markName)
    self.unmarkNetwork(sclass, markName)
    return

  ###########################
  # Source generation methods
  def generateSource(self, interface):
    if self.defnActive:
      if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
        self.generateStaticVariables(interface)
      if interface.isInstanceOf('SIDLLanguage.Class'):
        self.generateSkeletonDeclarations(interface)
        self.generateCast(interface)
        self.generateDelete(interface)
        self.generateInitEPV(interface)
        self.generateInitSEPV(interface)
        self.generateStatics(interface)
        self.generateNew(interface)
        self.generateInit(interface)
        self.generateFini(interface)
    self.generateArray(interface)
    if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
      self.generateExternalEntryPoints(interface)
    return

  def generateStaticVariables(self, interface):
    staticVars = self.getCode('ior static variables')
    staticVars.println('/* Static variables for managing EPV initialization */')
    if interface.isInstanceOf('SIDLLanguage.Class'):
      staticVars.println('static int s_method_initialized = 0;')
    if self.hasStaticMethod(interface):
      staticVars.println('static int s_static_initialized = 0;')
    staticVars.println('')
    if interface.isInstanceOf('SIDLLanguage.Class'):
      self.generateVTables(staticVars, interface)
    interface.addCode(staticVars)
    return

  def generateSkeletonDeclarations(self, sclass):
    skelDecl = self.getCode('ior skeleton declarations')
    skelDecl.println('/* Declare EPV routines defined in the skeleton */')
    skelDecl.indent()
    skelDecl.write('extern void ')
    self.composeCode(skelDecl, sclass, 'ior epv table accessor')
    skelDecl.write('(')
    self.composeCode(skelDecl, sclass, 'ior epv table type')
    skelDecl.write(' epv);\n')
    if self.hasStaticMethod(sclass):
      skelDecl.indent()
      skelDecl.write('extern void ')
      self.composeCode(skelDecl, sclass, 'ior static epv table accessor')
      skelDecl.write('(')
      self.composeCode(skelDecl, sclass, 'ior static function table type')
      skelDecl.write(' sepv);\n')
    skelDecl.println('')
    sclass.addCode(skelDecl)
    return

  def generateCast(self, sclass):
    castDefn = self.getCode('ior cast definition')
    castDefn.println('/* CAST: dynamic type casting support */')
    castDefn.println('static void *'+self.getIORName(sclass)+'__cast_ior(')
    castDefn.increaseTabLevel()
    castDefn.indent()
    self.composeCode(castDefn, sclass, 'ior type')
    castDefn.write('self,\n')
    castDefn.println('const char *name)')
    castDefn.decreaseTabLevel()
    castDefn.println('{')
    castDefn.increaseTabLevel()
    castDefn.println('void *cast = NULL;')
    castDefn.println('/* We are guaranteed here that the first member in any IOR is a BaseInterface object */')
    self.generateSelf(castDefn, sclass, selfName = '(('+self.getBaseObject(sclass, 'BaseInterface').getCodeBlock(self.language, 'ior type').getSource()+') self)->d_object')
    castDefn.println('')
    self.generateCastCheck(castDefn, sclass)
    castDefn.println('')
    castDefn.println('return cast;')
    castDefn.decreaseTabLevel()
    castDefn.println('}')
    castDefn.println('')
    sclass.addCode(castDefn)
    return

  def generateDelete(self, sclass):
    deleteDefn = self.getCode('ior delete definition')
    deleteDefn.println('/* DELETE: call destructor and free object memory */')
    deleteDefn.println('static void '+self.getIORName(sclass)+'__delete(')
    deleteDefn.increaseTabLevel()
    deleteDefn.indent()
    self.composeCode(deleteDefn, sclass, 'ior type')
    deleteDefn.write('self)\n')
    deleteDefn.decreaseTabLevel()
    deleteDefn.println('{')
    deleteDefn.increaseTabLevel()
    deleteDefn.indent()
    self.composeCode(deleteDefn, sclass, 'ior type')
    deleteDefn.write(' s0 = self->d_epv->f__cast_ior(self, "'+sclass.getFullIdentifier()+'");\n')
    deleteDefn.println(self.getIORName(sclass)+'__fini(s0);')
    deleteDefn.indent()
    deleteDefn.write('memset((void *) s0, 0, sizeof(')
    self.composeCode(deleteDefn, sclass, 'ior type struct')
    deleteDefn.write('));\n')
    deleteDefn.println('free((void *) s0);')
    deleteDefn.decreaseTabLevel()
    deleteDefn.println('}')
    deleteDefn.println('')
    sclass.addCode(deleteDefn)
    return

  def generateInitEPV(self, sclass):
    initEPVDefn = self.getCode('ior epv initialization definition')
    initEPVDefn.println('/* INIT_EPV: create method entry point vector (EPV) structure */')
    initEPVDefn.println('static void '+self.getIORName(sclass)+'__init_epv(')
    initEPVDefn.increaseTabLevel()
    initEPVDefn.indent()
    self.composeCode(initEPVDefn, sclass, 'ior type')
    initEPVDefn.write('self)\n')
    initEPVDefn.decreaseTabLevel()
    initEPVDefn.println('{')
    initEPVDefn.increaseTabLevel()
    self.generateSelf(initEPVDefn, sclass)
    initEPVDefn.println('')
    self.aliasEPV(initEPVDefn, sclass)
    for cls in sclass.getParents():
      self.storeEPV(initEPVDefn, SIDLLanguage.Class.Class(cls), level = 1)
    initEPVDefn.println('epv->f__cast_ior = '+self.getIORName(sclass)+'__cast_ior;')
    initEPVDefn.println('epv->f__delete = '+self.getIORName(sclass)+'__delete;')
    if not self.hasCtorNode(sclass):
      initEPVDefn.println('epv->f__ctor = NULL;')
      initEPVDefn.println('epv->f__dtor = NULL;')

    methods = []
    for cls in sclass.getParents():
      methods.extend(cls.getMethods())
    for method in sclass.getMethods():
      if method.getStatic(): continue
      same = 0
      for m in methods:
        if method.isSame(m): same = 1
      if same:
        initEPVDefn.indent()
        initEPVDefn.write('epv->f_'+method.getIdentifier()+' = ')
        self.composeCode(initEPVDefn, method, 'ior cast begin')
        self.composeCode(initEPVDefn, sclass, 'ior type')
        self.composeCode(initEPVDefn, method, 'ior cast end')
        initEPVDefn.write(' s1->d_epv->f_'+method.getIdentifier()+';\n')
      else:
        initEPVDefn.println('epv->f_'+method.getIdentifier()+' = NULL;')
    initEPVDefn.println('')

    initEPVDefn.indent()
    self.composeCode(initEPVDefn, sclass, 'ior epv table accessor')
    initEPVDefn.write('(epv);\n')
    initEPVDefn.println('')
    self.updateEPV(initEPVDefn, sclass)
    initEPVDefn.println('s_method_initialized = 1;')
    initEPVDefn.decreaseTabLevel()
    initEPVDefn.println('}')
    initEPVDefn.println('')
    sclass.addCode(initEPVDefn)
    return

  def generateInitSEPV(self, sclass):
    if not self.hasStaticMethod(sclass): return
    initSEPVDefn = self.getCode('ior static epv initialization definition')
    initSEPVDefn.println('/* INIT_SEPV: create static entry point vector (SEPV) structure */')
    initSEPVDefn.println('static void '+self.getIORName(sclass)+'__init_sepv(void)')
    initSEPVDefn.println('{')
    initSEPVDefn.increaseTabLevel()
    initSEPVDefn.indent()
    self.composeCode(initSEPVDefn, sclass, 'ior static function table type')
    initSEPVDefn.write(' s = &s_stc__'+self.getIORName(sclass).lower()+';\n')
    for cls in sclass.getParents():
      if self.hasStaticMethod(cls):
        initSEPVDefn.write('#if '+str(int(self.isBase(cls)))+'\n')
        initSEPVDefn.indent()
        initSEPVDefn.write('p = ')
        self.composeCode(initSEPVDefn, sclass, 'ior static function table accessor')
        initSEPVDefn.write('();\n')
        initSEPVDefn.write('#else\n')
        initSEPVDefn.indent()
        self.composeCode(initSEPVDefn, sclass, 'ior static function table type')
        initSEPVDefn.write(' (*dll_f)() = (')
        self.composeCode(initSEPVDefn, sclass, 'ior static function table type')
        initSEPVDefn.write(' (*)()) SIDL_Loader__externals()->getStaticEPV()->f_lookupSymbol("')
        self.composeCode(initSEPVDefn, sclass, 'ior static function table accessor')
        initSEPVDefn.write('");\n')
        initSEPVDefn.println('')
        initSEPVDefn.println('if (!dll_f) {')
        initSEPVDefn.increaseTabLevel()
        initSEPVDefn.println('fputs("Unable to find static function table for '+cls.getFullIdentifier()+'; add to SIDL_DLL_PATH\\n", stderr);')
        initSEPVDefn.println('exit(1);')
        initSEPVDefn.decreaseTabLevel()
        initSEPVDefn.println('} else {')
        initSEPVDefn.increaseTabLevel()
        initSEPVDefn.println('p = (*dll_f)();')
        initSEPVDefn.println('')
        initSEPVDefn.println('if (!p) {')
        initSEPVDefn.increaseTabLevel()
        initSEPVDefn.println('fputs("Invalid static function table for '+cls.getFullIdentifier()+'\\n", stderr);')
        initSEPVDefn.println('exit(1);')
        initSEPVDefn.decreaseTabLevel()
        initSEPVDefn.println('}')
        initSEPVDefn.decreaseTabLevel()
        initSEPVDefn.println('}')
        initSEPVDefn.write('#endif\n')
    initSEPVDefn.println('')
    methods = []
    for cls in sclass.getParents():
      methods.extend(cls.getMethods())
    for child in sclass.getChildren():
      method = SIDLLanguage.Method.Method(child)
      if method.getStatic():
        initSEPVDefn.indent()
        initSEPVDefn.write('s->f_'+method.getIdentifier()+' = ')
        same = 0
        for m in methods:
          if method.isSame(m): same = 1
        if same:
          initSEPVDefn.write('p->f_'+method.getIdentifier()+';\n')
        else:
          initSEPVDefn.write('NULL;\n')
    initSEPVDefn.println('')
    initSEPVDefn.indent()
    self.composeCode(initSEPVDefn, sclass, 'ior static epv table accessor')
    initSEPVDefn.write('(s);\n')
    initSEPVDefn.println('')
    initSEPVDefn.println('s_static_initialized = 1;')
    initSEPVDefn.decreaseTabLevel()
    initSEPVDefn.println('}')
    initSEPVDefn.println('')
    sclass.addCode(initSEPVDefn)
    return

  def generateStatics(self, sclass):
    if not self.hasStaticMethod(sclass): return
    staticDefn = self.getCode('ior static accessor definition')
    staticDefn.println('/* STATICS: return static EPV structure for static methods */')
    staticDefn.indent()
    self.composeCode(staticDefn, sclass, 'ior static function table type')
    staticDefn.write(' '+self.getIORName(sclass)+'__statics()\n')
    staticDefn.println('{')
    staticDefn.increaseTabLevel()
    staticDefn.println('if (!s_static_initialized) {')
    staticDefn.increaseTabLevel()
    staticDefn.println(self.getIORName(sclass)+'__init_sepv();')
    staticDefn.decreaseTabLevel()
    staticDefn.println('}')
    staticDefn.println('return &s_stc__'+self.getIORName(sclass).lower()+';')
    staticDefn.decreaseTabLevel()
    staticDefn.println('}')
    staticDefn.println('')
    sclass.addCode(staticDefn)
    return

  def generateNew(self, sclass):
    if sclass.getAbstract(): return
    newDefn = self.getCode('ior new definition')
    newDefn.println('/* NEW: allocate object and initialize it */')
    newDefn.indent()
    self.composeCode(newDefn, sclass, 'ior type')
    newDefn.write(self.getIORName(sclass)+'__new()\n')
    newDefn.println('{')
    newDefn.increaseTabLevel()
    newDefn.indent()
    self.composeCode(newDefn, sclass, 'ior type')
    newDefn.write('self = (')
    self.composeCode(newDefn, sclass, 'ior type')
    newDefn.write(') malloc(sizeof(')
    self.composeCode(newDefn, sclass, 'ior type struct')
    newDefn.write('));\n')
    newDefn.println(self.getIORName(sclass)+'__init(self);')
    newDefn.println('return self;')
    newDefn.decreaseTabLevel()
    newDefn.println('}')
    newDefn.println('')
    sclass.addCode(newDefn)
    return

  def generateInit(self, sclass):
    '''Generate the class initializer
    When is this called?
    The parent initializer must be dynamically loaded since the IOR resides in the implementation library of the parent'''
    initDefn = self.getCode('ior init definition')
    initDefn.println('/* INIT: initialize a new instance of the class object */')
    initDefn.println('void '+self.getIORName(sclass)+'__init(')
    initDefn.increaseTabLevel()
    initDefn.indent()
    self.composeCode(initDefn, sclass, 'ior type')
    initDefn.write('self)\n')
    initDefn.decreaseTabLevel()
    initDefn.println('{')
    initDefn.increaseTabLevel()
    self.generateSelf(initDefn, sclass)
    initDefn.println('')
    for cls in sclass.getParents():
      initDefn.write('#if '+str(int(self.isBase(cls)))+'\n')
      initDefn.println(self.getIORName(cls)+'__init(s1);')
      initDefn.write('#else\n')
      initDefn.indent()
      initDefn.write('void (*dll_f)(')
      self.composeCode(initDefn, cls, 'ior type')
      initDefn.write(') = (void (*)(')
      self.composeCode(initDefn, cls, 'ior type')
      initDefn.write(')) SIDL_Loader__externals()->getStaticEPV()->f_lookupSymbol("'+self.getIORName(cls)+'__init");\n')
      initDefn.println('')
      initDefn.println('if (!dll_f) {')
      initDefn.increaseTabLevel()
      initDefn.println('fputs("Unable to find initializer for '+cls.getFullIdentifier()+'; add to SIDL_DLL_PATH\\n", stderr);')
      initDefn.println('exit(1);')
      initDefn.decreaseTabLevel()
      initDefn.println('} else {')
      initDefn.increaseTabLevel()
      initDefn.println('(*dll_f)(s1);')
      initDefn.decreaseTabLevel()
      initDefn.println('}')
      initDefn.write('#endif\n')
      initDefn.println('')
    initDefn.println('if (!s_method_initialized) {')
    initDefn.increaseTabLevel()
    initDefn.println(self.getIORName(sclass)+'__init_epv(s0);')
    initDefn.decreaseTabLevel()
    initDefn.println('}')
    initDefn.println('')
    self.retargetEPV(initDefn, sclass)
    self.setupObjects(initDefn, sclass)
    initDefn.println('')
    self.setupData(initDefn, sclass)
    initDefn.decreaseTabLevel()
    initDefn.println('}')
    initDefn.println('')
    sclass.addCode(initDefn)
    return

  def generateFini(self, sclass):
    '''Generate the class finializer
    When is this called?
    The parent finalizer must be dynamically loaded since the IOR resides in the implementation library of the parent'''
    finiDefn = self.getCode('ior fini definition')
    finiDefn.println('/* FINI: deallocate a class instance (destructor) */')
    finiDefn.println('void '+self.getIORName(sclass)+'__fini(')
    finiDefn.increaseTabLevel()
    finiDefn.indent()
    self.composeCode(finiDefn, sclass, 'ior type')
    finiDefn.write('self)\n')
    finiDefn.decreaseTabLevel()
    finiDefn.println('{')
    finiDefn.increaseTabLevel()
    self.generateSelf(finiDefn, sclass)
    for cls in sclass.getParents():
      finiDefn.write('#if '+str(int(not self.isBase(cls)))+'\n')
      finiDefn.indent()
      finiDefn.write('void (*dll_f)(')
      self.composeCode(finiDefn, cls, 'ior type')
      finiDefn.write(') = (void (*)(')
      self.composeCode(finiDefn, cls, 'ior type')
      finiDefn.write(')) SIDL_Loader__externals()->getStaticEPV()->f_lookupSymbol("'+self.getIORName(cls)+'__fini");\n')
      finiDefn.write('#endif\n')
    finiDefn.println('')
    finiDefn.println('(*(self->d_epv->f__dtor))(self);')
    finiDefn.println('')
    for cls in sclass.getParents():
      self.retargetEPV(finiDefn, SIDLLanguage.Class.Class(cls), level = 1, new = 0)
      finiDefn.write('#if '+str(int(self.isBase(cls)))+'\n')
      finiDefn.println(self.getIORName(cls)+'__fini(s1);')
      finiDefn.write('#else\n')
      finiDefn.println('if (!dll_f) {')
      finiDefn.increaseTabLevel()
      finiDefn.println('fputs("Unable to find finializer for '+cls.getFullIdentifier()+'; add to SIDL_DLL_PATH\\n", stderr);')
      finiDefn.println('exit(1);')
      finiDefn.decreaseTabLevel()
      finiDefn.println('} else {')
      finiDefn.increaseTabLevel()
      finiDefn.println('(*dll_f)(s1);')
      finiDefn.decreaseTabLevel()
      finiDefn.println('}')
      finiDefn.write('#endif\n')
    finiDefn.decreaseTabLevel()
    finiDefn.println('}')
    finiDefn.println('')
    sclass.addCode(finiDefn)
    return

  def generateArray(self, interface):
    self.generateArrayStructure(interface)
    self.generateArrayAllocator(interface)
    self.generateArrayConstructor(interface)
    self.generateArrayBorrow(interface)
    self.generateArrayDestructor(interface)
    self.generateArrayDimension(interface)
    self.generateArrayLower(interface)
    self.generateArrayUpper(interface)
    self.generateArrayAccess(interface, 0)
    self.generateArrayAccess(interface, 1)
    self.generateFastArrayAccess(interface, 0)
    self.generateFastArrayAccess(interface, 1)
    return

  def generateArrayStructure(self, interface):
    if not self.defnActive: return
    code = self.getCode('ior array structure')
    code.println('/* The IOR array structure */')
    code.println('/*   Macros to read this are defined in SIDLArray.h */')
    code.indent()
    self.composeCode(code, interface, 'ior array struct')
    code.write(' {\n')
    code.increaseTabLevel()
    code.indent()
    self.composeCode(code, interface, 'ior type')
    code.write('* d_firstElement;\n')
    code.println(self.typeMap['int']+' *d_lower;')
    code.println(self.typeMap['int']+' *d_upper;')
    code.println(self.typeMap['int']+' *d_stride;')
    code.println(self.typeMap['bool']+' d_borrowed;')
    code.println(self.typeMap['int']+' d_dimen;')
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    interface.addCode(code)
    return

  def generateArrayAllocator(self, interface):
    if not self.defnActive: return
    code = self.getCode('ior array allocator')
    code.indent()
    code.write('static ')
    self.composeCode(code, interface, 'ior array type')
    code.write('newArray('+self.typeMap['int']+' dimen, const '+self.typeMap['int']+' lower[], const '+self.typeMap['int']+' upper[]) {\n')
    code.increaseTabLevel()
    code.indent()
    code.write('static const size_t arraySize = sizeof(')
    self.composeCode(code, interface, 'ior array struct')
    code.write(') + (sizeof('+self.typeMap['int']+') - (sizeof(')
    self.composeCode(code, interface, 'ior array struct')
    code.write(')%sizeof('+self.typeMap['int']+')))%sizeof('+self.typeMap['int']+');\n')
    code.indent()
    self.composeCode(code, interface, 'ior array type')
    code.write('result = (')
    self.composeCode(code, interface, 'ior array type')
    code.write(') malloc(arraySize + 3*sizeof('+self.typeMap['int']+')*dimen);\n')
    code.println('')
    code.println('result->d_dimen    = dimen;')
    code.println('result->d_borrowed = 0;')
    code.println('result->d_lower    = ('+self.typeMap['int']+' *) ((char *) result + arraySize);')
    code.println('result->d_upper    = result->d_lower + dimen;')
    code.println('result->d_stride   = result->d_upper + dimen;')
    code.println('memcpy(result->d_lower, lower, sizeof('+self.typeMap['int']+')*dimen);')
    code.println('memcpy(result->d_upper, upper, sizeof('+self.typeMap['int']+')*dimen);')
    code.println('return result;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateArrayConstructor(self, interface):
    code = self.getCode('ior array constructor comment')
    code.println('/* Create a dense array of the given dimension with specified index bounds */')
    code.println('/*   This array owns and manages its data */')
    code.println('/*   All object pointers are initialized to NULL */')
    interface.addCode(code)

    code = self.getCode('ior array constructor declaration begin')
    self.composeCode(code, interface, 'ior array type')
    interface.addCode(code)

    code = self.getCode('ior array constructor declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.println(self.typeMap['int']+' dimen,')
    code.println('const '+self.typeMap['int']+' lower[],')
    code.indent()
    code.write('const '+self.typeMap['int']+' upper[])')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array constructor declaration')
      self.composeCode(code, interface, 'ior array constructor comment')
      self.composeCode(code, interface, 'ior array constructor declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_create')
      self.composeCode(code, interface, 'ior array constructor declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array constructor')
      self.composeCode(code, interface, 'ior array constructor comment')
      self.composeCode(code, interface, 'ior array constructor declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_create')
      self.composeCode(code, interface, 'ior array constructor declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior array type')
      code.write('result = newArray(dimen, lower, upper);\n')
      code.println(self.typeMap['int']+' size;')
      code.println(self.typeMap['int']+' i;')
      code.println('')
      code.println('for(i = 0, size = 1; i < dimen; i++) {')
      code.increaseTabLevel()
      code.println('result->d_stride[i] = size;')
      code.println('size *= (upper[i] - lower[i] + 1);')
      code.decreaseTabLevel()
      code.println('}')
      code.indent()
      code.write('size *= sizeof(')
      self.composeCode(code, interface, 'ior type')
      code.write(');\n')
      code.indent()
      code.write('result->d_firstElement = (')
      self.composeCode(code, interface, 'ior type')
      code.write('*) malloc((size_t) size);\n')
      code.println('memset(result->d_firstElement, 0, (size_t) size);')
      code.println('return result;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayBorrow(self, interface):
    code = self.getCode('ior array borrow comment')
    code.println('/* Create an array using user-provided memory */')
    interface.addCode(code)

    code = self.getCode('ior array borrow declaration begin')
    self.composeCode(code, interface, 'ior array type')
    interface.addCode(code)

    code = self.getCode('ior array borrow declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    self.composeCode(code, interface, 'ior type')
    code.write('* firstElement,\n')
    code.println(self.typeMap['int']+' dimen,')
    code.println('const '+self.typeMap['int']+' lower[],')
    code.println('const '+self.typeMap['int']+' upper[],')
    code.indent()
    code.write('const '+self.typeMap['int']+' stride[])')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array borrow declaration')
      self.composeCode(code, interface, 'ior array borrow comment')
      self.composeCode(code, interface, 'ior array borrow declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_borrow')
      self.composeCode(code, interface, 'ior array borrow declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array borrow')
      self.composeCode(code, interface, 'ior array borrow comment')
      self.composeCode(code, interface, 'ior array borrow declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_borrow')
      self.composeCode(code, interface, 'ior array borrow declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior array type')
      code.write('result = newArray(dimen, lower, upper);\n')
      code.println('')
      code.println('memcpy(result->d_stride, stride, sizeof('+self.typeMap['int']+')*dimen);')
      code.println('result->d_firstElement = firstElement;')
      code.println('result->d_borrowed     = 1;')
      code.println('return result;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayDestructor(self, interface):
    code = self.getCode('ior array destructor comment')
    code.println('/* Destroy the array (NULL arrays are ignored) */')
    interface.addCode(code)

    code = self.getCode('ior array destructor declaration begin')
    code.write('void')
    interface.addCode(code)

    code = self.getCode('ior array destructor declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    self.composeCode(code, interface, 'ior array type')
    code.write('array)')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array destructor declaration')
      self.composeCode(code, interface, 'ior array destructor comment')
      self.composeCode(code, interface, 'ior array destructor declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_destroy')
      self.composeCode(code, interface, 'ior array destructor declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array destructor')
      self.composeCode(code, interface, 'ior array destructor comment')
      self.composeCode(code, interface, 'ior array destructor declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_destroy')
      self.composeCode(code, interface, 'ior array destructor declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.println('if (array) {')
      code.increaseTabLevel()
      code.println('const '+self.typeMap['int']+' dimen = array->d_dimen;')
      code.println('')
      code.println('if (!(array->d_borrowed)) {')
      code.increaseTabLevel()
      code.println('if (dimen > 0) {')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('*start = array->d_firstElement;\n')
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('*end;\n')
      code.println(self.typeMap['int']+' size = 1;')
      code.println('')
      code.println('if (dimen > 1) {')
      code.increaseTabLevel()
      code.println('size = array->d_stride[dimen-1];')
      code.decreaseTabLevel()
      code.println('}')
      code.println('size *= (array->d_upper[dimen-1] - array->d_lower[dimen-1] + 1);')
      code.println('end   = start + size;')
      if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
        code.println('while(start < end) {')
        code.increaseTabLevel()
        code.println('if (*start) {')
        code.increaseTabLevel()
        # This works since every class has a BaseClass object in front
        code.println('(*(((struct SIDL_BaseClass__object *) (*start)->d_object)->d_epv->f_deleteReference))((*start)->d_object);')
        code.println('*start = NULL;')
        code.decreaseTabLevel()
        code.println('}')
        code.println('++start;')
        code.decreaseTabLevel()
        code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      code.println('free(array->d_firstElement);')
      code.decreaseTabLevel()
      code.println('}')
      code.indent()
      code.write('memset(array, 0, sizeof(')
      self.composeCode(code, interface, 'ior array struct')
      code.write(') + 3*sizeof('+self.typeMap['int']+')*dimen);\n')
      code.println('free(array);')
      code.decreaseTabLevel()
      code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayDimension(self, interface):
    code = self.getCode('ior array dimension comment')
    code.println('/* Return the number of dimensions in the array */')
    code.println('/*   If the array pointer is NULL, zero is returned */')
    interface.addCode(code)

    code = self.getCode('ior array dimension declaration begin')
    code.write(self.typeMap['int'])
    interface.addCode(code)

    code = self.getCode('ior array dimension declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    code.write('const ')
    self.composeCode(code, interface, 'ior array type')
    code.write('array)')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array dimension declaration')
      self.composeCode(code, interface, 'ior array dimension comment')
      self.composeCode(code, interface, 'ior array dimension declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_dimen')
      self.composeCode(code, interface, 'ior array dimension declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array dimension')
      self.composeCode(code, interface, 'ior array dimension comment')
      self.composeCode(code, interface, 'ior array dimension declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_dimen')
      self.composeCode(code, interface, 'ior array dimension declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.println('return array ? array->d_dimen : 0;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayLower(self, interface):
    code = self.getCode('ior array lower comment')
    code.println('/* Return the lower bound on dimension ind */')
    code.println('/*   If the array pointer is NULL or ind is not a valid dimension, zero is returned */')
    interface.addCode(code)

    code = self.getCode('ior array lower declaration begin')
    code.write(self.typeMap['int'])
    interface.addCode(code)

    code = self.getCode('ior array lower declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    code.write('const ')
    self.composeCode(code, interface, 'ior array type')
    code.write('array,\n')
    code.indent()
    code.write(self.typeMap['int']+' ind)')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array lower declaration')
      self.composeCode(code, interface, 'ior array lower comment')
      self.composeCode(code, interface, 'ior array lower declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_lower')
      self.composeCode(code, interface, 'ior array lower declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array lower')
      self.composeCode(code, interface, 'ior array lower comment')
      self.composeCode(code, interface, 'ior array lower declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_lower')
      self.composeCode(code, interface, 'ior array lower declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.println('return (array && (ind < array->d_dimen) && (ind >= 0)) ? array->d_lower[ind] : 0;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayUpper(self, interface):
    code = self.getCode('ior array upper comment')
    code.println('/* Return the upper bound on dimension ind */')
    code.println('/*   If the array pointer is NULL or ind is not a valid dimension, -1 is returned */')
    interface.addCode(code)

    code = self.getCode('ior array upper declaration begin')
    code.write(self.typeMap['int'])
    interface.addCode(code)

    code = self.getCode('ior array upper declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    code.write('const ')
    self.composeCode(code, interface, 'ior array type')
    code.write('array,\n')
    code.indent()
    code.write(self.typeMap['int']+' ind)')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array upper declaration')
      self.composeCode(code, interface, 'ior array upper comment')
      self.composeCode(code, interface, 'ior array upper declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_upper')
      self.composeCode(code, interface, 'ior array upper declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array upper')
      self.composeCode(code, interface, 'ior array upper comment')
      self.composeCode(code, interface, 'ior array upper declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_upper')
      self.composeCode(code, interface, 'ior array upper declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.println('return (array && (ind < array->d_dimen) && (ind >= 0)) ? array->d_upper[ind] : -1;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateArrayAccess(self, interface, isSet = 0):
    if isSet:
      name = 'set'
    else:
      name = 'get'

    code = self.getCode('ior array '+name+' comment')
    code.println('/* '+name.capitalize()+' an element of a multi-dimensional array */')
    code.println('/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */')
    if isSet:
      code.println('/* If the incoming value is non-NULL, this function will increment the reference code of the object/interface */')
      code.println('/* If it is overwriting a non-NULL pointer, the reference count of the object/interface being overwritten will be decremented */')
    else:
      code.println('/* If the return value is non-NULL, the client owns one reference to the object/interface */')
      code.println('/* The client must decrement the reference count when done with the reference */')
    interface.addCode(code)

    code = self.getCode('ior array '+name+' declaration begin')
    if isSet:
      code.write('void')
    else:
      self.composeCode(code, interface, 'ior type')
    interface.addCode(code)

    code = self.getCode('ior array '+name+' declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    if not isSet: code.write('const ')
    self.composeCode(code, interface, 'ior array type')
    code.write('array,\n')
    code.indent()
    code.write('const '+self.typeMap['int']+' indices[]')
    if isSet:
      code.write(',\n')
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write(' value')
    code.write(')')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array '+name+' declaration')
      self.composeCode(code, interface, 'ior array '+name+' comment')
      self.composeCode(code, interface, 'ior array '+name+' declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_'+name)
      self.composeCode(code, interface, 'ior array '+name+' declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array '+name)
      self.composeCode(code, interface, 'ior array '+name+' comment')
      self.composeCode(code, interface, 'ior array '+name+' declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_'+name)
      self.composeCode(code, interface, 'ior array '+name+' declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('*result = NULL;\n')
      code.println('if (array && (array->d_dimen > 0)) {')
      code.increaseTabLevel()
      code.println(self.typeMap['int']+' i;')
      code.println('')
      code.println('result = array->d_firstElement;')
      code.println('for(i = 0; i < array->d_dimen; i++) {')
      code.increaseTabLevel()
      code.println('if ((indices[i] >= array->d_lower[i]) && (indices[i] <= array->d_upper[i])) {')
      code.increaseTabLevel()
      code.println('result += (array->d_stride[i]*(indices[i] - array->d_lower[i]));')
      code.decreaseTabLevel()
      code.println('} else {')
      code.increaseTabLevel()
      code.println('result = NULL;')
      code.println('break;')
      code.decreaseTabLevel()
      code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      if isSet:
        code.println('if (result) {')
        code.increaseTabLevel()
        if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.println('if (value) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) value->d_object)->d_epv->f_addReference))(value->d_object);')
          code.decreaseTabLevel()
          code.println('}')
          code.println('if (*result) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_deleteReference))((*result)->d_object);')
          code.decreaseTabLevel()
          code.println('}')
        code.println('*result = value;')
        code.decreaseTabLevel()
        code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      if not isSet:
        code.println('if (result) {')
        code.increaseTabLevel()
        if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.println('if (*result) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_addReference))((*result)->d_object);')
          code.decreaseTabLevel()
          code.println('}')
        code.println('return *result;')
        code.decreaseTabLevel()
        code.println('} else {')
        code.increaseTabLevel()
        if interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.indent()
          code.write('return (')
          self.composeCode(code, interface, 'ior type')
          code.write(') 0;\n')
        else:
          code.println('return NULL;')
        code.decreaseTabLevel()
        code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateFastArrayAccess(self, interface, isSet = 0):
    maxDim = 4
    if isSet:
      name = 'set'
    else:
      name = 'get'

    code = self.getCode('ior array fast '+name+' comment')
    code.println('/* '+name+' an element of a multi-dimensional array */')
    code.println('/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */')
    if isSet:
      code.println('/* If the incoming value is non-NULL, this function will increment the reference code of the object/interface */')
      code.println('/* If it is overwriting a non-NULL pointer, the reference count of the object/interface being overwritten will be decremented */')
    else:
      code.println('/* If the return value is non-NULL, the client owns one reference to the object/interface */')
      code.println('/* The client must decrement the reference count when done with the reference */')
    interface.addCode(code)

    code = self.getCode('ior array fast '+name+' declaration begin')
    self.composeCode(code, interface, 'ior array '+name+' declaration begin')
    interface.addCode(code)

    code = self.getCode('ior array fast '+name+' declaration end')
    code.write('(\n')
    code.increaseTabLevel()
    code.indent()
    if not isSet: code.write('const ')
    self.composeCode(code, interface, 'ior array type')
    code.write('array,\n')
    for d in range(1, maxDim):
      code.println(self.typeMap['int']+' i'+str(d)+',')
    code.indent()
    code.write(self.typeMap['int']+' i'+str(maxDim))
    if isSet:
      code.write(',\n')
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write(' value')
    code.write(')')
    interface.addCode(code)

    if self.declActive:
      code = self.getCode('ior array fast '+name+' declaration')
      self.composeCode(code, interface, 'ior array fast '+name+' comment')
      self.composeCode(code, interface, 'ior array fast '+name+' declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_'+name+str(maxDim))
      self.composeCode(code, interface, 'ior array fast '+name+' declaration end')
      code.write(';\n')
      code.println('')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior array fast '+name)
      self.composeCode(code, interface, 'ior array fast '+name+' comment')
      self.composeCode(code, interface, 'ior array fast '+name+' declaration begin')
      code.write(' '+self.getIORName(interface)+'__iorarray_'+name+str(maxDim))
      self.composeCode(code, interface, 'ior array fast '+name+' declaration end')
      code.write('\n')
      code.println('{')
      code.increaseTabLevel()
      code.indent()
      self.composeCode(code, interface, 'ior type')
      code.write('*result = NULL;\n')
      code.println('if (array) {')
      code.increaseTabLevel()
      code.println('result = array->d_firstElement;')
      code.println('switch(array->d_dimen) {')
      for d in range(maxDim, 0, -1):
        code.println('case '+str(d)+':')
        code.increaseTabLevel()
        code.println('if ((i'+str(d)+' >= array->d_lower['+str(d-1)+']) && (i'+str(d)+' <= array->d_upper['+str(d-1)+'])) {')
        code.increaseTabLevel()
        code.println('result += (array->d_stride['+str(d-1)+']*(i'+str(d)+' - array->d_lower['+str(d-1)+']));')
        code.decreaseTabLevel()
        code.println('} else {')
        code.increaseTabLevel()
        code.println('result = NULL;')
        code.println('break;')
        code.decreaseTabLevel()
        code.println('}')
        if d > 1: code.println('/* fall through */')
        code.decreaseTabLevel()
      code.increaseTabLevel()
      code.println('break;')
      code.decreaseTabLevel()
      code.println('default:')
      code.increaseTabLevel()
      code.println('result = NULL;')
      code.println('break;')
      code.decreaseTabLevel()
      code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      if isSet:
        code.println('if (result) {')
        code.increaseTabLevel()
        if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.println('if (value) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) value->d_object)->d_epv->f_addReference))(value->d_object);')
          code.decreaseTabLevel()
          code.println('}')
          code.println('if (*result) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_deleteReference))((*result)->d_object);')
          code.decreaseTabLevel()
          code.println('}')
        code.println('*result = value;')
        code.decreaseTabLevel()
        code.println('}')
      if not isSet:
        code.println('if (result) {')
        code.increaseTabLevel()
        if not interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.println('if (*result) {')
          code.increaseTabLevel()
          # This works since every class has a BaseClass object in front
          code.println('(*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_addReference))((*result)->d_object);')
          code.decreaseTabLevel()
          code.println('}')
        code.println('return *result;')
        code.decreaseTabLevel()
        code.println('} else {')
        code.increaseTabLevel()
        if interface.isInstanceOf('SIDLLanguage.Enumeration'):
          code.indent()
          code.write('return (')
          self.composeCode(code, interface, 'ior type')
          code.write(') 0;\n')
        else:
          code.println('return NULL;')
        code.decreaseTabLevel()
        code.println('}')
      code.decreaseTabLevel()
      code.println('}')
      code.println('')
      interface.addCode(code)
    return

  def generateExternalEntryPoint(self, code, interface, codeName, funcName):
    code.indent()
    self.composeCode(code, interface, 'ior array '+codeName+' declaration begin')
    code.write(' (*'+funcName+')')
    self.composeCode(code, interface, 'ior array '+codeName+' declaration end')
    code.write(';\n')
    return

  def generateExternalEntryPoints(self, interface):
    name = self.getIORName(interface)

    if self.declActive:
      # TODO: Must make these Method nodes, and use the make cast function
      code = self.getCode('ior external entry points declaration')
      code.indent()
      self.composeCode(code, interface, 'ior function table type struct')
      code.write(' {\n')
      code.increaseTabLevel()
      if interface.isInstanceOf('SIDLLanguage.Class'):
        if not SIDLLanguage.Class.Class(interface).getAbstract():
          code.indent()
          self.composeCode(code, interface, 'ior type')
          code.write('(*createObject)(void);\n')
        code.println('void (*remoteDummy)(void);')
        if self.hasStaticMethod(interface):
          code.indent()
          self.composeCode(code, interface, 'ior static function table type')
          code.write(' (*getStaticEPV)(void);\n')
      self.generateExternalEntryPoint(code, interface, 'constructor', 'createArray')
      self.generateExternalEntryPoint(code, interface, 'borrow',      'borrowArray')
      self.generateExternalEntryPoint(code, interface, 'destructor',  'destroyArray')
      self.generateExternalEntryPoint(code, interface, 'dimension',   'getDimen')
      self.generateExternalEntryPoint(code, interface, 'lower',       'getLower')
      self.generateExternalEntryPoint(code, interface, 'upper',       'getUpper')
      self.generateExternalEntryPoint(code, interface, 'get',         'getElement')
      self.generateExternalEntryPoint(code, interface, 'fast get',    'getElement4')
      self.generateExternalEntryPoint(code, interface, 'set',         'setElement')
      self.generateExternalEntryPoint(code, interface, 'fast set',    'setElement4')
      code.decreaseTabLevel()
      code.println('};')
      code.println('')

      code.println('/* This function returns a pointer to a static structure of pointers to function entry points */')
      code.println('/* It is called upon loading the DLL */')
      code.indent()
      code.write('const ')
      self.composeCode(code, interface, 'ior function table type')
      self.composeCode(code, interface, 'ior function table accessor')
      code.write('(void);\n')
      interface.addCode(code)

    if self.defnActive:
      code = self.getCode('ior external entry points')
      code.indent()
      code.write('static const ')
      self.composeCode(code, interface, 'ior function table type struct')
      code.write(' s_externalEntryPoints = {\n')
      code.increaseTabLevel()
      if interface.isInstanceOf('SIDLLanguage.Class'):
        if not SIDLLanguage.Class.Class(interface).getAbstract():
          code.println(name+'__new,')
        code.println('NULL /*'+name+'__remote*/,')
        if self.hasStaticMethod(interface):
          code.indent()
          self.composeCode(code, interface, 'ior static function table accessor')
          code.write(',\n')
      code.println(name+'__iorarray_create,')
      code.println(name+'__iorarray_borrow,')
      code.println(name+'__iorarray_destroy,')
      code.println(name+'__iorarray_dimen,')
      code.println(name+'__iorarray_lower,')
      code.println(name+'__iorarray_upper,')
      code.println(name+'__iorarray_get,')
      code.println(name+'__iorarray_get4,')
      code.println(name+'__iorarray_set,')
      code.println(name+'__iorarray_set4')
      code.decreaseTabLevel()
      code.println('};')
      code.println('')

      code.println('/* This function returns a pointer to a static structure of pointers to function entry points */')
      code.println('/* It is called upon loading the DLL */')
      code.indent()
      code.write('const ')
      self.composeCode(code, interface, 'ior function table type')
      self.composeCode(code, interface, 'ior function table accessor')
      code.write('()\n')
      code.println('{')
      code.increaseTabLevel()
      code.println('return &s_externalEntryPoints;')
      code.decreaseTabLevel()
      code.println('}')
      interface.addCode(code)
    return
# DO-NOT-DELETE splicer.end(_final)

