#ifndef included_SIDLLanguage_AccessOpaque_IOR
#include "SIDLLanguage/AccessOpaque_IOR/AccessOpaque_IOR.h"
#endif
#ifndef included_SIDL_BaseClass_IOR
#include "SIDL/BaseClass_IOR/BaseClass_IOR.h"
#endif

/* Static variables for managing EPV initialization */

/* The IOR array structure */
/*   Macros to read this are defined in SIDLArray.h */
struct SIDLLanguage_AccessOpaque__array {
  struct SIDLLanguage_AccessOpaque__object ** d_firstElement;
  int32_t *d_lower;
  int32_t *d_upper;
  int32_t *d_stride;
  SIDL_bool d_borrowed;
  int32_t d_dimen;
};

static struct SIDLLanguage_AccessOpaque__array *newArray(int32_t dimen, const int32_t lower[], const int32_t upper[]) {
  static const size_t arraySize = sizeof(struct SIDLLanguage_AccessOpaque__array) + (sizeof(int32_t) - (sizeof(struct SIDLLanguage_AccessOpaque__array)%sizeof(int32_t)))%sizeof(int32_t);
  struct SIDLLanguage_AccessOpaque__array *result = (struct SIDLLanguage_AccessOpaque__array *) malloc(arraySize + 3*sizeof(int32_t)*dimen);

  result->d_dimen    = dimen;
  result->d_borrowed = 0;
  result->d_lower    = (int32_t *) ((char *) result + arraySize);
  result->d_upper    = result->d_lower + dimen;
  result->d_stride   = result->d_upper + dimen;
  memcpy(result->d_lower, lower, sizeof(int32_t)*dimen);
  memcpy(result->d_upper, upper, sizeof(int32_t)*dimen);
  return result;
}

/* Create a dense array of the given dimension with specified index bounds */
/*   This array owns and manages its data */
/*   All object pointers are initialized to NULL */
struct SIDLLanguage_AccessOpaque__array * SIDLLanguage_AccessOpaque__iorarray_create(
  int32_t dimen,
  const int32_t lower[],
  const int32_t upper[])
{
  struct SIDLLanguage_AccessOpaque__array *result = newArray(dimen, lower, upper);
  int32_t size;
  int32_t i;

  for(i = 0, size = 1; i < dimen; i++) {
    result->d_stride[i] = size;
    size *= (upper[i] - lower[i] + 1);
  }
  size *= sizeof(struct SIDLLanguage_AccessOpaque__object *);
  result->d_firstElement = (struct SIDLLanguage_AccessOpaque__object **) malloc((size_t) size);
  memset(result->d_firstElement, 0, (size_t) size);
  return result;
}

/* Create an array using user-provided memory */
struct SIDLLanguage_AccessOpaque__array * SIDLLanguage_AccessOpaque__iorarray_borrow(
  struct SIDLLanguage_AccessOpaque__object ** firstElement,
  int32_t dimen,
  const int32_t lower[],
  const int32_t upper[],
  const int32_t stride[])
{
  struct SIDLLanguage_AccessOpaque__array *result = newArray(dimen, lower, upper);

  memcpy(result->d_stride, stride, sizeof(int32_t)*dimen);
  result->d_firstElement = firstElement;
  result->d_borrowed     = 1;
  return result;
}

/* Destroy the array (NULL arrays are ignored) */
void SIDLLanguage_AccessOpaque__iorarray_destroy(
  struct SIDLLanguage_AccessOpaque__array *array)
{
  if (array) {
    const int32_t dimen = array->d_dimen;

    if (!(array->d_borrowed)) {
      if (dimen > 0) {
        struct SIDLLanguage_AccessOpaque__object **start = array->d_firstElement;
        struct SIDLLanguage_AccessOpaque__object **end;
        int32_t size = 1;

        if (dimen > 1) {
          size = array->d_stride[dimen-1];
        }
        size *= (array->d_upper[dimen-1] - array->d_lower[dimen-1] + 1);
        end   = start + size;
        while(start < end) {
          if (*start) {
            (*(((struct SIDL_BaseClass__object *) (*start)->d_object)->d_epv->f_deleteReference))((*start)->d_object);
            *start = NULL;
          }
          ++start;
        }
      }
      free(array->d_firstElement);
    }
    memset(array, 0, sizeof(struct SIDLLanguage_AccessOpaque__array) + 3*sizeof(int32_t)*dimen);
    free(array);
  }
}

/* Return the number of dimensions in the array */
/*   If the array pointer is NULL, zero is returned */
int32_t SIDLLanguage_AccessOpaque__iorarray_dimen(
  const struct SIDLLanguage_AccessOpaque__array *array)
{
  return array ? array->d_dimen : 0;
}

/* Return the lower bound on dimension ind */
/*   If the array pointer is NULL or ind is not a valid dimension, zero is returned */
int32_t SIDLLanguage_AccessOpaque__iorarray_lower(
  const struct SIDLLanguage_AccessOpaque__array *array,
  int32_t ind)
{
  return (array && (ind < array->d_dimen) && (ind >= 0)) ? array->d_lower[ind] : 0;
}

/* Return the upper bound on dimension ind */
/*   If the array pointer is NULL or ind is not a valid dimension, -1 is returned */
int32_t SIDLLanguage_AccessOpaque__iorarray_upper(
  const struct SIDLLanguage_AccessOpaque__array *array,
  int32_t ind)
{
  return (array && (ind < array->d_dimen) && (ind >= 0)) ? array->d_upper[ind] : -1;
}

/* Get an element of a multi-dimensional array */
/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */
/* If the return value is non-NULL, the client owns one reference to the object/interface */
/* The client must decrement the reference count when done with the reference */
struct SIDLLanguage_AccessOpaque__object * SIDLLanguage_AccessOpaque__iorarray_get(
  const struct SIDLLanguage_AccessOpaque__array *array,
  const int32_t indices[])
{
  struct SIDLLanguage_AccessOpaque__object **result = NULL;
  if (array && (array->d_dimen > 0)) {
    int32_t i;

    result = array->d_firstElement;
    for(i = 0; i < array->d_dimen; i++) {
      if ((indices[i] >= array->d_lower[i]) && (indices[i] <= array->d_upper[i])) {
        result += (array->d_stride[i]*(indices[i] - array->d_lower[i]));
      } else {
        result = NULL;
        break;
      }
    }
  }
  if (result) {
    if (*result) {
      (*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_addReference))((*result)->d_object);
    }
    return *result;
  } else {
    return NULL;
  }
}

/* get an element of a multi-dimensional array */
/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */
/* If the return value is non-NULL, the client owns one reference to the object/interface */
/* The client must decrement the reference count when done with the reference */
struct SIDLLanguage_AccessOpaque__object * SIDLLanguage_AccessOpaque__iorarray_get4(
  const struct SIDLLanguage_AccessOpaque__array *array,
  int32_t i1,
  int32_t i2,
  int32_t i3,
  int32_t i4)
{
  struct SIDLLanguage_AccessOpaque__object **result = NULL;
  if (array) {
    result = array->d_firstElement;
    switch(array->d_dimen) {
    case 4:
      if ((i4 >= array->d_lower[3]) && (i4 <= array->d_upper[3])) {
        result += (array->d_stride[3]*(i4 - array->d_lower[3]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 3:
      if ((i3 >= array->d_lower[2]) && (i3 <= array->d_upper[2])) {
        result += (array->d_stride[2]*(i3 - array->d_lower[2]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 2:
      if ((i2 >= array->d_lower[1]) && (i2 <= array->d_upper[1])) {
        result += (array->d_stride[1]*(i2 - array->d_lower[1]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 1:
      if ((i1 >= array->d_lower[0]) && (i1 <= array->d_upper[0])) {
        result += (array->d_stride[0]*(i1 - array->d_lower[0]));
      } else {
        result = NULL;
        break;
      }
      break;
    default:
      result = NULL;
      break;
    }
  }
  if (result) {
    if (*result) {
      (*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_addReference))((*result)->d_object);
    }
    return *result;
  } else {
    return NULL;
  }
}

/* Set an element of a multi-dimensional array */
/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */
/* If the incoming value is non-NULL, this function will increment the reference code of the object/interface */
/* If it is overwriting a non-NULL pointer, the reference count of the object/interface being overwritten will be decremented */
void SIDLLanguage_AccessOpaque__iorarray_set(
  struct SIDLLanguage_AccessOpaque__array *array,
  const int32_t indices[],
  struct SIDLLanguage_AccessOpaque__object * value)
{
  struct SIDLLanguage_AccessOpaque__object **result = NULL;
  if (array && (array->d_dimen > 0)) {
    int32_t i;

    result = array->d_firstElement;
    for(i = 0; i < array->d_dimen; i++) {
      if ((indices[i] >= array->d_lower[i]) && (indices[i] <= array->d_upper[i])) {
        result += (array->d_stride[i]*(indices[i] - array->d_lower[i]));
      } else {
        result = NULL;
        break;
      }
    }
    if (result) {
      if (value) {
        (*(((struct SIDL_BaseClass__object *) value->d_object)->d_epv->f_addReference))(value->d_object);
      }
      if (*result) {
        (*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_deleteReference))((*result)->d_object);
      }
      *result = value;
    }
  }
}

/* set an element of a multi-dimensional array */
/* This will use the indices provided up to the actual dimension of the array, ignoring extra indices */
/* If the incoming value is non-NULL, this function will increment the reference code of the object/interface */
/* If it is overwriting a non-NULL pointer, the reference count of the object/interface being overwritten will be decremented */
void SIDLLanguage_AccessOpaque__iorarray_set4(
  struct SIDLLanguage_AccessOpaque__array *array,
  int32_t i1,
  int32_t i2,
  int32_t i3,
  int32_t i4,
  struct SIDLLanguage_AccessOpaque__object * value)
{
  struct SIDLLanguage_AccessOpaque__object **result = NULL;
  if (array) {
    result = array->d_firstElement;
    switch(array->d_dimen) {
    case 4:
      if ((i4 >= array->d_lower[3]) && (i4 <= array->d_upper[3])) {
        result += (array->d_stride[3]*(i4 - array->d_lower[3]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 3:
      if ((i3 >= array->d_lower[2]) && (i3 <= array->d_upper[2])) {
        result += (array->d_stride[2]*(i3 - array->d_lower[2]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 2:
      if ((i2 >= array->d_lower[1]) && (i2 <= array->d_upper[1])) {
        result += (array->d_stride[1]*(i2 - array->d_lower[1]));
      } else {
        result = NULL;
        break;
      }
      /* fall through */
    case 1:
      if ((i1 >= array->d_lower[0]) && (i1 <= array->d_upper[0])) {
        result += (array->d_stride[0]*(i1 - array->d_lower[0]));
      } else {
        result = NULL;
        break;
      }
      break;
    default:
      result = NULL;
      break;
    }
  }
  if (result) {
    if (value) {
      (*(((struct SIDL_BaseClass__object *) value->d_object)->d_epv->f_addReference))(value->d_object);
    }
    if (*result) {
      (*(((struct SIDL_BaseClass__object *) (*result)->d_object)->d_epv->f_deleteReference))((*result)->d_object);
    }
    *result = value;
  }
}

static const struct SIDLLanguage_AccessOpaque__external s_externalEntryPoints = {
  SIDLLanguage_AccessOpaque__iorarray_create,
  SIDLLanguage_AccessOpaque__iorarray_borrow,
  SIDLLanguage_AccessOpaque__iorarray_destroy,
  SIDLLanguage_AccessOpaque__iorarray_dimen,
  SIDLLanguage_AccessOpaque__iorarray_lower,
  SIDLLanguage_AccessOpaque__iorarray_upper,
  SIDLLanguage_AccessOpaque__iorarray_get,
  SIDLLanguage_AccessOpaque__iorarray_get4,
  SIDLLanguage_AccessOpaque__iorarray_set,
  SIDLLanguage_AccessOpaque__iorarray_set4
};

/* This function returns a pointer to a static structure of pointers to function entry points */
/* It is called upon loading the DLL */
const struct SIDLLanguage_AccessOpaque__external *SIDLLanguage_AccessOpaque__externals()
{
  return &s_externalEntryPoints;
}
