# DO-NOT-DELETE splicer.begin(_initial)
import SIDLLanguage.CodePurpose
# DO-NOT-DELETE splicer.end(_initial)

import SIDL.NullIORException
import SIDL.BaseInterface
import SIDL.BaseException
import SIDLLanguage.AST
import SIDLLanguage.VisitorException
import SIDLLanguage.Specification
import SIDLLanguage.Version
import SIDLLanguage.Package
import SIDLLanguage.Interface
import SIDLLanguage.Class
import SIDLLanguage.Method
import SIDLLanguage.Type
import SIDLLanguage.Enumeration
import SIDLLanguage.Enumerator
import SIDLLanguage.Array
import SIDLLanguage.Parameter
import SIDLLanguage.Comment

# DO-NOT-DELETE splicer.begin(_before_type)
import sys
# DO-NOT-DELETE splicer.end(_before_type)

class Generator:
  '''
  // This class generates all code for Python
  '''
  # All calls to SIDL methods should use __IORself

  def __init__(self, IORself):
    self.__IORself = IORself
    # DO-NOT-DELETE splicer.begin(__init__)
    self.language      = 'Python'
    self.indentString  = '  '
    self.markName      = 'pythonGeneratorMark'
    self.typeOnly      = [0]
    self.purposeActive = {'stub': [0],
                          'IOR':  [0],
                          'skel': [0],
                          'impl': [0]}
    self.declActive    = 0
    self.defnActive    = 0
    self.stub          = SIDLLanguage.CodePurpose.CodePurpose(SIDL.Loader.createClass('ANL.SIDLCompilerI.Stub'))
    self.skel          = SIDLLanguage.CodePurpose.CodePurpose(SIDL.Loader.createClass('ANL.SIDLCompilerI.Skeleton'))
    self.impl          = SIDLLanguage.CodePurpose.CodePurpose(SIDL.Loader.createClass('ANL.SIDLCompilerI.Implementation'))
    # Map SIDL types to Python/C types
    self.typeMap = {"bool":     "int",
                    "char":     "char",
                    "dcomplex": "Py_complex",
                    "double":   "double",
                    "fcomplex": "Py_complex",
                    "float":    "float",
                    "int":      "int",
                    "long":     "long long int",
                    "opaque":   "void *",
                    "string":   "char *",
                    "void":     "void"}
    self.iorDefaultValue = {"bool":     "FALSE",
                            "char":     "'\\0'",
                            "dcomplex": "{0.0, 0.0}",
                            "double":   "0.0",
                            "fcomplex": "{0.0, 0.0}",
                            "float":    "0.0",
                            "int":      "0",
                            "long":     "0L",
                            "opaque":   "NULL",
                            "string":   "NULL",
                            "void":     ""}
    # DO-NOT-DELETE splicer.end(__init__)

  def getPurposeActive(self, purpose):
    '''
    // Get the activity state of the given generation purpose
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # purpose string
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getPurposeActive)
    return self.purposeActive[purpose][-1]
    # DO-NOT-DELETE splicer.end(getPurposeActive)

  def setPurposeActive(self, purpose, isActive):
    '''
    // Set the activity state of the given generation purpose
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # purpose string
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setPurposeActive)
    if self.purposeActive.has_key(purpose):
      self.purposeActive[purpose][-1] = isActive
    else:
      self.purposeActive[purpose] = [isActive]
    # DO-NOT-DELETE splicer.end(setPurposeActive)

  def getDeclarationActive(self):
    '''
    // Get the flag determining whether to generate declarations (headers)
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getDeclarationActive)
    return self.declActive
    # DO-NOT-DELETE splicer.end(getDeclarationActive)

  def setDeclarationActive(self, isActive):
    '''
    // Set the flag determining whether to generate declarations (headers)
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setDeclarationActive)
    self.declActive = isActive
    # DO-NOT-DELETE splicer.end(setDeclarationActive)

  def getDefinitionActive(self):
    '''
    // Get the flag determining whether to generate definitions (source)
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getDefinitionActive)
    return self.defnActive
    # DO-NOT-DELETE splicer.end(getDefinitionActive)

  def setDefinitionActive(self, isActive):
    '''
    // Set the flag determining whether to generate definitions (source)
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # isActive bool
    # DO-NOT-DELETE splicer.begin(setDefinitionActive)
    self.defnActive = isActive
    # DO-NOT-DELETE splicer.end(setDefinitionActive)

  def getRepository(self):
    '''
    // Retrieve auxilliary ASTs
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # array _return
    # DO-NOT-DELETE splicer.begin(getRepository)
    return []
    # DO-NOT-DELETE splicer.end(getRepository)

  def setRepository(self, repository):
    '''
    // Set auxilliary ASTs
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # repository array
    # DO-NOT-DELETE splicer.begin(setRepository)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(setRepository)

  def visitAST(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.AST
    # DO-NOT-DELETE splicer.begin(visitAST)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitAST)

  def visitSpecification(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Specification
    # DO-NOT-DELETE splicer.begin(visitSpecification)
    if self.isMarked(node): return
    self.markNode(node)

    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitSpecification)

  def visitVersion(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Version
    # DO-NOT-DELETE splicer.begin(visitVersion)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitVersion)

  def visitPackage(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Package
    # DO-NOT-DELETE splicer.begin(visitPackage)
    if self.isMarked(node): return
    self.markNode(node)

    purpose = self.stub
    if self.getPurposeActive(purpose.getName()):
      self.generatePackage(node, purpose)
    purpose = self.impl
    if self.getPurposeActive(purpose.getName()):
      self.generatePackage(node, purpose)
    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitPackage)

  def visitInterface(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Interface
    # DO-NOT-DELETE splicer.begin(visitInterface)
    self.generateInterfaceType(node)

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    # All method code must be generated
    self.defaultVisit(node)
    try:
      for interface in node.getParents():
        interface.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    try:
      purpose = self.stub
      if self.getPurposeActive(purpose.getName()):
        self.generateInterfaceStub(node, purpose)
    except Exception, e:
      import traceback
      self.raiseVisitorException(str(e), traceback.format_tb(sys.exc_info()[2]))
    # DO-NOT-DELETE splicer.end(visitInterface)

  def visitClass(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Class
    # DO-NOT-DELETE splicer.begin(visitClass)
    self.generateInterfaceType(node)

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    # All method code must be generated
    self.defaultVisit(node)
    try:
      for interface in node.getInterfaces():
        interface.accept(self.__IORself)
      for interface in node.getParents():
        interface.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    try:
      purpose = self.stub
      if self.getPurposeActive(purpose.getName()):
        self.generateInterfaceStub(node, purpose)
      purpose = self.skel
      if self.getPurposeActive(purpose.getName()):
        self.generateClassSkel(node, purpose)
      purpose = self.impl
      if self.getPurposeActive(purpose.getName()):
        self.generateClassImpl(node, purpose)
    except Exception, e:
      import traceback
      self.raiseVisitorException(str(e), traceback.format_tb(sys.exc_info()[2]))
    # DO-NOT-DELETE splicer.end(visitClass)

  def visitMethod(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Method
    # DO-NOT-DELETE splicer.begin(visitMethod)
    if self.isMarked(node): return
    self.markNode(node)

    try:
      self.typeOnly.append(1)
      node.getReturnType().accept(self.__IORself)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    self.defaultVisit(node)
    try:
      purpose = self.stub
      if self.getPurposeActive(purpose.getName()):
        self.generateMethodComponents(node, purpose)
      purpose = self.skel
      if self.getPurposeActive(purpose.getName()):
        self.generateMethodComponents(node, purpose)
    except SIDLLanguage.CodeGeneratorException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())
    except Exception, e:
      import traceback
      self.raiseVisitorException(str(e), traceback.format_tb(sys.exc_info()[2]))
    # DO-NOT-DELETE splicer.end(visitMethod)

  def visitType(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Type
    # DO-NOT-DELETE splicer.begin(visitType)
    if self.isMarked(node): return
    self.markNode(node)

    pythonType = self.getCode('python type')
    try:
      pythonType.write(self.typeMap[node.getIdentifier()])
    except KeyError:
      self.raiseVisitorException('No base type: '+str(node.getIdentifier()))
    node.addCode(pythonType)
    self.marshallReturnType(node, node)
    # DO-NOT-DELETE splicer.end(visitType)

  def visitEnumeration(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumeration
    # DO-NOT-DELETE splicer.begin(visitEnumeration)
    pythonType = self.getCode('python type')
    pythonType.write(node.getFullIdentifier())
    node.addCode(pythonType)
    self.marshallReturnType(node, node)

    if self.typeOnly[-1]: return
    if self.isMarked(node): return
    self.markNode(node)

    purpose = self.stub
    if self.getPurposeActive(purpose.getName()):
      code = self.getCode(purpose.getName()+' contents')
      for child in node.getChildren():
        enum = SIDLLanguage.Enumerator.Enumerator(child)
        code.println(enum.getIdentifier()+' = '+str(enum.getValue()))
      node.addCode(code)
    # DO-NOT-DELETE splicer.end(visitEnumeration)

  def visitEnumerator(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumerator
    # DO-NOT-DELETE splicer.begin(visitEnumerator)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitEnumerator)

  def visitArray(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Array
    # DO-NOT-DELETE splicer.begin(visitArray)
    if self.isMarked(node): return
    self.markNode(node)

    try:
      self.typeOnly.append(1)
      node.getType().accept(self.__IORself)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    pythonType = self.getCode('python type')
    pythonType.write('SIDL::array<')
    self.composeCode(pythonType, node.getType(), 'python type')
    pythonType.write('>')
    node.addCode(pythonType)
    self.marshallReturnType(node, node)
    # DO-NOT-DELETE splicer.end(visitArray)

  def visitParameter(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Parameter
    # DO-NOT-DELETE splicer.begin(visitParameter)
    if self.isMarked(node): return
    self.markNode(node)
    type = node.getType()

    try:
      self.typeOnly.append(1)
      type.accept(self.__IORself)
      self.typeOnly.pop()
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    name = node.getIdentifier()
    mode = node.getAccessMode()

    pythonParam = self.getCode('python parameter')
    pythonParam.write(self.getModeComment(mode))
    self.composeCode(pythonParam, type, 'python type')
    if self.isOut(mode):
      pythonParam.write('&')
    pythonParam.write(' '+name)
    node.addCode(pythonParam)

    self.marshallType(node, node.getType(), name, mode)
    # DO-NOT-DELETE splicer.end(visitParameter)

  def visitComment(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Comment
    # DO-NOT-DELETE splicer.begin(visitComment)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitComment)

# DO-NOT-DELETE splicer.begin(_final)
  ##########################
  # Basic Visitor Operations
  def raiseVisitorException(self, message, trace = ''):
    e = SIDLLanguage.VisitorException.VisitorException()
    e.setMessage(message)
    if isinstance(trace, list):
      for line in trace: e.addToStackTrace(line.strip())
    elif trace:
      e.addToStackTrace(trace)
    e.addToTrace(__file__, sys._getframe(1).f_lineno, sys._getframe().f_code.co_name)
    raise SIDLLanguage.VisitorException.Exception, e

  def defaultVisit(self, ast):
    try:
      for child in ast.getChildren():
        child.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())
    return

  def isMarked(self, node, name = ''):
    if not name: name = self.markName
    return node.hasAttribute(name)

  def markNode(self, node, name = ''):
    if not name: name = self.markName
    seen = SIDLLanguage.Attribute.Attribute(SIDL.Loader.createClass('ANL.SIDLCompilerI.Attribute'))
    seen.setName(name)
    node.addAttribute(seen)
    return

  def unmarkNode(self, node, name = ''):
    if not name: name = self.markName
    if node.hasAttribute(name):
      node.deleteAttribute(name)
    return

  def unmarkNetwork(self, node, name = ''):
    self.unmarkNode(node, name)
    for interface in node.getParents():
      self.unmarkNetwork(interface, name)
    if node.isInstanceOf('SIDLLanguage.Class'):
      for interface in SIDLLanguage.Class.Class(node).getInterfaces():
        self.unmarkNetwork(interface, name)
    return

  #########################
  # Should go in Code class
  def composeCode(self, parent, node, label, fail = 1, forceBreak = 0, language = None):
    if not language: language = self.language
    if node.hasCodeBlock(language, label):
      return self.composeCodeBlock(parent, node.getCodeBlock(language, label), forceBreak)
    elif fail:
      self.raiseVisitorException('Unable to find code block "'+label+'" in '+node.getFullIdentifier())
    return None

  def composeCodeBlock(self, parent, code, forceBreak = 0):
    self.composeSourceBlock(parent, code.getSource(), forceBreak)
    return code

  def composeSourceBlock(self, parent, source, forceBreak = 0):
    indent = ''
    for i in range(parent.getTabLevel()):
      indent += self.indentString
    if indent and (source.find('\n') >= 0 or forceBreak):
      # Indent start and all newlines
      source = indent+source.replace('\n', '\n'+indent)
      # Remove extra space from the end
      index  = 0
      while source[index-1] == ' ': index -= 1
      if index < 0:
        source = source[:index]
    if forceBreak and not source[-1] == '\n':
      source = source+'\n'
    parent.write(source)
    return

  ###################
  # IOR naming scheme
  def getIORCastMethod(self):
    return 'f__cast_ior'

  def getIORName(self, node):
    return node.getFullIdentifier().replace('.', '_')

  def getIORStructName(self, interface, name):
    if not interface.isInstanceOf('SIDLLanguage.Class'):
      return name+'->d_object'
    else:
      return name

  #############################
  # Python/C specific code structures
  def getModeName(self, mode):
    if mode == SIDLLanguage.ParameterAccessMode.IN:
      return 'in'
    elif mode == SIDLLanguage.ParameterAccessMode.INOUT:
      return 'inout'
    elif mode == SIDLLanguage.ParameterAccessMode.OUT:
      return 'out'
    else:
      self.raiseVisitorException('Invalid parameter access mode: '+str(mode))

  def getModeComment(self, mode):
    return '/*'+self.getModeName(mode)+'*/ '

  ##################################
  # Basic Code Generation Operations
  def getCode(self, label = '', language = ''):
    code = SIDLLanguage.Code.Code(SIDL.Loader.createClass('ANL.SIDLCompilerI.Code'))
    if language:
      code.setLanguage(language)
    else:
      code.setLanguage(self.language)
    if label:
      code.setLabel(label)
    return code

  def isIn(self, mode):
    return mode == SIDLLanguage.ParameterAccessMode.IN or mode == SIDLLanguage.ParameterAccessMode.INOUT

  def isOut(self, mode):
    return mode == SIDLLanguage.ParameterAccessMode.OUT or mode == SIDLLanguage.ParameterAccessMode.INOUT

  def isInOut(self, mode):
    return mode == SIDLLanguage.ParameterAccessMode.INOUT

  def isBase(self, interface):
    return interface.getFullIdentifier() in ['SIDL.BaseInterface', 'SIDL.BaseClass', 'SIDL.BaseException', 'SIDL.NullIORException', 'SIDL.Loader', 'SIDL.DLL']

  def isException(self, interface):
    '''This will have to change is BaseException becomes an interface'''
    if interface.getFullIdentifier() == 'SIDL.BaseException': return 1
    for parent in interface.getParents():
      isExc = self.isException(parent)
      if isExc: return isExc
    return 0

  def hasStaticMethod(self, interface):
    for method in interface.getMethods():
      if method.getStatic():
        return 1
    return 0

  def getBase(self, interface):
    if interface.isInstanceOf('SIDLLanguage.Class'):
      if interface.getFullIdentifier() == 'SIDL.BaseClass': return interface
    else:
      if interface.getFullIdentifier() == 'SIDL.BaseInterface': return interface
    for parent in interface.getParents():
      base = self.getBase(parent)
      if base: return base
    return None

  def getBaseException(self, exception):
    if exception.getFullIdentifier() == 'SIDL.BaseException':
      return exception
    for sclass in exception.getParents():
      base = self.getBaseException(sclass)
      if base:
        return base
    return None

  def getIORExceptions(self, method):
    exceptions = []
    for exception in method.getExceptions():
      if not exception.getFullIdentifier() == 'SIDL.NullIORException':
        exceptions.append(exception)
    return exceptions

  def escapeComment(self, line):
    return line.replace('\\', '\\\\').replace('"', '\\"').replace("'", "\\'")

  def commentsToString(self, node):
    end = '\\n\\\n'
    str = ''
    for comment in node.getComments():
      for line in comment.getComment().split('\n'):
        if line: str += self.escapeComment(line)+end
    return str

  def commentsToDocString(self, code, node):
    if len(node.getComments()):
      code.println('\'\'\'')
      for comment in node.getComments():
        for line in comment.getComment().split('\n'):
          if line: code.println(self.escapeComment(line))
      code.println('\'\'\'')
    return

  ########################
  # Marshalling Operations
  def needLocalVariable(self, type):
    if type.getBaseType():
      if type.getIdentifier() in ['bool', 'dcomplex', 'fcomplex', 'int', 'long']:
        useLocal = 1
      else:
        useLocal = 0
    elif type.isInstanceOf('SIDLLanguage.Enumeration'):
      useLocal = 1
    elif type.isInstanceOf('SIDLLanguage.Array'):
      useLocal = 0
    else:
      useLocal = 0
    return useLocal

  def hasIOR(self, type):
    if type.getBaseType() or type.isInstanceOf('SIDLLanguage.Enumeration'):
      return 0
    return 1

  def getFormatString(self, type, pack):
    if type.getBaseType():
      typeName = type.getIdentifier()

      if   typeName == 'bool':     format = 'i'
      elif typeName == 'char':     format = 'c'
      elif typeName == 'dcomplex' or typeName == 'fcomplex':
        if pack:                   format = 'D'
        else:                      format = 'N'
      elif typeName == 'double':   format = 'd'
      elif typeName == 'float':    format = 'f'
      elif typeName == 'int':      format = 'l'
      elif typeName == 'long':     format = 'L'
      elif typeName == 'opaque':   format = 'O&'
      elif typeName == 'string':   format = 'z'
      else:                        format = ''
    elif type.isInstanceOf('SIDLLanguage.Enumeration'):
      format = 'l'
    else:
      format = 'O&'
    return format

  def marshallReturnType(self, node, type):
    return self.marshallType(node, type, '_result', SIDLLanguage.ParameterAccessMode.OUT, isReturn = 1)

  def marshallType(self, node, type, name, mode, isReturn = 0):
    pythonDeclare      = self.getCode('python argument declare')
    pythonPackArg      = self.getCode('python argument packing')
    pythonPackFormat   = self.getCode('python argument packing format')
    pythonPackParam    = self.getCode('python argument packing parameter')
    pythonPackSkel     = self.getCode('python argument packing skeleton')
    pythonUnpackArg    = self.getCode('python argument unpacking')
    pythonUnpackFormat = self.getCode('python argument unpacking format')
    pythonUnpackParam  = self.getCode('python argument unpacking parameter')
    pythonUnpackSkel   = self.getCode('python argument unpacking skeleton')
    pythonCleanup      = self.getCode('python argument cleanup')
    iorPackArg         = self.getCode('ior argument packing')
    iorUnpackArg       = self.getCode('ior argument unpacking')

    pythonPackFormat.write(self.getFormatString(type, 1))
    pythonUnpackFormat.write(self.getFormatString(type, 0))

    if type.getBaseType():
      self.marshallBaseType(type, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn)
    elif type.isInstanceOf('SIDLLanguage.Enumeration'):
      self.marshallEnumeration(type, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn)
    elif type.isInstanceOf('SIDLLanguage.Array'):
      self.marshallArray(SIDLLanguage.Array.Array(type), name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonPackSkel, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn)
    else:
      self.marshallInterface(type, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonPackSkel, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn)

    if self.needLocalVariable(type):
      pythonPackParam.write('_local_')
    pythonPackParam.write(name)

    if self.needLocalVariable(type):
      pythonUnpackSkel.write('_local_')
    pythonUnpackSkel.write(name)

    node.addCode(pythonDeclare)
    node.addCode(pythonPackArg)
    node.addCode(pythonUnpackArg)
    node.addCode(pythonPackParam)
    node.addCode(pythonUnpackParam)
    node.addCode(pythonPackFormat)
    node.addCode(pythonUnpackFormat)
    node.addCode(pythonPackSkel)
    node.addCode(pythonUnpackSkel)
    node.addCode(pythonCleanup)
    node.addCode(iorPackArg)
    node.addCode(iorUnpackArg)
    return

  def marshallBaseType(self, type, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn):
    typeName   = type.getIdentifier()
    iorType    = type.getCodeBlock('C', 'ior type').getSource()
    pythonType = self.typeMap[typeName]
    if self.isOut(mode) and not isReturn:
      iorName = '*'+name
    else:
      iorName = name
    localName = '_local_'+name

    if typeName == 'void': return
    pythonDeclare.println(iorType+' '+name+';')
    if typeName == 'opaque':
      pythonPackParam.write('(void *) SIDL_Opaque_Convert, ')
      pythonUnpackSkel.write('(void *) SIDL_Opaque_Convert, ')
      pythonUnpackParam.write('(void *) SIDL_Opaque_Create, ')
    elif typeName == 'dcomplex' or typeName == 'fcomplex':
      pythonUnpackParam.write('PyComplex_FromCComplex(')
    pythonPackParam.write('&')
    if self.needLocalVariable(type):
      pythonUnpackParam.write('_local_')
    pythonUnpackParam.write(name)
    if typeName == 'dcomplex' or typeName == 'fcomplex':
      pythonUnpackParam.write(')')
    if typeName == 'string':
      if self.isOut(mode):
        if self.isIn(mode):
          pythonPackArg.println(name+' = ('+name+' ? strcpy(malloc(strlen('+name+')+1), '+name+') : NULL);')
        iorUnpackArg.println(iorName+' = ('+iorName+' ? strcpy(malloc(strlen('+iorName+')+1), '+iorName+') : NULL);')
        pythonCleanup.println('free((void *) '+name+');')
    if self.needLocalVariable(type) or not self.isOut(mode) or isReturn:
      pythonUnpackSkel.write('&')

    if self.needLocalVariable(type):
      pythonDeclare.println(pythonType+' '+localName+';')
      iorPackArg.println(pythonType+' '+localName+';')
      if typeName == 'bool':
        if self.isIn(mode):
          pythonPackArg.println(name+' = ('+localName+' != FALSE);')
          iorPackArg.println(localName+' = ('+iorName+' != FALSE);')
        if self.isOut(mode):
          pythonUnpackArg.println(localName+' = ('+name+' != FALSE);')
          iorUnpackArg.println(iorName+' = ('+localName+' ? TRUE : FALSE);')
      elif typeName == 'fcomplex' or typeName == 'dcomplex':
        if self.isIn(mode):
          pythonPackArg.println('('+name+').real      = '+localName+'.real;')
          pythonPackArg.println('('+name+').imaginary = '+localName+'.imag;')
          iorPackArg.println(localName+'.real = ('+iorName+').real;')
          iorPackArg.println(localName+'.imag = ('+iorName+').imaginary;')
        if self.isOut(mode):
          pythonUnpackArg.println(localName+'.real = ('+name+').real;')
          pythonUnpackArg.println(localName+'.imag = ('+name+').imaginary;')
          iorUnpackArg.println('('+iorName+').real      = '+localName+'.real;')
          iorUnpackArg.println('('+iorName+').imaginary = '+localName+'.imag;')
      elif typeName == 'int' or typeName == 'long':
        if self.isIn(mode):
          pythonPackArg.println(name+' = '+localName+';')
          iorPackArg.println(localName+' = '+iorName+';')
        if self.isOut(mode):
          pythonUnpackArg.println(localName+' = '+name+';')
          iorUnpackArg.println(iorName+' = '+localName+';')
    return

  def marshallEnumeration(self, type, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn):
    typeName   = type.getIdentifier()
    iorType    = type.getCodeBlock('C', 'ior type').getSource()
    pythonType = 'long'
    if self.isOut(mode) and not isReturn:
      iorName = '*'+name
    else:
      iorName = name
    localName = '_local_'+name

    pythonDeclare.println(iorType+' '+name+';')
    pythonPackParam.write('&')
    pythonUnpackSkel.write('&')
    if self.needLocalVariable(type):
      pythonUnpackParam.write('_local_')
    pythonUnpackParam.write(name)

    if self.needLocalVariable(type):
      pythonDeclare.println(pythonType+' '+localName+';')
      iorPackArg.println(pythonType+' '+localName+';')
      if self.isIn(mode):
        pythonPackArg.println(name+' = '+localName+';')
        iorPackArg.println(localName+' = '+iorName+';')
      if self.isOut(mode):
        pythonUnpackArg.println(localName+' = '+name+';')
        iorUnpackArg.println(iorName+' = '+localName+';')
    return

  def marshallArray(self, array, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonPackSkel, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn):
    pythonType = array.getCodeBlock(self.language, 'python type').getSource()
    iorType    = array.getCodeBlock('C', 'ior type').getSource()
    if self.isOut(mode) and not isReturn:
      iorName = '*'+name
    else:
      iorName = name

    pythonDeclare.println(iorType+' '+name+' = NULL;')
    if array.getType().getBaseType():
      typeName = array.getType().getIdentifier()
      pythonPackParam.write('(void *) SIDL_'+typeName+'__borrow_python_array, ')
      pythonUnpackSkel.write('(void *) SIDL_'+typeName+'__clone_python_array, ')
      pythonPackSkel.write('(void *) SIDL_'+typeName+'__python_borrow_array, ')
      pythonUnpackParam.write('(void *) SIDL_'+typeName+'__python_clone_array, ')
      pythonCleanup.println('SIDL_'+typeName+'__array_destroy('+name+');')
    else:
      typeName = array.getType().getFullIdentifier().replace('.', '_')
      pythonPackParam.write('(void *) '+typeName+'__convert_python_array, ')
      pythonUnpackSkel.write('(void *) '+typeName+'__convert_python_array, ')
      pythonPackSkel.write('(void *) '+typeName+'__convert_sidl_array, ')
      pythonUnpackParam.write('(void *) '+typeName+'__convert_sidl_array, ')
      pythonCleanup.println(typeName+'__destroy_sidl_array('+name+');')
    if self.isInOut(mode):
      pythonPackSkel.write('*')
    pythonPackParam.write('&')
    if not self.isOut(mode) or isReturn:
      pythonUnpackSkel.write('&')
    pythonPackSkel.write(name)
    pythonUnpackParam.write(name)
    return

  def marshallInterface(self, interface, name, mode, pythonDeclare, pythonPackArg, pythonPackParam, pythonPackSkel, pythonUnpackArg, pythonUnpackParam, pythonUnpackSkel, pythonCleanup, iorPackArg, iorUnpackArg, isReturn):
    pythonType = interface.getCodeBlock(self.language, 'python type').getSource()
    iorType = interface.getCodeBlock('C', 'ior type').getSource()
    if self.isOut(mode) and not isReturn:
      iorName = '*'+name
    else:
      iorName = name
    localName = '_local_'+name

    pythonDeclare.println(iorType+' '+name+' = NULL;')
    pythonPackParam.write('(void *) '+interface.getFullIdentifier().replace('.', '_')+'__convert, ')
    pythonUnpackParam.write('(void *) '+interface.getFullIdentifier().replace('.', '_')+'__wrap, ')
    pythonPackSkel.write('(void *) '+interface.getFullIdentifier().replace('.', '_')+'__newRef, ')
    pythonUnpackSkel.write('(void *) '+interface.getFullIdentifier().replace('.', '_')+'__convert, ')
    if not self.isOut(mode) or isReturn:
      pythonUnpackSkel.write('&')
    if self.isOut(mode):
      pythonPackSkel.write('*')
    pythonPackParam.write('&')
    pythonPackSkel.write(name)
    pythonUnpackParam.write(name)
    if not self.isOut(mode):
      pythonCleanup.println(interface.getFullIdentifier().replace('.', '_')+'__deref('+name+');')
    return

  ##################################
  # Package Generation Operations
  def generatePackage(self, package, purpose):
    code = self.getCode(purpose.getName()+' contents')
    code.println('all = [')
    code.increaseTabLevel()
    children = package.getChildren()
    for i in range(len(children)):
      if i < len(children)-1:
        code.println('"'+children[i].getIdentifier()+'",')
      else:
        code.println('"'+children[i].getIdentifier()+'"')
    code.decreaseTabLevel()
    code.println(']')
    package.addCode(code)
    return

  #################################
  # Interface Generation Operations
  def generateInterfaceType(self, interface):
    try:
      code = self.getCode('python type')
      code.write(interface.getFullIdentifier())
      interface.addCode(code)
      self.marshallReturnType(interface, interface)
    except SIDLLanguage.CodeGeneratorException.Exception, e:
      self.raiseVisitorException('In '+interface.getIdentifier()+': '+e.getMessage(), e.getStackTrace().strip())
    return

  def generateInterfaceStub(self, interface, purpose):
    self.generateInterfaceDefines(interface, purpose)
    self.generateStubIncludes(interface, purpose)
    self.generateEPV(interface, purpose)

    self.generateHeaderBegin(interface, purpose)
    self.generateExternBegin(interface, purpose)
    self.generateStubForwardDeclarations(interface, purpose)
    self.generateInternalDeclarations(interface, purpose)
    self.generateExternEnd(interface, purpose)
    self.generateHeaderEnd(interface, purpose)

    castMethod = self.createCastMethod(interface)
    self.generateCast(interface, castMethod, purpose)
    self.generateMethods(interface, purpose)
    self.generateStaticVTable(interface, castMethod, purpose)
    self.generateVTable(interface, purpose)
    self.generateWrapper(interface, purpose)
    self.generateWeakReference(interface, purpose)
    self.generateDeleteReference(interface, purpose)
    self.generateNewReference(interface, purpose)
    self.generateAddReference(interface, purpose)
    self.generateConvert(interface, purpose)
    self.generateConvertToElement(interface, purpose)
    self.generateConvertToArray(interface, purpose)
    self.generateConvertFromElement(interface, purpose)
    self.generateConvertFromArray(interface, purpose)
    self.generateDestroyArray(interface, purpose)
    self.generateInitModule(interface, purpose)
    return

  def generateClassSkel(self, sclass, purpose):
    self.generateSkelIncludes(sclass, purpose)
    self.generateMethods(sclass, purpose)
    return

  def generateClassImpl(self, sclass, purpose):
    code = self.getCode(purpose.getName()+' class begin')
    code.println('class '+sclass.getIdentifier()+':')
    code.increaseTabLevel()
    self.commentsToDocString(code, sclass)
    code.println('# All calls to SIDL methods should use __IORself')
    code.println('')
    sclass.addCode(code)

    self.generateMethodsImpl(sclass, purpose)
    return

  def generateHeaderBegin(self, interface, purpose):
    code  = self.getCode(purpose.getName()+' begin header')
    guard = 'included_'+interface.getFullIdentifier().replace('.', '_')+'_'+purpose.getName()
    code.write('#ifndef '+guard+'\n')
    code.write('#define '+guard+'\n')
    code.println('')
    interface.addCode(code)
    return

  def generateHeaderEnd(self, interface, purpose):
    code  = self.getCode(purpose.getName()+' end header')
    guard = 'included_'+interface.getFullIdentifier().replace('.', '_')+'_'+purpose.getName()
    code.write('#endif /* '+guard+' */\n')
    interface.addCode(code)
    return

  def generateInterfaceDefines(self, interface, purpose):
    code = self.getCode(purpose.getName()+' header shared defines')
    name = interface.getFullIdentifier().replace('.', '_')
    i    = 0
    func = name+'__wrap'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN PyObject *\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj)\n')
    code.println('')
    i   += 1
    func = name+'__convert'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN int\n')
    code.write('#define '+func+'_PROTO  (PyObject *obj, ')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('*sidlobj)\n')
    code.println('')
    i += 1
    func = name+'__convert_python_array'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN int\n')
    code.write('#define '+func+'_PROTO  (PyObject *obj, ')
    self.composeCode(code, interface, 'ior array type', language = 'C')
    code.write('*sidlarray)\n')
    code.println('')
    i += 1
    func = name+'__convert_sidl_array'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN PyObject *\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior array type', language = 'C')
    code.write('sidlarray)\n')
    code.println('')
    i += 1
    func = name+'__destroy_sidl_array'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN void\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior array type', language = 'C')
    code.write('sidlarray)\n')
    code.println('')
    i += 1
    func = name+'__weakRef'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN PyObject *\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj)\n')
    code.println('')
    i += 1
    func = name+'__deref'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN void\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj)\n')
    code.println('')
    i += 1
    func = name+'__newRef'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN PyObject *\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj)\n')
    code.println('')
    i += 1
    func = name+'__addRef'
    code.write('#define '+func+'_NUM    '+str(i)+'\n')
    code.write('#define '+func+'_RETURN void\n')
    code.write('#define '+func+'_PROTO  (')
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj)\n')
    code.println('')
    i += 1
    if self.isException(interface):
      code.write('#define '+name+'__type_NUM '+str(i)+'\n')
      code.println('')
      i += 1
    code.write('#define '+name+'__API_NUM '+str(i)+'\n')
    code.println('')
    interface.addCode(code)

    api  = name+'__API'
    code = self.getCode(purpose.getName()+' header client defines')
    code.write('#ifndef '+name+'_INTERNAL\n')
    code.println('static void **'+api+';')
    for ext in ['wrap', 'convert', 'convert_python_array', 'convert_sidl_array', 'destroy_sidl_array', 'weakRef', 'deref', 'newRef', 'addRef']:
      func = name+'__'+ext
      code.write('#define '+func+' (*(('+func+'_RETURN (*)'+func+'_PROTO)('+api+'['+func+'_NUM])))\n')
    if self.isException(interface):
      code.write('#define '+name+'__type ((PyObject *)('+api+'['+name+'__type_NUM]))\n')
    code.println('')
    code.write('#define '+name+'__import() \\\n')
    code.println('{ \\')
    code.increaseTabLevel()
    code.println('PyObject *module = PyImport_ImportModule((char *) "'+interface.getFullIdentifier()+'"); \\')
    code.println('if (module != NULL) { \\')
    code.increaseTabLevel()
    code.println('PyObject *module_dict  = PyModule_GetDict(module); \\')
    code.println('PyObject *c_api_object = PyDict_GetItemString(module_dict, (char *) "_C_API"); \\')
    code.println('if ((c_api_object != NULL) && PyCObject_Check(c_api_object)) { \\')
    code.increaseTabLevel()
    code.println(api+' = (void **) PyCObject_AsVoidPtr(c_api_object); \\')
    code.decreaseTabLevel()
    code.println('} \\')
    code.println('Py_XDECREF(module); \\')
    code.decreaseTabLevel()
    code.println('} \\')
    code.decreaseTabLevel()
    code.println('}')
    code.write('#endif /* not '+name+'_INTERNAL */\n')
    code.println('')
    interface.addCode(code)

    code = self.getCode(purpose.getName()+' source defines')
    code.write('#define '+name+'_INTERNAL 1\n')
    code.println('')
    interface.addCode(code)
    return

  def generateStubIncludes(self, interface, purpose):
    code = self.getCode(purpose.getName()+' header includes')
    code.write('#include "sidl_config.h"\n')
    code.write('#include <Python.h>\n')
    code.println('')
    interface.addCode(code)

    code = self.getCode(purpose.getName()+' source includes')
    code.write('#include "SIDLObjA.h"\n')
    code.write('#include "SIDLPyArrays.h"\n')
    code.write('#include "Numeric/arrayobject.h"\n')
    code.println('')
    interface.addCode(code)
    return

  def generateSkelIncludes(self, sclass, purpose):
    code = self.getCode(purpose.getName()+' source includes')
    code.write('#include "SIDLObjA.h"\n')
    code.write('#include "SIDLPyArrays.h"\n')
    code.write('#define  NO_IMPORT_ARRAY\n')
    code.write('#include "Numeric/arrayobject.h"\n')
    code.println('')
    code.println('/* The SIDL Runtime function which spawns the interpreter */')
    code.println('void SIDL_Python_Init(void);')
    code.println('')
    sclass.addCode(code)
    return

  def generateExternBegin(self, interface, purpose):
    code = self.getCode(purpose.getName()+' begin extern')
    code.write('#ifdef __cplusplus\n')
    code.println('extern "C" {')
    code.write('#endif\n')
    code.println('')
    interface.addCode(code)
    return

  def generateStubForwardDeclarations(self, interface, purpose):
    code = self.getCode(purpose.getName()+' forward declarations')
    code.println('/* Forward declaration of IOR structure */')
    code.indent()
    self.composeCode(code, interface, 'ior type struct', language = 'C')
    code.write(';\n')
    code.indent()
    self.composeCode(code, interface, 'ior array struct', language = 'C')
    code.write(';\n')
    code.println('')
    interface.addCode(code)
    return

  def generateInternalDeclarations(self, interface, purpose):
    name = interface.getFullIdentifier().replace('.', '_')
    code = self.getCode(purpose.getName()+' internal declarations')
    code.write('#ifdef '+name+'_INTERNAL\n')
    for ext in ['wrap', 'convert', 'convert_python_array', 'convert_sidl_array', 'destroy_sidl_array', 'weakRef', 'deref', 'newRef', 'addRef']:
      func = name+'__'+ext
      code.println('static '+func+'_RETURN '+func+' '+func+'_PROTO;')
    if self.isException(interface):
      code.println('static PyObject *'+name+'__type;')
    code.println('void init'+interface.getIdentifier()+'(void);')
    code.write('#endif /* '+name+'_INTERNAL */\n')
    code.println('')
    interface.addCode(code)
    return

  def generateExternEnd(self, interface, purpose):
    code = self.getCode(purpose.getName()+' end extern')
    code.write('#ifdef __cplusplus\n')
    code.println('}')
    code.write('#endif\n')
    code.println('')
    interface.addCode(code)
    return

  def generateEPV(self, interface, purpose):
    code = self.getCode(purpose.getName()+' epv definition')
    if self.hasStaticMethod(interface):
      code.indent()
      code.write('static const ')
      self.composeCode(code, interface, 'ior static function table type', language = 'C')
      code.write(' _sepv = NULL;\n')
    code.indent()
    code.write('static const ')
    self.composeCode(code, interface, 'ior function table type', language = 'C')
    code.write(' _epv = NULL;\n')
    code.println('')
    interface.addCode(code)
    return

  ##############################
  # Method Generation Operations
  def generateMethodReturnType(self, method, purpose):
    defn = self.getCode(purpose.getName()+' definition return type')
    defn.indent()
    defn.write('static ')
    if purpose.getIORDispatch():
      defn.write('PyObject *\n')
    else:
      self.composeCode(defn, method.getReturnType(), 'ior type', forceBreak = 1, language = 'C')
    method.addCode(defn)
    return

  def generateMethodBegin(self, method, purpose):
    code = self.getCode(purpose.getName()+' definition begin')
    code.println('{')
    code.increaseTabLevel()
    method.addCode(code)
    return

  def generateMethodIORCheck(self, method, purpose):
    if not purpose.getGenerateMethods() or not purpose.getCheckIOR() or method.getStatic(): return
    code = self.getCode(purpose.getName()+' definition IOR check')

    code.println('if (_self_ior == 0) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_TypeError, "Null IOR pointer in '+method.getIdentifier()+'()");')
    code.decreaseTabLevel()
    code.println('}')
    method.addCode(code)
    return

  def generateMethodReturnSetup(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition return setup')

    code.println('/* Return value setup */')
    type = method.getReturnType()
    if purpose.getIORDispatch():
      code.println('PyObject *_return_value = NULL;')
      self.composeCode(code, method.getReturnType(), 'python argument declare')
    else:
      code.println('PyObject *_context;')
      code.println('PyObject *_pfunc;')
      code.println('PyObject *_args;')
      code.println('PyObject *_return_value;')
      code.println('int       _okay = 0;')
      if not type.getIdentifier() == 'void':
        typeName = type.getIdentifier()
        iorType  = type.getCodeBlock('C', 'ior type').getSource()

        if type.getBaseType() and self.iorDefaultValue[typeName]:
          code.println(iorType+' _result = '+self.iorDefaultValue[typeName]+';')
        else:
          code.println(iorType+' _result;')
      self.composeCode(code, method.getReturnType(), 'ior argument packing')
    if len(self.getIORExceptions(method)):
      if not purpose.getEmulateExceptions():
        code.indent()
        self.composeCode(code, self.getBaseException(self.getIORExceptions(method)[0]), 'ior type', language = 'C')
        code.write('_exception = 0;\n')
    method.addCode(code)
    return

  def generateMethodArgumentPacking(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code         = self.getCode(purpose.getName()+' definition argument packing')
    parameters   = method.getChildren()
    inParameters = []

    if purpose.getIORDispatch():
      code.println('/* Pack arguments for dispatch to IOR */')
      for parameter in parameters:
        parameter = SIDLLanguage.Parameter.Parameter(parameter)
        self.composeCode(code, parameter, 'python argument declare')
        if self.isIn(parameter.getAccessMode()):
          inParameters.append(parameter)

      code.println('static char *_kwlist[] = {')
      code.increaseTabLevel()
      for parameter in inParameters:
        code.println('(char *) "'+parameter.getIdentifier()+'",')
      code.println('NULL')
      code.decreaseTabLevel()
      code.println('};')
      code.println('const int _okay = PyArg_ParseTupleAndKeywords(')
      code.increaseTabLevel()
      code.println('_args, _kwdict,')
      code.indent()
      code.write('(char *) "')
      optionalArgs = 0
      for parameter in inParameters:
        if not optionalArgs and not parameter.getRequired():
          code.write('|')
          optionalArgs = 1
        self.composeCode(code, parameter, 'python argument packing format')
      code.write('", _kwlist')
      if len(inParameters): code.write(',')
      code.write('\n')
      for i in range(len(inParameters)):
        code.indent()
        self.composeCode(code, inParameters[i], 'python argument packing parameter')
        if i < len(inParameters)-1: code.write(',')
        code.write('\n')
      code.decreaseTabLevel()
      code.write(');\n')
      code.println('if (_okay == 0) {')
      code.increaseTabLevel()
      code.println('PyErr_SetString(PyExc_TypeError, "Could not parse arguments for '+method.getIdentifier()+'");')
      code.println('return _return_value;');
      code.decreaseTabLevel()
      code.println('}')
      for parameter in parameters:
        self.composeCode(code, parameter, 'python argument packing')
    else:
      code.println('/* Pack arguments for dispatch to implementation */')
      for parameter in method.getChildren():
        self.composeCode(code, parameter, 'ior argument packing')
      inParameters = []
      for parameter in method.getChildren():
        parameter = SIDLLanguage.Parameter.Parameter(parameter)
        if self.isIn(parameter.getAccessMode()):
          inParameters.append(parameter)

      code.println('_args = Py_BuildValue(')
      code.increaseTabLevel()
      code.indent()
      code.write(' (char *) "(')
      for parameter in inParameters:
        self.composeCode(code, parameter, 'python argument unpacking format')
      code.write(')"')
      if len(inParameters) > 0: code.write(',')
      code.write('\n')
      for i in range(len(inParameters)):
        code.indent()
        if inParameters[i].getType().getBaseType() or inParameters[i].getType().isInstanceOf('SIDLLanguage.Enumeration'):
          if not self.needLocalVariable(inParameters[i].getType()) and self.isInOut(inParameters[i].getAccessMode()):
            code.write('*')
          self.composeCode(code, inParameters[i], 'python argument unpacking parameter')
        else:
          self.composeCode(code, inParameters[i], 'python argument packing skeleton')
        if i < len(inParameters)-1: code.write(',')
        code.write('\n')
      code.decreaseTabLevel()
      code.write(');\n')
    code.println('if (!_args) {')
    code.increaseTabLevel()
    code.println('if (PyErr_Occurred()) PyErr_Print();')
    self.composeCode(code, method, purpose.getName()+' definition return')
    code.decreaseTabLevel()
    code.println('}')
    method.addCode(code)
    return

  def generateMethodExceptionSetup(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition exception setup')

    code.println('/* Exception handling setup */')
    if len(self.getIORExceptions(method)):
      if purpose.getEmulateExceptions():
        code.println('*_exception = 0;')
    method.addCode(code)
    return

  def generateMethodIORDispatchBegin(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition IOR dispatch begin')

    code.println('/* Dispatch to IOR */')
    code.indent()
    if not method.getReturnType().getIdentifier() == 'void':
      if self.needLocalVariable(method.getReturnType()):
        code.write('_result = ')
      else:
        code.write('_result = ')
    if method.getStatic():
      code.write('(_sepv->f_'+method.getIdentifier()+')(')
    else:
      code.write('(*(_self_ior->d_epv->f_'+method.getIdentifier()+'))(')
    method.addCode(code)
    return

  def generateMethodIORDispatchEnd(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code       = self.getCode(purpose.getName()+' definition IOR dispatch end')
    parameters = method.getChildren()
    exceptions = self.getIORExceptions(method)

    if not method.getStatic() and len(parameters) > 0:
      code.write(', ')
    for p in range(len(parameters)):
      param = SIDLLanguage.Parameter.Parameter(parameters[p])
      name = param.getIdentifier()
      mode = param.getAccessMode()
      type = param.getType()

      code.write(self.getModeComment(mode))
      if self.isOut(mode):
        code.write('&')
      if self.needLocalVariable(type) and not self.isOut(mode) and self.hasIOR(type):
        code.write(name+'._get_ior()')
      else:
        code.write(name)
      if (p < len(parameters)-1): code.write(', ')
    if len(exceptions):
      if not method.getStatic() or len(parameters) > 0:
        code.write(', ')
      code.write('&_exception')
    code.write(');\n')

    method.addCode(code)
    return

  def generateMethodImplDispatchBegin(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition impl dispatch begin')

    code.println('/* Dispatch to implementation */')
    method.addCode(code)
    return

  def generateMethodImplDispatchEnd(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition impl dispatch end')

    code.println('if (!_context) {')
    code.increaseTabLevel()
    if method.getStatic():
      code.println('if (PyErr_Occurred()) PyErr_Print();')
    else:
      code.println('fprintf(stderr, "Null implementation object for '+method.getFullIdentifier()+'");')
    self.composeCode(code, method, purpose.getName()+' definition return')
    code.decreaseTabLevel()
    code.println('}')
    code.println('_pfunc = PyObject_GetAttrString(_context, (char *) "'+method.getIdentifier()+'");')
    code.println('if (!_pfunc) {')
    code.increaseTabLevel()
    code.println('if (PyErr_Occurred()) PyErr_Print();')
    if method.getStatic():
      code.println('Py_DECREF(_context);')
    self.composeCode(code, method, purpose.getName()+' definition return')
    code.decreaseTabLevel()
    code.println('}')
    code.println('_return_value = PyObject_CallObject(_pfunc, _args);')
    code.println('Py_DECREF(_pfunc);')
    if method.getStatic():
      code.println('Py_DECREF(_context);')
    method.addCode(code)
    return

  def generateMethodExceptionCleanup(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code       = self.getCode(purpose.getName()+' definition exception cleanup')
    exceptions = self.getIORExceptions(method)

    code.println('/* Exception handling cleanup */')
    method.addCode(code)
    if not len(exceptions):
      if purpose.getEmulateExceptions():
        code.println('if (PyErr_Occurred()) PyErr_Print();')
      return

    if purpose.getEmulateExceptions():
      code.println('if (PyErr_Occurred()) {')
      code.increaseTabLevel()
      code.println('PyObject *_exType;')
      code.println('PyObject *_exValue;')
      code.println('PyObject *_exTrace;')
      code.println('')
      code.println('PyErr_Fetch(&_exType, &_exValue, &_exTrace);')
      for i in range(len(exceptions)):
        exception = exceptions[i]
        iorName   = self.getIORName(exception)

        code.indent()
        if i > 0: code.write('} else ')
        code.write('if (PyErr_GivenExceptionMatches(_exType, '+iorName+'__type)) {\n')
        code.increaseTabLevel()
        code.indent()
        self.composeCode(code, exception, 'ior type', language = 'C')
        code.write('_local_exception;\n')
        code.println('')
        code.println('(void) '+iorName+'__convert(_exValue, &_local_exception);')
        code.println('if (_local_exception) {')
        code.increaseTabLevel()
        code.println(iorName+'__addRef(_local_exception);')
        code.indent()
        code.write('*_exception = (')
        self.composeCode(code, self.getBaseException(exception), 'ior type', language = 'C')
        code.write(') _local_exception;\n')
        code.decreaseTabLevel()
        code.println('}')
        code.decreaseTabLevel()
      code.println('} else {')
      code.increaseTabLevel()
      exception = self.getBaseException(exceptions[0])
      iorName   = self.getIORName(exception)
      code.println('PyObject *module         = PyImport_ImportModule((char *) "SIDL.BaseException");')
      code.println('PyObject *modDict        = PyModule_GetDict(module);')
      code.println('PyObject *msg            = PyObject_Str(_exValue);')
      code.println('char     *message        = PyString_AsString(msg);')
      code.println('PyObject *_new_exception = PyObject_CallFunction(PyDict_GetItemString(modDict, (char *) "BaseException"), (char *) "");')
      code.indent()
      self.composeCode(code, exception, 'ior type', language = 'C')
      code.write('_local_exception;\n')
      code.println('')
      code.println('PyObject_CallMethod(_new_exception, (char *) "setMessage", (char *) "s", message);')
      code.println('Py_XDECREF(msg);')
      code.println('(void) '+iorName+'__convert(_new_exception, &_local_exception);')
      code.println('if (_local_exception) {')
      code.increaseTabLevel()
      code.println(iorName+'__addRef(_local_exception);')
      code.indent()
      code.write('*_exception = (')
      self.composeCode(code, exception, 'ior type', language = 'C')
      code.write(') _local_exception;\n')
      code.decreaseTabLevel()
      code.println('}')
      code.println('Py_XDECREF(_new_exception);')
      code.println('Py_XDECREF(module);')
      code.decreaseTabLevel()
      code.println('}')
      code.println('Py_XDECREF(_exType);')
      code.println('Py_XDECREF(_exValue);')
      code.println('Py_XDECREF(_exTrace);')
      code.println('PyErr_Clear();')
      code.decreaseTabLevel()
      code.println('}')
    else:
      code.println('if (_exception != 0) {')
      code.increaseTabLevel()
      code.println('int found = 0;')
      for i in range(len(exceptions)):
        exception     = exceptions[i]
        exceptionNum  = '_ex'+str(i)
        exceptionName = exception.getFullIdentifier()
        iorType       = exception.getCodeBlock('C', 'ior type').getSource()

        code.println(iorType+exceptionNum+';')
      code.println('')
      for i in range(len(exceptions)):
        exception     = exceptions[i]
        exceptionNum  = '_ex'+str(i)
        exceptionName = exception.getFullIdentifier()
        iorType       = exception.getCodeBlock('C', 'ior type').getSource()

        code.println('if (!found && ('+exceptionNum+' = ('+iorType+') SIDL_PyExceptionCast(_exception, "'+exceptionName+'"))) {')
        code.increaseTabLevel()
        code.println('PyObject *obj = '+self.getIORName(exception)+'__wrap('+exceptionNum+');')
        code.println('')
        code.println('PyErr_SetObject('+self.getIORName(exception)+'__type, obj);')
        code.println('Py_XDECREF(obj);')
        code.println('found = 1;')
        code.decreaseTabLevel()
        code.println('}')
      code.println('if (!found) {')
      code.increaseTabLevel()
      code.println('PyObject *module  = PyImport_ImportModule((char *) "SIDL.BaseException");')
      code.println('PyObject *modDict = PyModule_GetDict(module);')
      code.println('PyObject *obj     = PyObject_CallFunction(PyDict_GetItemString(modDict, (char *) "BaseException"), (char *) "");')
      code.println('')
      code.println('PyObject_CallMethod(obj, (char *) "setMessage", (char *) "s", (char *) "Invalid exception thrown by implementation");')
      code.println('PyErr_SetObject(SIDL_BaseException__type, obj);')
      code.println('Py_XDECREF(module);')
      code.decreaseTabLevel()
      code.println('}')
      code.println('return _return_value;')
      code.decreaseTabLevel()
      code.println('}')
    return

  def generateMethodArgumentUnpacking(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition argument unpacking')

    parameters = method.getChildren()
    if purpose.getIORDispatch():
      code.println('/* Unpack arguments from dispatch to IOR */')
      for parameter in parameters:
        self.composeCode(code, parameter, 'python argument unpacking')
    else:
      returnType = method.getReturnType()
      hasReturn  = not (returnType.getBaseType() and returnType.getIdentifier() == 'void')
      outParameters = []
      for parameter in method.getChildren():
        parameter = SIDLLanguage.Parameter.Parameter(parameter)
        if self.isOut(parameter.getAccessMode()):
          outParameters.append(parameter)

      code.println('/* Unpack arguments from dispatch to implementation */')
      if hasReturn or len(outParameters):
        code.println('if (_return_value) {')
        code.increaseTabLevel()
        if hasReturn + len(outParameters) == 1:
          code.println('PyObject *_tmp = PyTuple_New(1);')
          code.println('')
          code.println('PyTuple_SET_ITEM(_tmp, 0, _return_value);')
          code.println('_return_value = _tmp;')
        code.println('_okay   = PyArg_ParseTuple(_return_value,')
        code.increaseTabLevel()
        code.indent()
        code.write('(char *) "')
        self.composeCode(code, returnType, 'python argument packing format')
        for parameter in outParameters:
          self.composeCode(code, parameter, 'python argument packing format')
        code.write('"')
        if hasReturn or len(outParameters): code.write(',')
        code.write('\n')
        if hasReturn:
          code.indent()
          self.composeCode(code, returnType, 'python argument unpacking skeleton')
          if len(outParameters): code.write(',')
          code.write('\n')
        for i in range(len(outParameters)):
          code.indent()
          self.composeCode(code, outParameters[i], 'python argument unpacking skeleton')
          if i < len(outParameters)-1: code.write(',')
          code.write('\n')
        code.decreaseTabLevel()
        code.println(');')
        code.decreaseTabLevel()
        code.println('}')
      else:
        code.println('_okay = (_return_value == Py_None);')
        code.println('if (!_okay) {')
        code.increaseTabLevel()
        code.println('printf("Invalid return from void function '+method.getIdentifier()+'\\n");')
        code.decreaseTabLevel()
        code.println('}')
      code.println('Py_DECREF(_args);')
      code.println('if (PyErr_Occurred()) {')
      code.println('PyErr_Print();')
      code.println('}')
      for parameter in outParameters:
        self.composeCode(code, parameter, 'ior argument unpacking')
    method.addCode(code)
    return

  def generateMethodReturnCleanup(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code       = self.getCode(purpose.getName()+' definition return cleanup')
    returnType = method.getReturnType()

    code.println('/* Return value cleanup */')
    if purpose.getIORDispatch():
      hasReturn     = not (returnType.getBaseType() and returnType.getIdentifier() == 'void')
      outParameters = []
      for parameter in method.getChildren():
        parameter = SIDLLanguage.Parameter.Parameter(parameter)
        if self.isOut(parameter.getAccessMode()):
          outParameters.append(parameter)

      self.composeCode(code, returnType, 'python argument unpacking')
      if hasReturn or len(outParameters):
        code.println('_return_value = Py_BuildValue(')
        code.increaseTabLevel()
        code.indent()
        code.write(' (char *) "')
        self.composeCode(code, returnType, 'python argument unpacking format')
        for parameter in outParameters:
          self.composeCode(code, parameter, 'python argument unpacking format')
        code.write('",\n')
        if hasReturn:
          code.indent()
          self.composeCode(code, returnType, 'python argument unpacking parameter')
          if len(outParameters): code.write(',')
          code.write('\n')
        for i in range(len(outParameters)):
          code.indent()
          self.composeCode(code, outParameters[i], 'python argument unpacking parameter')
          if i < len(outParameters)-1: code.write(',')
          code.write('\n')
        code.decreaseTabLevel()
        code.write(');\n')
      else:
        code.println('_return_value = Py_None;')
        code.println('Py_INCREF(_return_value);')
      for parameter in method.getChildren():
        self.composeCode(code, parameter, 'python argument cleanup')
      self.composeCode(code, returnType, 'python argument cleanup')
    else:
      if not returnType.getIdentifier() == 'void':
        self.composeCode(code, returnType, 'ior argument unpacking')
      code.println('if (_return_value && _return_value != Py_None) {Py_DECREF(_return_value);}')
    method.addCode(code)
    return

  def generateMethodReturn(self, method, purpose):
    if not purpose.getGenerateMethods(): return
    code = self.getCode(purpose.getName()+' definition return')

    if purpose.getIORDispatch():
      code.println('return _return_value;')
    else:
      if not method.getReturnType().getIdentifier() == 'void':
        code.println('return _result;')
      else:
        code.println('return;')
    method.addCode(code)
    return

  def generateMethodEnd(self, method, purpose):
    code = self.getCode(purpose.getName()+' definition end')
    code.println('}')
    method.addCode(code)
    return

  def generateMethodComponents(self, method, purpose, isDeclaration = 1):
    self.generateMethodReturn(method, purpose)

    self.generateMethodReturnType(method, purpose)
    self.generateMethodBegin(method, purpose)
    self.generateMethodIORCheck(method, purpose)
    self.generateMethodReturnSetup(method, purpose)
    self.generateMethodArgumentPacking(method, purpose)
    self.generateMethodExceptionSetup(method, purpose)
    if purpose.getIORDispatch():
      self.generateMethodIORDispatchBegin(method, purpose)
      self.generateMethodIORDispatchEnd(method, purpose)
    else:
      self.generateMethodImplDispatchBegin(method, purpose)
      self.generateMethodImplDispatchEnd(method, purpose)
    self.generateMethodExceptionCleanup(method, purpose)
    self.generateMethodArgumentUnpacking(method, purpose)
    self.generateMethodReturnCleanup(method, purpose)
    self.generateMethodEnd(method, purpose)
    return

  def getMethodName(self, interface, method, purpose):
    pathExt = purpose.getPathExtension()
    if purpose.getIORDispatch():
      # C naming
      methodName = interface.getFullIdentifier().replace('.', '_')+'_'+method.getIdentifier()
      if pathExt: methodName = methodName+'_'+pathExt
    else:
      # C naming
      methodName = interface.getFullIdentifier().replace('.', '_')+'_'+method.getIdentifier()
      if pathExt: methodName = methodName+'_'+pathExt
    return methodName

  def generateMethodParameters(self, interface, method, purpose, code):
    parameters    = method.getChildren()
    hasExceptions = purpose.getEmulateExceptions() and len(self.getIORExceptions(method))
    hasArgs       = (purpose.getUseSelf() and not method.getStatic()) or len(parameters) or hasExceptions
    if hasArgs or purpose.getIORDispatch():
      code.println('(')
      code.increaseTabLevel()
      if purpose.getUseSelf() and not method.getStatic():
        code.indent()
        self.composeCode(code, interface, 'ior type', language = 'C')
        code.write('self')
        if len(parameters) or hasExceptions: code.write(',\n')
      if purpose.getIORDispatch():
        if method.getStatic():
          code.println('PyObject *_ignored,')
        else:
          code.println('PyObject *_self,')
        code.println('PyObject *_args,')
        code.println('PyObject *_kwdict')
      else:
        for p in range(len(parameters)):
          code.indent()
          self.composeCode(code, parameters[p], 'ior parameter', language = 'C')
          if p < len(parameters)-1 or hasExceptions: code.write(',\n')
        if purpose.getEmulateExceptions():
          self.generateMethodExceptions(interface, method, purpose, code)
      code.decreaseTabLevel()
      code.println(')')
    else:
      code.println('(void)')
    return

  def generateMethodExceptions(self, interface, method, purpose, code):
    if purpose.getEmulateExceptions():
      exceptions = self.getIORExceptions(method)
      if len(exceptions):
        code.indent()
        self.composeCode(code, self.getBaseException(exceptions[0]), 'ior type', language = 'C')
        code.write('*_exception')
    return

  def generateMethodIOR(self, code, interface, method, purpose):
    if not purpose.getGenerateMethods() or method.getStatic(): return

    if purpose.getIORDispatch():
      code.indent()
      self.composeCode(code, interface, 'ior type', language = 'C')
      code.write('_self_ior = ((')
      self.composeCode(code, interface, 'ior type', language = 'C')
      code.write(') SIDL_Cast(_self, "'+interface.getFullIdentifier()+'"));\n')
    return

  def generateMethod(self, interface, method, purpose, use):
    if purpose.getEmbedded(): self.raiseVisitorException('Python methods cannot be embedded')
    pre        = purpose.getName()+' '+use
    methodName = self.getMethodName(interface, method, purpose)
    if use == 'declaration':
      code      = self.getCode(purpose.getName()+' method '+use+' '+method.getIdentifier())
    else:
      codeBegin = self.getCode(purpose.getName()+' method '+use+' begin '+method.getIdentifier())
      codeEnd   = self.getCode(purpose.getName()+' method '+use+' end '+method.getIdentifier())
      code      = codeBegin

    # Create method signature
    if not method.hasCodeBlock(self.language, pre+' return type'): return
    self.composeCode(code, method, pre+' return type')
    code.write(methodName)
    self.generateMethodParameters(interface, method, purpose, code)
    if not purpose.getEmulateExceptions():
      self.generateMethodExceptions(interface, method, purpose, code)
    if use == 'declaration':
      code.println(';')

    # Create method body
    self.composeCode(code, method, pre+' begin', 0)
    code.increaseTabLevel()
    if use == 'definition':
      interface.addCode(code)
      code = codeEnd
      code.increaseTabLevel()
    self.generateMethodIOR(code, interface, method, purpose)
    self.composeCode(code, method, pre+' return setup', 0)
    self.composeCode(code, method, pre+' argument packing', 0)
    self.composeCode(code, method, pre+' exception setup', 0)
    self.composeCode(code, method, pre+' IOR check', 0)
    self.composeCode(code, method, pre+' IOR dispatch begin', 0)
    if method.hasCodeBlock(self.language, pre+' IOR dispatch begin'):
      if not method.getStatic():
        code.write(self.getIORStructName(interface, '_self_ior'))
    self.composeCode(code, method, pre+' IOR dispatch end', 0)
    self.composeCode(code, method, pre+' impl dispatch begin', 0)
    if method.hasCodeBlock(self.language, pre+' impl dispatch begin'):
      if method.getStatic():
        code.println('_context = PyImport_ImportModule((char *) "'+interface.getParent().getFullIdentifier()+'_impl'+'.'+interface.getIdentifier()+'_impl");')
      else:
        code.println('_context = (self ? self->d_data : NULL);')
      self.composeCode(code, method, pre+' impl dispatch end', 0)
    self.composeCode(code, method, pre+' exception cleanup', 0)
    self.composeCode(code, method, pre+' argument unpacking', 0)
    self.composeCode(code, method, pre+' return cleanup', 0)
    self.composeCode(code, method, pre+' return', 0)
    code.decreaseTabLevel()
    self.composeCode(code, method, pre+' end', 0)
    code.println('')
    interface.addCode(code)
    return

  def generateMethodImpl(self, sclass, method, purpose, use):
    if purpose.getEmbedded(): self.raiseVisitorException('Python methods cannot be embedded')
    pre = purpose.getName()+' '+use
    returnType = method.getReturnType()
    hasReturn  = not (returnType.getBaseType() and returnType.getIdentifier() == 'void')
    inParams   = []
    outParams  = []
    for child in method.getChildren():
      parameter = SIDLLanguage.Parameter.Parameter(child)
      if self.isIn(parameter.getAccessMode()):
        inParams.append(parameter)
      if self.isOut(parameter.getAccessMode()):
        outParams.append(parameter)

    code = self.getCode(purpose.getName()+' method '+use+' begin '+method.getIdentifier())
    code.indent()
    code.write('def '+method.getIdentifier()+'(')
    if not method.getStatic():
      code.write('self')
      if len(inParams): code.write(', ')
    for i in range(len(inParams)):
      code.write(inParams[i].getIdentifier())
      if i < len(inParams)-1: code.write(', ')
    code.write('):\n')
    code.increaseTabLevel()
    self.commentsToDocString(code, method)
    if len(inParams):
      code.println('# SIDL EXPECTED INCOMING TYPES')
      code.println('# ============================')
      for parameter in inParams:
        code.println('# '+parameter.getIdentifier()+' '+parameter.getType().getFullIdentifier())
    if hasReturn or len(outParams):
      code.println('# SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)')
      code.println('# ================================================================')
      if hasReturn:
        code.println('# '+returnType.getFullIdentifier()+' _return')
      for parameter in outParams:
        code.println('# '+parameter.getIdentifier()+' '+parameter.getType().getFullIdentifier())
    sclass.addCode(code)
    return

  def createConstructorMethod(self, sclass):
    method = SIDLLanguage.Method.Method(SIDL.Loader.createClass('ANL.SIDLCompilerI.Method'))
    void   = SIDLLanguage.Type.Type(SIDL.Loader.createClass('ANL.SIDLCompilerI.Type'))
    void.setBaseType(1)
    void.setIdentifier('void')
    method.setReturnType(void)
    method.setIdentifier('_ctor')
    method.setAbstract(0)
    method.setStatic(0)
    method.setParent(sclass)
    return method

  def createDestructorMethod(self, sclass):
    method = SIDLLanguage.Method.Method(SIDL.Loader.createClass('ANL.SIDLCompilerI.Method'))
    void   = SIDLLanguage.Type.Type(SIDL.Loader.createClass('ANL.SIDLCompilerI.Type'))
    void.setBaseType(1)
    void.setIdentifier('void')
    method.setReturnType(void)
    method.setIdentifier('_dtor')
    method.setAbstract(0)
    method.setStatic(0)
    method.setParent(sclass)
    return method

  def createCastMethod(self, interface):
    method = SIDLLanguage.Method.Method(SIDL.Loader.createClass('ANL.SIDLCompilerI.Method'))
    method.setReturnType(interface)
    method.setIdentifier(interface.getIdentifier())
    method.setAbstract(0)
    method.setStatic(1)
    parameter = SIDLLanguage.Parameter.Parameter(SIDL.Loader.createClass('ANL.SIDLCompilerI.Parameter'))
    parameter.setIdentifier('optarg')
    parameter.setType(interface)
    parameter.setAccessMode(SIDLLanguage.ParameterAccessMode.IN)
    parameter.setRequired(0)
    method.addChildren([parameter])
    comment = SIDLLanguage.Comment.Comment(SIDL.Loader.createClass('ANL.SIDLCompilerI.Comment'))
    comment.setComment('Constructor and/or cast method for '+interface.getFullIdentifier())
    method.addComments([comment])
    return method

  def generateCast(self, interface, method, purpose):
    method.accept(self.__IORself)
    # Since we do not dispatch, we do not want to dereference the argument
    method.removeCodeBlock(self.language, purpose.getName()+' definition IOR dispatch begin')
    code = method.getCodeBlock(self.language, purpose.getName()+' definition IOR dispatch end')
    code.setSource('')
    code.println('/* Implementation code */')
    code.println('if (optarg) {')
    code.increaseTabLevel()
    code.indent()
    code.write('(*(optarg->d_epv->f_addReference))(')
    code.write(self.getIORStructName(interface, 'optarg'))
    code.write(');\n')
    code.decreaseTabLevel()
    if interface.isInstanceOf('SIDLLanguage.Class') and not interface.getAbstract():
      code.println('} else {')
      code.increaseTabLevel()
      code.println('optarg = (*(_epv->createObject))();')
      code.decreaseTabLevel()
    code.println('}')
    code.println('/* Since we may merely up the reference count and return the argument, add an extra reference */')
    code.println(self.getIORName(interface)+'__addRef(optarg);')
    code.write('_result = optarg;')
    self.generateMethod(interface, method, purpose, 'definition')
    return

  def generateMethods(self, interface, purpose):
    if not self.defnActive: return
    # Special methods
    if purpose.isInstanceOf('ANL.SIDLCompilerI.Stub'):
      pass
    elif purpose.isInstanceOf('ANL.SIDLCompilerI.Skeleton'):
      self.generateSkelInit(interface, purpose)
      self.generateSkelConstructors(interface, purpose)
      self.generateFunctionTableSetup(interface, purpose)
    elif purpose.isInstanceOf('ANL.SIDLCompilerI.Implementation'):
      pass

    # Normal methods
    methods    = []
    for parent in interface.getParents():
      methods.extend(parent.getMethods())
    isAbstract = 0
    if interface.isInstanceOf('SIDLLanguage.Class') and interface.getAbstract(): isAbstract = 1

    for method in interface.getMethods():
      same = 0
      for m in methods:
        if m.isSame(method):
          same = 1
          break
      if purpose.isInstanceOf('ANL.SIDLCompilerI.Stub') or (not same and not isAbstract):
        self.generateMethod(interface, method, purpose, 'declaration')
        self.generateMethod(interface, method, purpose, 'definition')
    return

  def generateMethodsImpl(self, sclass, purpose):
    if not self.defnActive: return
    for method in sclass.getImplementedMethods():
      if not sclass.getAbstract():
        self.generateMethodImpl(sclass, method, purpose, 'definition')
    # Special methods
    self.generateImplInit(sclass, purpose)
    return

  def getMethodComment(self, method):
    returnType = method.getReturnType()
    hasReturn  = not (returnType.getBaseType() and returnType.getIdentifier() == 'void')
    children   = method.getChildren()
    parameters = []
    for child in children:
      parameter = SIDLLanguage.Parameter.Parameter(child)
      if self.isOut(parameter.getAccessMode()):
        parameters.append(parameter)
    end = '\\n\\\n'

    str  = method.getIdentifier()+'('
    for i in range(len(children)):
      parameter = SIDLLanguage.Parameter.Parameter(children[i])
      str += self.getModeName(parameter.getAccessMode())+' '+parameter.getType().getIdentifier()+' '+parameter.getIdentifier()
      if i < len(children)-1: str += ', '
    str += ')'+end
    str += 'RETURNS'+end
    if hasReturn or len(parameters):
      if hasReturn:
        str += '  '+returnType.getFullIdentifier()+' _result'+end
      for parameter in parameters:
        str += '  '+self.getModeName(parameter.getAccessMode())+' '+parameter.getType().getFullIdentifier()+' '+parameter.getIdentifier()+end
    else:
      str += '   None'+end
    str += self.commentsToString(method)
    return '(char *) "'+str+'"'

  def generateVTableEntry(self, interface, method, purpose, code):
    code.println('{(char *) "'+method.getIdentifier()+'", (PyCFunction) '+self.getMethodName(interface, method, purpose)+
                 ', (METH_VARARGS | METH_KEYWORDS), '+self.getMethodComment(method)+'},')
    return

  def generateStaticVTable(self, interface, castMethod, purpose):
    name = interface.getIdentifier()
    code = self.getCode(purpose.getName()+' static virtual table')
    code.println('static PyMethodDef _'+name+'ModuleMethods[] = {')
    code.increaseTabLevel()
    self.generateVTableEntry(interface, castMethod, purpose, code)
    for method in interface.getMethods():
      if method.getStatic():
        self.generateVTableEntry(interface, method, purpose, code)
    code.println('{NULL, NULL}')
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    interface.addCode(code)
    return

  def generateVTable(self, interface, purpose):
    name = interface.getIdentifier()
    code = self.getCode(purpose.getName()+' virtual table')
    code.println('static PyMethodDef _'+name+'ObjectMethods[] = {')
    code.increaseTabLevel()
    for method in interface.getMethods():
      if not method.getStatic():
        self.generateVTableEntry(interface, method, purpose, code)
    code.println('{NULL, NULL}')
    code.decreaseTabLevel()
    code.println('};')
    code.println('')
    interface.addCode(code)
    return

  def generatePythonObject(self, code, interface, refType):
    name = interface.getIdentifier()
    code.println('return SIDL_Object_Create(sidlobj->d_epv->f__cast_ior('+self.getIORStructName(interface, 'sidlobj')+
                 ', "'+interface.getFullIdentifier()+'"), _'+name+'ObjectMethods, '+refType+');')
    return

  def generateWrapper(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition wrapper')
    name = interface.getFullIdentifier().replace('.', '_')+'__wrap'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('if (!sidlobj) {')
    code.increaseTabLevel()
    code.println('return Py_None;')
    code.decreaseTabLevel()
    code.println('} else {')
    code.increaseTabLevel()
    self.generatePythonObject(code, interface, 'SIDL_PyStealRef')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateWeakReference(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition weakRef')
    name = interface.getFullIdentifier().replace('.', '_')+'__weakRef'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    self.generatePythonObject(code, interface, 'SIDL_PyWeakRef')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateDeleteReference(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition deref')
    name = interface.getFullIdentifier().replace('.', '_')+'__deref'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('if (sidlobj) {')
    code.increaseTabLevel()
    code.indent()
    code.write('(*(sidlobj->d_epv->f_deleteReference))(')
    code.write(self.getIORStructName(interface, 'sidlobj'))
    code.write(');\n')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateNewReference(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition newRef')
    name = interface.getFullIdentifier().replace('.', '_')+'__newRef'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    self.generatePythonObject(code, interface, 'SIDL_PyNewRef')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateAddReference(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition addRef')
    name = interface.getFullIdentifier().replace('.', '_')+'__addRef'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('if (sidlobj) {')
    code.increaseTabLevel()
    code.indent()
    code.write('(*(sidlobj->d_epv->f_addReference))(')
    code.write(self.getIORStructName(interface, 'sidlobj'))
    code.write(');\n')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateConvert(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition convert')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('*sidlobj = SIDL_Cast(obj, "'+interface.getFullIdentifier()+'");')
    code.println('if (*sidlobj) {')
    code.increaseTabLevel()
    code.indent()
    code.write('(*((*sidlobj)->d_epv->f_addReference))(')
    code.write(self.getIORStructName(interface, '(*sidlobj)'))
    code.write(');\n')
    code.decreaseTabLevel()
    code.println('} else if (obj != Py_None) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_TypeError, "Argument is not of type '+interface.getFullIdentifier()+'");')
    code.println('return 0;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('return 1;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateConvertToElement(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition convertToElement')
    name = '_convertPython'
    code.println('static int')
    code.println(name+'(void *sidlarray, const int *ind, PyObject *pyobj) {')
    code.increaseTabLevel()
    code.indent()
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj;\n')
    code.println('')
    code.println('if ('+interface.getFullIdentifier().replace('.', '_')+'__convert(pyobj, &sidlobj)) {')
    code.increaseTabLevel()
    code.indent()
    code.write('(*_epv->setElement)((')
    self.composeCode(code, interface, 'ior array type', language = 'C')
    code.write(') sidlarray, ind, sidlobj);\n')
    code.println('return 0;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('return 1;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateConvertToArray(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition convertToArray')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert_python_array'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('PyObject *pya    = PyArray_FromObject(obj, PyArray_OBJECT, 0, 0);')
    code.println('int       result = 0;')
    code.println('')
    code.println('*sidlarray = NULL;')
    code.println('if (pya) {')
    code.increaseTabLevel()
    code.println('if (PyArray_OBJECT == ((PyArrayObject *) pya)->descr->type_num) {')
    code.increaseTabLevel()
    code.println('int dimen, *lower, *upper, *stride;')
    code.println('')
    code.println('if (SIDL_array__extract_python_info(pya, &dimen, &lower, &upper, &stride)) {')
    code.increaseTabLevel()
    code.println('*sidlarray = (*_epv->createArray)(dimen, lower, upper);')
    code.println('free(lower);')
    code.println('free(upper);')
    code.println('free(stride);')
    code.println('result = SIDL_array__convert_python(pya, dimen, *sidlarray, _convertPython);')
    code.println('if (*sidlarray && !result) {')
    code.increaseTabLevel()
    code.println('(*_epv->destroyArray)(*sidlarray);')
    code.println('*sidlarray = NULL;')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('Py_DECREF(pya);')
    code.decreaseTabLevel()
    code.println('}')
    code.println('return result;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateConvertFromElement(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition convertFromElement')
    name = '_convertSIDL'
    code.println('static int')
    code.println(name+'(void *sidlarray, const int *ind, PyObject **dest) {')
    code.increaseTabLevel()
    code.indent()
    self.composeCode(code, interface, 'ior type', language = 'C')
    code.write('sidlobj = (*_epv->getElement)((')
    self.composeCode(code, interface, 'ior array type', language = 'C')
    code.write(') sidlarray, ind);\n')
    code.println('')
    code.println('*dest = '+interface.getFullIdentifier().replace('.', '_')+'__wrap(sidlobj);')
    code.println('return (*dest == NULL);')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateConvertFromArray(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition convertFromArray')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert_sidl_array'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('PyObject *pya = NULL;')
    code.println('')
    code.println('if (sidlarray) {')
    code.increaseTabLevel()
    code.println('const int dimen   = (*_epv->getDimen)(sidlarray);')
    code.println('int      *lower   = (int *) malloc(dimen * sizeof(int));')
    code.println('int      *upper   = (int *) malloc(dimen * sizeof(int));')
    code.println('int      *numElem = (int *) malloc(dimen * sizeof(int));')
    code.println('int       i;')
    code.println('')
    code.println('for(i = 0; i < dimen; i++) {')
    code.increaseTabLevel()
    code.println('lower[i]   = (*_epv->getLower)(sidlarray, i);')
    code.println('upper[i]   = (*_epv->getUpper)(sidlarray, i);')
    code.println('numElem[i] = upper[i] - lower[i] + 1;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('pya = PyArray_FromDims(dimen, numElem, PyArray_OBJECT);;')
    code.println('if (pya) {')
    code.increaseTabLevel()
    code.println('if (!SIDL_array__convert_sidl(pya, dimen, lower, upper, numElem, sidlarray, _convertSIDL)) {')
    code.increaseTabLevel()
    code.println('Py_DECREF(pya);')
    code.println('pya = NULL;')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('free(lower);')
    code.println('free(upper);')
    code.println('free(numElem);')
    code.decreaseTabLevel()
    code.println('}')
    code.println('return pya;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateDestroyArray(self, interface, purpose):
    code = self.getCode(purpose.getName()+' method definition destroyArray')
    name = interface.getFullIdentifier().replace('.', '_')+'__destroy_sidl_array'
    code.println(name+'_RETURN')
    code.println(name+' '+name+'_PROTO {')
    code.increaseTabLevel()
    code.println('if (sidlarray) {')
    code.increaseTabLevel()
    code.println('(*_epv->destroyArray)(sidlarray);')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    interface.addCode(code)
    return

  def generateInitModule(self, interface, purpose):
    code    = self.getCode(purpose.getName()+' method definition initModule')
    name    = 'init'+interface.getIdentifier()
    comment = self.commentsToString(interface)

    code.println('void '+name+'(void) {')
    code.increaseTabLevel()
    code.println('PyObject *module, *dict, *c_api;')
    code.println('static void *ExternalAPI['+interface.getFullIdentifier().replace('.', '_')+'__API_NUM];')
    if not self.isBase(interface):
      code.indent()
      self.composeCode(code, interface, 'ior function table type', language = 'C')
      code.write(' (*_extFunc)(void) = NULL;\n')
    code.indent()
    if comment:
      code.write('module = Py_InitModule3(')
    else:
      code.write('module = Py_InitModule(')
    code.write('(char *) "'+interface.getIdentifier()+'", _'+interface.getIdentifier()+'ModuleMethods')
    if comment:
      code.write(', (char *) "'+comment+'"')
    code.write(');\n')
    if not self.isBase(interface):
      code.indent()
      code.write('_extFunc = (')
      self.composeCode(code, interface, 'ior function table type', language = 'C')
      code.write(' (*)(void)) SIDL_Loader__externals()->getStaticEPV()->f_lookupSymbol("')
      self.composeCode(code, interface, 'ior function table accessor', language = 'C')
      code.write('");\n')
      code.println('if (_extFunc) {')
      code.increaseTabLevel()
      code.println('_epv = (*_extFunc)();')
      code.decreaseTabLevel()
      code.println('}')
    else:
      code.indent()
      code.write('_epv = ')
      self.composeCode(code, interface, 'ior function table accessor', language = 'C')
      code.write('();\n')
    code.println('if (!_epv) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_TypeError, "Cannot load SIDL implementation for '+interface.getFullIdentifier()+'");')
    code.println('return;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('dict = PyModule_GetDict(module);')
    name = interface.getFullIdentifier().replace('.', '_')+'__wrap'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert_python_array'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__convert_sidl_array'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__destroy_sidl_array'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__weakRef'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__deref'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__newRef'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    name = interface.getFullIdentifier().replace('.', '_')+'__addRef'
    code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    code.println('import_SIDLObjA();')
    code.println('if (PyErr_Occurred()) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_ImportError, "Cannot import SIDLObjA module");')
    code.println('return;')
    code.decreaseTabLevel()
    code.println('}')
    if self.isException(interface):
      name = interface.getFullIdentifier().replace('.', '_')+'__type'
      code.println(name+' = PyErr_NewException((char *) "'+interface.getFullIdentifier()+'.Exception", SIDL_PyType(), NULL);')
      code.println('PyDict_SetItemString(dict, (char *) "Exception", '+name+');')
      code.println('ExternalAPI['+name+'_NUM] = '+name+';')
    code.println('c_api = PyCObject_FromVoidPtr((void *) ExternalAPI, NULL);')
    code.println('PyDict_SetItemString(dict, (char *) "_C_API", c_api);')
    code.println('Py_XDECREF(c_api);')
    if self.hasStaticMethod(interface):
      code.println('_sepv = (*_epv->getStaticEPV)();')
    code.println('if (PyErr_Occurred()) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_TypeError, "Cannot initialize Python module '+interface.getFullIdentifier()+'");')
    code.println('return;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('import_SIDLPyArrays();')
    code.println('if (PyErr_Occurred()) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_ImportError, "Cannot import SIDLPyArrays module");')
    code.println('return;')
    code.decreaseTabLevel()
    code.println('}')
    code.println('import_array();')
    code.println('if (PyErr_Occurred()) {')
    code.increaseTabLevel()
    code.println('PyErr_SetString(PyExc_ImportError, "Cannot import Numeric Python array module");')
    code.println('return;')
    code.decreaseTabLevel()
    code.println('}')
    for dep in interface.getDependencies():
      if not dep.isInstanceOf('SIDLLanguage.Enumeration') and not dep.getFullIdentifier() == 'SIDL.Loader':
        code.println(dep.getFullIdentifier().replace('.', '_')+'__import();')
    code.decreaseTabLevel()
    code.println('}')
    interface.addCode(code)
    return

  def generateSkelInit(self, sclass, purpose):
    iorName = sclass.getFullIdentifier().replace('.', '_')

    code = self.getCode(purpose.getName()+' method definition _importModules')
    code.println('static void _importModules(void) {')
    code.increaseTabLevel()
    code.println('static int _import = 1;')
    code.println('PyObject *module;')
    code.println('')
    code.println('if (_import) {')
    code.increaseTabLevel()
    code.println('_import = 0;')
    code.println('module = PyImport_ImportModule((char *) "user");')
    code.println('module = PyImport_ImportModule((char *) "importer");')
    code.println('if (PyErr_Occurred()) {PyErr_Print();}')
    code.println('import_SIDLObjA();')
    code.println('import_SIDLPyArrays();')
    code.println(iorName+'__import();')
    for dep in sclass.getDependencies():
      if not dep.isInstanceOf('SIDLLanguage.Enumeration') and not dep.getFullIdentifier() == 'SIDL.Loader':
        code.println(dep.getFullIdentifier().replace('.', '_')+'__import();')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)
    return

  def generateSkelConstructors(self, sclass, purpose):
    iorName = sclass.getFullIdentifier().replace('.', '_')
    iorType = sclass.getCodeBlock('C', 'ior type').getSource()
    package = sclass.getParent().getFullIdentifier()+'_impl'

    code = self.getCode(purpose.getName()+' method definition _ctor')
    code.println('static void '+iorName+'__ctor_skel('+iorType+'self) {')
    code.increaseTabLevel()
    if sclass.getAbstract():
      code.println('/* No implementation exists for an abstract class */')
    else:
      code.println('PyObject *_context;')
      code.println('PyObject *_pfunc;')
      code.println('PyObject *_args;')
      code.println('PyObject *_result;')
      code.println('')
      code.println('_context = PyImport_ImportModule((char *) "'+package+'.'+sclass.getIdentifier()+'_impl");')
      code.println('if (!_context) {')
      code.increaseTabLevel()
      code.println('if (PyErr_Occurred()) PyErr_Print();')
      code.println('return;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('_pfunc = PyObject_GetAttrString(_context, (char *) "'+sclass.getIdentifier()+'");')
      code.println('if (!_pfunc) {')
      code.increaseTabLevel()
      code.println('if (PyErr_Occurred()) PyErr_Print();')
      code.println('Py_DECREF(_context);')
      code.println('return;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('_args = Py_BuildValue((char *) "(O&)", (void *) '+iorName+'__weakRef, self);')
      code.println('if (!_args) {')
      code.increaseTabLevel()
      code.println('if (PyErr_Occurred()) PyErr_Print();')
      code.println('Py_DECREF(_pfunc);')
      code.println('Py_DECREF(_context);')
      code.println('return;')
      code.decreaseTabLevel()
      code.println('}')
      code.println('_result = PyInstance_New(_pfunc, _args, NULL);')
      code.println('if (_result) {')
      code.increaseTabLevel()
      code.println('self->d_data = _result;')
      code.decreaseTabLevel()
      code.println('} else {')
      code.increaseTabLevel()
      code.println('if (PyErr_Occurred()) PyErr_Print();')
      code.decreaseTabLevel()
      code.println('}')
      code.println('Py_DECREF(_args);')
      code.println('Py_DECREF(_pfunc);')
      code.println('Py_DECREF(_context);')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)

    code = self.getCode(purpose.getName()+' method definition _dtor')
    code.println('static void '+iorName+'__dtor_skel('+iorType+'self) {')
    code.increaseTabLevel()
    if sclass.getAbstract():
      code.println('/* No data was allocated */')
    else:
      code.println('/* Remove only reference to Python object */')
      code.println('Py_XDECREF(((PyObject *) self->d_data));')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)
    return

  def generateFunctionTableEntry(self, code, sclass, method, purpose):
    if method.getStatic():
      table = 'sepv'
    else:
      table = 'epv'
    if method.getAbstract():
      code.println(table+'->f_'+method.getIdentifier()+' = NULL;')
    else:
      code.println(table+'->f_'+method.getIdentifier()+' = '+self.getMethodName(sclass, method, purpose)+';')
    return

  def generateFunctionTableSetup(self, sclass, purpose):
    iorName = sclass.getFullIdentifier().replace('.', '_')
    vtable  = sclass.getCodeBlock('C', 'ior epv table type').getSource()
    stable  = sclass.getCodeBlock('C', 'ior static function table type').getSource()

    code = self.getCode(purpose.getName()+' method definition initImplementation')
    code.println('static void')
    code.println('initImplementation(void) {')
    code.increaseTabLevel()
    code.println('static int notInitialized = 1;')
    code.println('')
    code.println('if (notInitialized) {')
    code.increaseTabLevel()
    code.println('notInitialized = 0;')
    code.println('SIDL_Python_Init();')
    code.decreaseTabLevel()
    code.println('}')
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)

    code = self.getCode(purpose.getName()+' method definition _set_epv')
    code.println('void '+iorName+'__set_epv('+vtable+' epv);')
    code.println('void')
    code.println(iorName+'__set_epv('+vtable+' epv) {')
    code.increaseTabLevel()
    code.println('initImplementation();')
    code.println('_importModules();')
    code.println('/* Initialize constructor and destructor */')
    self.generateFunctionTableEntry(code, sclass, self.createConstructorMethod(sclass), purpose)
    self.generateFunctionTableEntry(code, sclass, self.createDestructorMethod(sclass), purpose)
    methods = sclass.getImplementedMethods()
    header  = 0
    for method in methods:
      if not method.getStatic():
        if not header:
          code.println('/* Initialize user methods */')
          header = 1
        self.generateFunctionTableEntry(code, sclass, method, purpose)
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)

    if not self.hasStaticMethod(sclass): return
    code = self.getCode(purpose.getName()+' method definition _set_sepv')
    code.println('void '+iorName+'__set_sepv('+stable+' sepv);')
    code.println('void')
    code.println(iorName+'__set_sepv('+stable+' sepv) {')
    code.increaseTabLevel()
    code.println('initImplementation();')
    code.println('_importModules();')
    methods = sclass.getImplementedMethods()
    code.println('/* Initialize user methods */')
    for method in methods:
      if method.getStatic():
        self.generateFunctionTableEntry(code, sclass, method, purpose)
    code.decreaseTabLevel()
    code.println('}')
    code.println('')
    sclass.addCode(code)
    return

  def generateImplInit(self, sclass, purpose):
    iorName = sclass.getFullIdentifier().replace('.', '_')

    code = self.getCode(purpose.getName()+' method definition begin __init__')
    code.println('def __init__(self, IORself):')
    code.increaseTabLevel()
    code.println('self.__IORself = IORself')
    sclass.addCode(code)
    return
# DO-NOT-DELETE splicer.end(_final)

