#include "SIDL_Python.h"
#ifndef INCLUDED_SIDL_CONFIG_H
#include "sidl_config.h"
#endif
#include <stdlib.h>

#if HAVE_LIBDL && defined(PYTHON_SHARED_LIBRARY)
#include <stdio.h>

#ifdef HAVE_DLFCN_H
#  include <dlfcn.h>
#endif

#ifdef HAVE_SYS_DL_H
#  include <sys/dl.h>
#endif

/* This is ridiculous. This should all be figured out by configure beforehand. */
#ifdef RTLD_GLOBAL
#  define LT_GLOBAL		RTLD_GLOBAL
#else
#  ifdef DL_GLOBAL
#    define LT_GLOBAL		DL_GLOBAL
#  endif
#endif /* !RTLD_GLOBAL */
#ifndef LT_GLOBAL
#  define LT_GLOBAL		0
#endif /* !LT_GLOBAL */

/* We may have to define LT_LAZY_OR_NOW in the command line if we
   find out it does not work in some platform. */
#ifndef LT_LAZY_OR_NOW
#  ifdef RTLD_LAZY
#    define LT_LAZY_OR_NOW	RTLD_LAZY
#  else
#    ifdef DL_LAZY
#      define LT_LAZY_OR_NOW	DL_LAZY
#    endif
#  endif /* !RTLD_LAZY */
#endif
#ifndef LT_LAZY_OR_NOW
#  ifdef RTLD_NOW
#    define LT_LAZY_OR_NOW	RTLD_NOW
#  else
#    ifdef DL_NOW
#      define LT_LAZY_OR_NOW	DL_NOW
#    endif
#  endif /* !RTLD_NOW */
#endif
#ifndef LT_LAZY_OR_NOW
#  define LT_LAZY_OR_NOW	0
#endif /* !LT_LAZY_OR_NOW */

void SIDL_Python_Init(void)
{
  static int pythonInitialized = 0;

  if (!pythonInitialized) {
    static const char libName[]   = PYTHON_SHARED_LIBRARY;
    static const char initName[]  = "Py_Initialize";
    static const char finalName[] = "Py_Finalize";
    void             *handle      = dlopen(libName, LT_GLOBAL | LT_LAZY_OR_NOW);
    void            (*pyinit)(void);
    void            (*pyfini)(void);

    if (handle) {
      pyinit = (void (*)(void)) dlsym(handle, initName);
      if (pyinit) {
        pythonInitialized = 1;
        (*pyinit)();
        
        pyfini = (void (*)(void)) dlsym(handle, finalName);
        if (pyfini) {
          atexit(pyfini);
        } else {
          fprintf(stderr, "Unable to find symbol %s in %s: %s", finalName, libName, dlerror());
        }
        return;
      } else {
        fprintf(stderr, "Unable to find symbol %s in %s: %s", initName, libName, dlerror());
        dlclose(handle);
      }
    } else {
      fprintf(stderr, "Unable to load library %s: %s", libName, dlerror());
    }
  }
}

#else

#include "Python.h"

void SIDL_Python_Init(void)
{
  static int pythonInitialized = 0;

  if (!pythonInitialized) {
    Py_Initialize();
    atexit(Py_Finalize);
    pythonInitialized = 1;
  }
}

#endif
