# DO-NOT-DELETE splicer.begin(_initial)
# Put your code here...
pass
# DO-NOT-DELETE splicer.end(_initial)

import SIDLLanguage.AST
import SIDL.NullIORException
import SIDL.BaseException
import SIDL.BaseInterface
import SIDLLanguage.VisitorException
import SIDLLanguage.Specification
import SIDLLanguage.Version
import SIDLLanguage.Package
import SIDLLanguage.Interface
import SIDLLanguage.Class
import SIDLLanguage.Method
import SIDLLanguage.Type
import SIDLLanguage.Enumeration
import SIDLLanguage.Enumerator
import SIDLLanguage.Array
import SIDLLanguage.Parameter
import SIDLLanguage.Comment

# DO-NOT-DELETE splicer.begin(_before_type)
import sys
# DO-NOT-DELETE splicer.end(_before_type)

class Resolver:
  '''
  // Name resolution within a set of ASTs
  '''
  # All calls to SIDL methods should use __IORself

  def __init__(self, IORself):
    self.__IORself = IORself
    # DO-NOT-DELETE splicer.begin(__init__)
    import SIDL.Loader
    self.repository    = []
    self.markName      = 'resolverMark'
    self.loader        = None
    self.baseClass     = None
    self.baseInterface = None
    self.baseException = None
    self.templater     = SIDLLanguage.Visitor.Visitor(SIDL.Loader.createClass('ANL.TemplatingI.Templater'))
    # DO-NOT-DELETE splicer.end(__init__)

  def resolve(self, node, root):
    '''
    // Resolve \"node\" to its actual referent in the AST
    //   The \"root\" argument is an optional AST for resolving \"node\"
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.AST
    # root SIDLLanguage.AST
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # SIDLLanguage.AST _return
    # DO-NOT-DELETE splicer.begin(resolve)
    # We could have static instances for the base types
    if node.getBaseType():
      return node
    elif node.isInstanceOf('SIDLLanguage.Array'):
      array = SIDLLanguage.Array.Array(node)
      array.setType(self.resolve(array.getType(), root))
      return node

    idChain = node.getIdentifier().split('.')
    name    = idChain[-1]
    scope   = idChain[:-1]

    # Find enclosing scope
    if scope:
      # Find enclosing packages
      trees = []
      if root: trees.append(root)
      trees.extend(self.getRepository())
      for tree in trees:
        for packageName in scope:
          package = self.findPackage(tree, packageName)
          if package:
            root = package
          else:
            root = None
            break
        if root:
          node = self.findType(root, name)
          if node: return node
      if not root:
        self.raiseVisitorException('Could not resolve '+name+' in '+str(scope))
    else:
      # Must lie in same package
      root = node
      while not root.isInstanceOf('SIDLLanguage.Package'):
        root = root.getParent()
      node = self.findType(root, name)
      if node: return node

    self.raiseVisitorException('Could not resolve '+str(idChain))
    # DO-NOT-DELETE splicer.end(resolve)

  def getRepository(self):
    '''
    // Retrieve auxilliary ASTs
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # array _return
    # DO-NOT-DELETE splicer.begin(getRepository)
    return self.repository
    # DO-NOT-DELETE splicer.end(getRepository)

  def setRepository(self, repository):
    '''
    // Set auxilliary ASTs
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # repository array
    # DO-NOT-DELETE splicer.begin(setRepository)
    self.repository = repository
    # DO-NOT-DELETE splicer.end(setRepository)

  def visitAST(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.AST
    # DO-NOT-DELETE splicer.begin(visitAST)
    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitAST)

  def visitSpecification(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Specification
    # DO-NOT-DELETE splicer.begin(visitSpecification)
    self.resolveLoader(node, force = 1)
    self.resolveBaseInterface(node, force = 1)
    self.resolveBaseClass(node, force = 1)
    self.resolveBaseException(node, force = 1)
    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitSpecification)

  def visitVersion(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Version
    # DO-NOT-DELETE splicer.begin(visitVersion)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitVersion)

  def visitPackage(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Package
    # DO-NOT-DELETE splicer.begin(visitPackage)
    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitPackage)

  def visitInterface(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Interface
    # DO-NOT-DELETE splicer.begin(visitInterface)
    if self.isMarked(node): return

    root = self.getRoot(node)
    node.addDependency(self.loader)

    parents = []
    for parent in node.getParents():
      parents.append(self.resolve(parent, root))
    found = 0
    # Could do recursive search here
    if node.isSame(self.baseInterface) or len(parents) > 0:
      found = 1
    if not found: parents.append(self.baseInterface)
    node.setParents(parents)

    try:
      for parent in node.getParents():
        parent.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())

    if self.templater:
      for method in node.getChildren():
        method.accept(self.templater)
    self.defaultVisit(node)

    self.markNode(node)
    # DO-NOT-DELETE splicer.end(visitInterface)

  def visitClass(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Class
    # DO-NOT-DELETE splicer.begin(visitClass)
    if self.isMarked(node): return
    self.markNode(node)

    root = self.getRoot(node)
    node.addDependency(self.loader)

    parents = []
    for parent in node.getParents():
      parents.append(self.resolve(parent, root))
    found = 0
    # Could do recursive search here
    if node.isSame(self.baseClass) or len(parents) > 0:
      found = 1
    if not found:
      parents.append(self.baseClass)
    node.setParents(parents)

    interfaces = []
    for interface in node.getInterfaces():
      interfaces.append(self.resolve(interface, root))
    node.setInterfaces(interfaces)

    if self.templater:
      for method in node.getChildren():
        method.accept(self.templater)
    self.defaultVisit(node)
    try:
      for parent in node.getParents():
        parent.accept(self.__IORself)
      for interface in node.getInterfaces():
        interface.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())
    # DO-NOT-DELETE splicer.end(visitClass)

  def visitMethod(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Method
    # DO-NOT-DELETE splicer.begin(visitMethod)
    root = self.getRoot(node)

    typeNode = self.resolve(node.getReturnType(), root)
    node.setReturnType(typeNode)
    if typeNode.isInstanceOf('SIDLLanguage.Enumeration'):
      typeNode.accept(self.__IORself)

    exceptions = []
    for exception in node.getExceptions():
      exceptions.append(self.resolve(exception, root))
    if not self.isBase(node.getParent()) and not self.baseException in exceptions:
      exceptions.append(self.baseException)
    node.setExceptions(exceptions)
    for exception in node.getExceptions():
      self.resolveException(exception)

    self.defaultVisit(node)
    # DO-NOT-DELETE splicer.end(visitMethod)

  def visitType(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Type
    # DO-NOT-DELETE splicer.begin(visitType)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitType)

  def visitEnumeration(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumeration
    # DO-NOT-DELETE splicer.begin(visitEnumeration)
    if self.isMarked(node): return

    valuesUsed = []
    for child in node.getChildren():
      enumerator = SIDLLanguage.Enumerator.Enumerator(child)
      if enumerator.getInitialized():
        valuesUsed.append(enumerator.getValue())
    valuesUsed.sort()

    values = []
    for child in node.getChildren():
      enumerator = SIDLLanguage.Enumerator.Enumerator(child)
      if enumerator.getInitialized():
        value = enumerator.getValue()
        if value in values:
          self.raiseVisitorException('Repeated enumeration value '+value+' for "'+enumerator.getIdentifier()+'" in '+node.getFullIdentifier())
      else:
        if len(valuesUsed):
          value = valuesUsed[-1] + 1
        else:
          value = 0
        if value in valuesUsed: self.raiseVisitorException('Invalid processing of enumeration')
        valuesUsed.append(value)
      enumerator.setValue(value)
      values.append(value)

    self.markNode(node)
    # DO-NOT-DELETE splicer.end(visitEnumeration)

  def visitEnumerator(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Enumerator
    # DO-NOT-DELETE splicer.begin(visitEnumerator)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitEnumerator)

  def visitArray(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Array
    # DO-NOT-DELETE splicer.begin(visitArray)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitArray)

  def visitParameter(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Parameter
    # DO-NOT-DELETE splicer.begin(visitParameter)
    typeNode = self.resolve(node.getType(), self.getRoot(node))
    node.setType(typeNode)
    if typeNode.isInstanceOf('SIDLLanguage.Enumeration'):
      typeNode.accept(self.__IORself)
    # DO-NOT-DELETE splicer.end(visitParameter)

  def visitComment(self, node):
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.Comment
    # DO-NOT-DELETE splicer.begin(visitComment)
    # Put your code here...
    pass
    # DO-NOT-DELETE splicer.end(visitComment)

# DO-NOT-DELETE splicer.begin(_final)
  def raiseVisitorException(self, message, trace = ''):
    e = SIDLLanguage.VisitorException.VisitorException()
    e.setMessage(message)
    if isinstance(trace, list):
      for line in trace: e.addToStackTrace(line.strip())
    elif trace:
      e.addToStackTrace(trace)
    e.addToTrace(__file__, sys._getframe(1).f_lineno, sys._getframe().f_code.co_name)
    raise SIDLLanguage.VisitorException.Exception, e

  def defaultVisit(self, ast):
    try:
      for child in ast.getChildren():
        child.accept(self.__IORself)
    except SIDLLanguage.ASTException.Exception, e:
      self.raiseVisitorException(e.getMessage(), e.getStackTrace().strip())
    return

  def isMarked(self, node):
    return node.hasAttribute(self.markName)

  def markNode(self, node):
    seen = SIDLLanguage.Attribute.Attribute(SIDL.Loader.createClass('ANL.SIDLCompilerI.Attribute'))
    seen.setName(self.markName)
    node.addAttribute(seen)
    return

  def isBase(self, interface):
    return interface.getFullIdentifier() in ['SIDL.BaseInterface', 'SIDL.BaseClass', 'SIDL.BaseException', 'SIDL.NullIORException', 'SIDL.Loader', 'SIDL.DLL']

  def getRoot(self, node):
    root = node
    while root.getParent():
      root = root.getParent()
    return root

  def findPackage(self, root, package):
    # Breadth first search
    if root.isInstanceOf('SIDLLanguage.Specification') or root.isInstanceOf('SIDLLanguage.Package'):
      if root.getIdentifier() == package:
        return root
      else:
        for child in root.getChildren():
          node = self.findPackage(child, package)
          if node:
            return node
    return None

  def findType(self, node, typeName):
    for child in node.getChildren():
      if typeName == child.getIdentifier():
        return child
    return None

  def resolveLoader(self, root = None, force = 0):
    if not self.loader or force:
      loader = SIDLLanguage.Interface.Interface(SIDL.Loader.createClass('ANL.SIDLCompilerI.Interface'))
      loader.setIdentifier('SIDL.Loader')
      loader.setBaseType(0)
      self.loader = self.resolve(loader, root)
    return

  def resolveBaseInterface(self, root = None, force = 0):
    if not self.baseInterface or force:
      baseInterface = SIDLLanguage.Interface.Interface(SIDL.Loader.createClass('ANL.SIDLCompilerI.Interface'))
      baseInterface.setIdentifier('SIDL.BaseInterface')
      baseInterface.setBaseType(0)
      self.baseInterface = self.resolve(baseInterface, root)
    return

  def resolveBaseClass(self, root = None, force = 0):
    if not self.baseClass or force:
      baseClass = SIDLLanguage.Interface.Interface(SIDL.Loader.createClass('ANL.SIDLCompilerI.Class'))
      baseClass.setIdentifier('SIDL.BaseClass')
      baseClass.setBaseType(0)
      self.baseClass = self.resolve(baseClass, root)
    return

  def resolveBaseException(self, root = None, force = 0):
    if not self.baseException or force:
      baseException = SIDLLanguage.Interface.Interface(SIDL.Loader.createClass('ANL.SIDLCompilerI.Class'))
      baseException.setIdentifier('SIDL.BaseException')
      baseException.setBaseType(0)
      self.baseException = self.resolve(baseException, root)
    return

  def resolveException(self, exception):
    '''Must resolve all exception parents due to bad exception paradigm (needing BaseException to pass anything)'''
    root    = self.getRoot(exception)
    parents = []
    for parent in exception.getParents():
      parents.append(self.resolve(parent, root))
    exception.setParents(parents)
    for parent in exception.getParents():
      self.resolveException(parent)
    return
# DO-NOT-DELETE splicer.end(_final)

