# DO-NOT-DELETE splicer.begin(_initial)
# Put your code here...
pass
# DO-NOT-DELETE splicer.end(_initial)

import SIDL.NullIORException
import SIDL.BaseException
import ANL.SplicerI.TransferMode
import SIDL.BaseInterface
import SIDLLanguage.AST
import SIDLLanguage.CodeGeneratorException

# DO-NOT-DELETE splicer.begin(_before_type)
import os
import re
# DO-NOT-DELETE splicer.end(_before_type)

class Splicer:
  '''
  // Implementation of the Babel splicer technology
  '''
  # All calls to SIDL methods should use __IORself

  def __init__(self, IORself):
    self.__IORself = IORself
    # DO-NOT-DELETE splicer.begin(__init__)
    self.tag          = 'DO-NOT-DELETE splicer'
    self.beginRE      = re.compile(self.tag+r'\.begin\((?P<name>.*)\)')
    self.endRE        = re.compile(self.tag+r'\.end\((?P<name>.*)\)')
    self.doTransfer   = 0
    self.transferMode = ANL.SplicerI.TransferMode.BABEL_TO_SCANDAL
    # DO-NOT-DELETE splicer.end(__init__)

  def getDoTransfer(self):
    '''
    // Retrieve the flag determining whether an implementation transfer will be done
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # bool _return
    # DO-NOT-DELETE splicer.begin(getDoTransfer)
    # Put your code here...
    return self.doTransfer
    # DO-NOT-DELETE splicer.end(getDoTransfer)

  def setDoTransfer(self, doTransfer):
    '''
    // Retrieve the flag determining whether an implementation transfer will be done
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # doTransfer bool
    # DO-NOT-DELETE splicer.begin(setDoTransfer)
    # Put your code here...
    self.doTransfer = doTransfer
    # DO-NOT-DELETE splicer.end(setDoTransfer)

  def getTransferMode(self):
    '''
    // Retrieve the transfer mode
    '''
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # ANL.SplicerI.TransferMode _return
    # DO-NOT-DELETE splicer.begin(getTransferMode)
    # Put your code here...
    return self.transferMode
    # DO-NOT-DELETE splicer.end(getTransferMode)

  def setTransferMode(self, transferMode):
    '''
    // Set the transfer mode
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # transferMode ANL.SplicerI.TransferMode
    # DO-NOT-DELETE splicer.begin(setTransferMode)
    # Put your code here...
    self.transferMode = transferMode
    # DO-NOT-DELETE splicer.end(setTransferMode)

  def parse(self, filename, language, node):
    '''
    // Retrieves code blocks from a source file, and load them into an AST node
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # filename string
    # language string
    # node SIDLLanguage.AST
    # DO-NOT-DELETE splicer.begin(parse)
    if self.doTransfer and self.transferMode == ANL.SplicerI.TransferMode.BABEL_TO_SCANDAL:
      newFilename = self.filenameToBabel(language, filename)
      #print 'Error: '+filename+' --> '+newFilename
      filename    = newFilename
    if not os.path.isfile(filename): return
    f     = file(filename)
    lines = iter(f.xreadlines())
    try:
      while 1:
        line = lines.next()
        m = self.beginRE.search(line)
        if m:
          line = lines.next()
          code = SIDLLanguage.Code.Code(SIDL.Loader.createClass('ANL.SIDLCompilerI.Code'))
          code.setLanguage(language)
          code.setLabel(self.getLabel(m.group('name')))
          while not self.endRE.search(line):
            code.write(line)
            line = lines.next()
          node.addCode(code)
    except StopIteration:
      pass
    f.close()
    # DO-NOT-DELETE splicer.end(parse)

  def getLabel(self, name):
    '''
    // Get the code block label corresponding to a given splicer name
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # name string
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # string _return
    # DO-NOT-DELETE splicer.begin(getLabel)
    return self.tag+' '+name
    # DO-NOT-DELETE splicer.end(getLabel)

  def splice(self, node, language, name, defaultCode):
    '''
    // Return the code block corresponding to the splicer name, or the defaultCode
    '''
    # SIDL EXPECTED INCOMING TYPES
    # ============================
    # node SIDLLanguage.AST
    # language string
    # name string
    # defaultCode string
    # SIDL EXPECTED RETURN TYPES (returned as a tuple or single value)
    # ================================================================
    # string _return
    # DO-NOT-DELETE splicer.begin(splice)
    code = ''
    code = code + self.getComment(language, self.tag+'.begin('+name+')')
    if node.hasCodeBlock(language, self.getLabel(name)):
      code = code + node.getCodeBlock(language, self.getLabel(name)).getSource()
    elif defaultCode:
      code = code + defaultCode
      if not defaultCode[-1] == '\n':
        code = code + '\n'
    code = code + self.getComment(language, self.tag+'.end('+name+')')
    return code
    # DO-NOT-DELETE splicer.end(splice)

# DO-NOT-DELETE splicer.begin(_final)
  def getComment(self, language, line, newline = 1):
    if language == 'C':
      line = '/* '+line+' */'
    elif language == 'C++':
      line = '// '+line
    elif language == 'F77':
      line = 'C '+line
    elif language == 'F90':
      line = '! '+line
    elif language == 'Python':
      line = '# '+line
    elif language == 'Matlab':
      line = '% '+line
    elif language == 'Java':
      line = '// '+line
    if newline: line = line+'\n'
    return line

  def filenameToBabel(self, language, filename):
    (fullDir, file) = os.path.split(filename)
    (dir, implDir)  = os.path.split(fullDir)
    fullDir         = os.path.join(dir, implDir.replace('_impl', '_Impl'))
    if language == 'C++':
      # Handle directory
      (head, lastDir) = os.path.split(fullDir)
      tail            = ''
      while not re.match(r'^server-c\+\+-.*$', lastDir):
        tail            = os.path.join(lastDir, tail)
        (head, lastDir) = os.path.split(head)
      fullDir = os.path.join(head, lastDir, 'src', tail)
      # Handle header
      (base, ext) = os.path.splitext(file)
      if ext == '.hh':
        file = tail[:-1].replace('/', '_')+ext
      # Handle constructor and destructor
      if base == '_ctor' or base == '_dtor':
        file = tail[:-1].replace('/', '_')+ext
    elif language == 'Python':
      file = file.replace('_impl', '_Impl')
    return os.path.join(fullDir, file)
# DO-NOT-DELETE splicer.end(_final)

